#|
	Stole these routines from Gerald Roylance's clmath.

						Andy Long, 11/97.

;;; -*- Mode: LISP; Syntax: Common-lisp; Package: USER -*-

;;;; Beta Functions

;;; Reference
;;;   M. Abramowitz and I. Stegun, eds,
;;;     Handbook of Mathematical Functions,
;;;     National Bureau of Standards, 1964
;;;   William H. Press, Brian P. Flannery, Saul A. Teukolsky, William
;;;   T. Vetterling 
;;;     Numerical Recipes in C, The Art of Scientific Computing
;;;     Cambridge University Press, 1988
;;;     cf 178

;;;  (c) Copyright Gerald Roylance 1982, 1984, 1985, 1989
;;;      All Rights Reserved.
;;;  This file may be distributed noncommercially provided
;;;  that this notice is not removed.

;;; Bugs and fixes
;;;   betai has a bug:  returns complex numbers.
;;; no bug: he thought that a,b = 0 was permissible.
|#


(defun clmath-poly-expand (var coefs)
  (cond ((null coefs)       0.0)
	((null (cdr coefs)) (car coefs))
	(t
	 `(+ ,(car coefs) (* ,var ,(clmath-poly-expand var (cdr coefs)))))))

(defmacro clmath-poly (var . coefs)
  (clmath-poly-expand var coefs))



;;;; Beta Function

;;; The beta function is B(a,b)
;;;    B(a,b) = B (b,a)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                beta
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun beta(m n)
  "
;; The beta function:

	B(z,w)	= integral from 0 to 1 of t^(z-1)(1-t)^(w-1)dt
		= Gamma(z)Gamma(w)/Gamma(z+w)

;; Examples:
 (beta 3 4)
 (spin-domain .5 3 .5 3)
 (plot (func f(x y) (beta x y)))
"
 (exp (log-beta m n)))			;m,n>0

(defun beta-naive (m n)		;m,n>0
  "
;; The beta function:

	B(z,w)	= integral from 0 to 1 of t^(z-1)(1-t)^(w-1)dt
		= Gamma(z)Gamma(w)/Gamma(z+w)

Works for small m and n, but otherwise overflows before it can give the correct
answer.

;; See Also: beta, log-beta

;; Examples:
 (beta-naive 300 4)	; #.NOT-A-NUMBER
 (beta 300 4)		; 7.261292453418591E-10
 (spin-domain .5 3 .5 3)
 (plot (func f(x y) (beta-naive x y) :dash t))
"
  (/ (* (gamma m) (gamma n))
     (gamma (+ m n))))

(defun log-beta (m n)
  "
Helper for beta: computes the log of beta, which is better conditioned:
	Log Beta(z,w) = log-gamma(z) + log-gamma(w) - log-gamma(z+w)
"
  (- (+ (log-gamma m) (log-gamma n))
     (log-gamma (+ m n))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                betai
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun betai (a b x)
  "
	fun betai (a b x)

where a,b > 0 and 0<=x<=1

;;;; Incomplete Beta Function I_x

;;; These routines are lifted from Press, et al.

;;; reference:
;;;   William H. Press, Brian P. Flannery, Saul A. Teukolsky, William
;;;   T. Vetterling 
;;;     Numerical Recipes in C, The Art of Scientific Computing
;;;     Cambridge University Press, 1988
;;;     cf 178

;;; I_x(a,b) = {x^a (1-x)^b \over a B(a,b)}
;;;              [{1\over 1+}{d_1\over 1+}{d_2\over 1+}...]
;;; d_{2m+1} = - {(a+m)(a+b+m)x \over (a+2m)(a+2m+1)}
;;; d_{2m  } =   {m(b-m)x \over (a+2m-1)(a+2m)}

;;; iterations are O(sqrt(max(a,b))) for x < (a+1)/(a+b+1)

;;; symmetry
;;;   I_x(a,b) = 1 - I_{1-x}(b,a)

;; See also: betacf

;; Examples:
 (betai 1 1 .98)
 (betai 2 1 .98)

 (title \"Plot of Incomplete Beta Functions\")
 (plot
   (lambda(x) (betai .5 5 x))
   (lambda(x) (betai .5 .5 x))
   (lambda(x) (betai 1 3 x))
   (lambda(x) (betai 8 10 x))
   (lambda(x) (betai 5 .5 x))
   0 1)
"
  (declare (float a b x))
  (if (or (< x 0.0) (> x 1.0))
      (error "BETAI:  Bad argument, x=~f" x)
      (let ((bt (if (or (= x 0.0) (= x 1.0))
		    0.0
		    (exp (+ (- (log-gamma (+ a b))
			       (log-gamma a)
			       (log-gamma b))
			    (* a (log x))
			    (* b (log (- 1.0 x))))))))
	(declare (float bt))
	(if (< x (/ (+ a 1.0) (+ a b 2.0)))	; which converges quickest?
	    (* bt (/ (betacf a b x) a))
	    (- 1.0 (/ (* bt (betacf b a (- 1.0 x))) b))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                betacf
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun betacf (a b x)
  "
;; Helper for betai, incomplete beta function.
"
  (declare (float a b x))
  (let* ((eps 3.0e-7)
	 (qab (+ a b))
	 (qap (+ a 1.0))
	 (qam (- a 1.0))
	 (bz  (- 1.0 (/ (* qab x) qap)))
	 (bm  1.0)
	 (az  1.0)
	 (am  1.0))
    (declare (float eps qab qap qam bz bm az am))
    (do ((m 1 (1+ m)))
	((>= m 100)
	 (error "too many iterations"))
      (declare (fixnum m))

      (let* ((em  (float m))
	     (tem (* 2.0 em))
	     (d   (/ (* em (- b em) x)
		     (* (+ qam tem) (+ a tem))))
	     (ap  (+ az (* d am)))		; even step of recurrance
	     (bp  (+ bz (* d bm)))
	     (d   (- (/ (* (+ a em) (+ qab em) x)
			(* (+ qap tem) (+ a tem)))))
	     (app (+ ap (* d az)))		; odd step of recurrance
	     (bpp (+ bp (* d bz))))
	(declare (float em tem d ap bp app bpp))

	(let ((aold az))			; save old answer
	  (declare (float aold))
	  ;; renormalize
	  (setq am (/ ap  bpp))
	  (setq bm (/ bp  bpp))
	  (setq az (/ app bpp))
	  (setq bz 1.0)
	  (if (< (abs (- az aold)) (* EPS (abs az)))
	      (return az))
	  )))
    ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                elliptic-integral-k
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun elliptic-integral-k (m)
  "
;;;; ELLIPTIC INTEGRAL K(M)

	This is equal to the Legendre elliptic integral of the first kind
	F(pi/2,m).

;;; (integral 0 pi/2 (1 - m*sin^2(theta))^-0.5 theta)
;;; k'(m) = k(1-m)
;;; 0 <= M < 1 

;; Examples:
 (elliptic-integral-k 1)
 (plot elliptic-integral-k  0 .999)
"
  (if (= 1 m) POSITIVE-INFINITY (elliptic-integral-kC (- 1.0 (float m)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                elliptic-integral-kC
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun elliptic-integral-kC (m) 			;eps < 2E-8
  (+ (clmath-poly m
	   1.38629436112
	   0.09666344259
	   0.03590092383
	   0.03742563713
	   0.01451196212)
     (* (- (log m))
	(clmath-poly m
	      0.50000000000
	      0.12498593597
	      0.06880248576
	      0.03328355346
	      0.00441787012))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                elliptic-integral-e
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun elliptic-integral-e (m)
  "
;;;; ELLIPTIC INTEGRAL E(M)
;;; integral from 0 to pi/2 d[theta] (sqrt 1 - m*sin^2(theta))

	This is equal to the Legendre elliptic integral of the second kind
	E(pi/2,m).

;; Examples:
 (plot elliptic-integral-e 0 1)
"
  (if (= m 1) 1 (elliptic-integral-ec (- 1.0 m))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                elliptic-integral-eC
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun elliptic-integral-eC (m) 			;eps < 2E-8
  (+ (clmath-poly m
	   1.00000000000
	   0.44325141463
	   0.06260601220
	   0.04757383546
	   0.01736506451)
     (* (- (log m))
	(clmath-poly m
	      0.00000000000
	      0.24998368310
	      0.09200180037
	      0.04069697526
	      0.00526449639))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                elliptic-sine
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun elliptic-sine (u M)
  "
;;;; ELLIPTIC-SINE (u M)
;;; u(phi, M) = (Integral (0 to phi) (1 - M sin^2 x)^-0.5  (d x))
;;; elliptic-sine(u M) = sin(phi)

This is equal to the Legendre elliptic integral of the first kind
	F(pi/2,m).

;; Examples:
 (plot (mapcar (lambda(x) (elliptic-sine pi/2 x)) (rseq 0.01 .99 100)))
"
  (declare (single-float K K1 v q v2N1 qN05 sn))
  (let* ((K   (elliptic-integral-k M))
	 (K1  (elliptic-integral-kc M))
	 (v   (/ (* (float pi 1.0) u) (* 2.0 K)))
	 (q   (exp (- (/ (* (float pi 1.0) K1) K)))))
    (do ((v2N1 v (+ v2N1 v v))			;(2N+1)v
	 (qN05 (sqrt q) (* qN05 q))		;q^(N + 0.5)
	 (sn 0.0) )
	((progn (setq sn (+ sn (/ (* qN05 (sin v2N1))
				  (- 1.0 (* qN05 qN05)))))
		(< qN05 1.0E-8))
	 (/ (* (* 2.0 (float pi 1.0)) sn)
	    (* (sqrt M) K)))
      )))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                elliptic-cosine
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun elliptic-cosine (u M)
  (declare (single-float K K1 v q v2N1 qN05 sn))
  (let* ((K   (elliptic-integral-k M))
	 (K1  (elliptic-integral-kc M))
	 (v   (/ (* (float pi 1.0) u) (* 2.0 K)))
	 (q   (exp (- (/ (* (float pi 1.0) K1) K)))))
    (do ((v2N1 v (+ v2N1 v v))			;(2N+1)v
	 (qN05 (sqrt q) (* qN05 q))		;q^(N + 0.5)
	 (sn 0.0) )
	((progn (setq sn (+ sn (/ (* qN05 (cos v2N1))
				  (+ 1.0 (* qN05 qN05)))))
		(< qN05 1.0E-6))
	 (/ (* (* 2.0 (float pi 1.0)) sn)
	    (* (sqrt M) K)))
      )))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                BESSEL-J
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun BESSEL-J (order x)
  "
;;;; BESSEL FUNCTIONS
;;; The Bessel functions J0(x) J1(x) . . .

;; Examples:

 (title \"Plot of Bessel Functions\")
 (plot
   (lambda(x) (bessel-j 0 x))
   (lambda(x) (bessel-j 1 x))
   (lambda(x) (bessel-j 2 x))
   (lambda(x) (bessel-j 3 x))
   0 12)
"
  (declare (float z2))
  (cond ((= order 0)				;NBS 9.4.1 eps < 5e-8
	 (cond ((<= (abs x) 3.0)
		(let ((z2 (expt (/ x 3.0) 2)))
		  (clmath-poly z2
			1.0000000  -2.2499997
			1.2656208  -0.3163866
			0.0444479  -0.0039444
			0.0002100)))
	       ((>= x 3.0)			;NBS 9.4.3
		(let* ((z      (/ 3.0 x))
		       (f0     (clmath-poly z		; eps 1.6E-8
				     0.79788456 -.00000077 -.00552740
				     -.00009512 0.00137237 -.00072805
				     0.00014476))
		       (theta0 (+ x
				  (clmath-poly z	; eps 7.0E-8
					-.78539816 -.04166397
					-.00003954 0.00262537
					-.00054125 -.00029333
					0.00013558))))
		  (declare (float z f0 theta0))
		  ;; Y0 is ... (/$ (*$ f0 (sin theta0)) (sqrt x))
		  (/ (* f0 (cos theta0)) (sqrt x))))
	       (t
		(BESSEL-J ORDER (- X)))))
	((= order 1)
	 (cond ((<= (abs x) 3.0)
		(let ((z2 (expt (/ x 3.0) 2)))
		  (* (* x
			(clmath-poly z2		;eps < x * 1.3e-8
			      0.50000000 -0.56249985
			      0.21093573 -0.03954289
			      0.00443319 -0.00031761
			      0.00001109)
			))))
	       ((>= x 3.0)			;NBS 9.4.3
		(let* ((z      (/ 3.0 x))
		       (f1     (clmath-poly z		; eps 4.0E-8
				     0.79788456 0.00000156 0.01659667
				     0.00017105 -.00249511 0.00113653
				     -.00020033))
		       (theta1 (+ x
				  (clmath-poly z	; eps 9.0E-8
					-2.35619449 +0.12499612
					+0.00005650 -0.00637879
					+0.00074348 +0.00079824
					-0.00029166))))
		  (declare (float z f1 theta1))
		  ;; Y1 is ... (/$ (*$ f1 (sin theta1)) (sqrt x))
		  (/ (* f1 (cos theta1)) (sqrt x))))
	       (t
		(- (bessel-j order (- x))))))
	((< order 0)
	 (setq order (- order))
	 (cond ((oddp order)
		(- (BESSEL-J order x)))
	       (t (BESSEL-J order x))))
	(t
	 (if (= x 0) 0
	   (- (* (/ (* 2 (1- order)) x)
		 (BESSEL-J (1- order) x))
	      (BESSEL-J (- order 2) x))
	   )
	 )
	)
  )

#|
(defun bessel-i(order x)
  (* (^ -1 order) (bessel-j (order (* #c(0 1) x)))))
|#
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                BESSEL-I
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun BESSEL-I (order x)			
  "
;;;; MODIFIED BESSEL FUNCTIONS
;;; modified Bessel functions I0(x) I1(x) ...
;;; (exp z cos theta) = I0(z) + 2 (sigma (k = 1 inf) Ik(z) cos (k theta))
;;; I[v+1] = I[v-1] - (2 v / z) I[v]
;;; NBS 9.8.1

;; Examples:

 (title \"Plot of Modified Bessel Functions\")
 (plot
   (lambda(x) (bessel-i 0 x))
   (lambda(x) (bessel-i 1 x))
   (lambda(x) (bessel-i 2 x))
   0 4)
"
  (setq order (abs order))
  (cond ((= order 0)
	 (cond ((< (abs x) 3.75)
		(let ((z2 (expt (/ x 3.75) 2)))	;-3.75 <= x <= 3.75
		  (clmath-poly z2			;eps < 1.6e-7
			1.0000000 3.5156229
			3.0899424 1.2067492
			0.2659732 0.0360768
			0.0045813)))
	       ((>= x 3.75)			; 3.75 <= x
		(let ((zi (/ 3.75 x)))
		  (* (/ (exp x) (sqrt x))
		     (clmath-poly zi
			   0.39894228 0.01328592
			   0.00225319 -.00157565
			   0.00916281 -.02057706
			   0.02635537 -.01647633
			   0.00392377)))	;eps < 1.9e-7 (exp x) / (sqrt x)
		)
	       (T (BESSEL-I ORDER (- X)))))
	((= order 1)
	 (cond ((< (abs x) 3.75)
		(let ((z2 (expt (/ x 3.75) 2)))	;-3.75 <= x <= 3.75
		  (* x				;eps < x * 8e-9
		     (clmath-poly z2
			   0.50000000 0.87890594
			   0.51498869 0.15084934
			   0.02658733 0.00301532
			   0.00032411))))
	       ((>= x 3.75)			; 3.75 <= x
		(let ((zi (/ 3.75 x)))
		  (* (/ (exp x) (sqrt x))
		     (clmath-poly zi
			   0.39894228 -.03988024
			   -.00362018 0.00163801
			   -.01031555 0.02282967
			   -.02895312 0.01787654
			   -.00420059)))	;eps < 2.2e-7 (exp x) / (sqrt x)
		)
	       (T (- (BESSEL-I ORDER (- X))))))
	(t 
	 (if (= x 0) 0
	   (- (BESSEL-I (- order 2) x)		; *** doubly recursive
	      (* (/ (float (* 2 (1- order))) x)
		 (BESSEL-I (1- order) x)))))))

#|

;;;; Incomplete Gamma

;;; *** DOES NOT WORK -- gamma* not total & does not produce correct answers
;;; *** GAMMA-INCOMPLETE bombs on large x

;;; NBS 6.5.4
;;;   \gamma^*(a,x) is a single valued analytic function with no finite
;;;   singularities 
;;;   \gamma^*(a,x) = { x^{-a} \over \Gamma(a) } \gamma(a,x)
;;; NBS 6.5.14
;;;   (gamma* -n x) = (expt x n)
;;; NBS 6.5.29
;;;   \gamma^*(a,x) = (*$ (/$ 1.0 (gamma a))
;;;                      (sigma (n 0 inf) (/$ (expt (- z) n) (+ a n) n!)))
;;;

;;; gamma*aux(a,x) = \Gamma(a) \gamma(a,x)
(defun gamma*aux (a z)				; *** GROSS ROUNDOFF ERRORS
  ;; a > 0 if a is an integer
  (do ((n    0.0 (1+ n))
       (term 1.0
	     (/ (* term (- z))
		(+ n 1.0)))
       (sum  0.0))
      ((progn (setq sum (+ sum (/ term (+ a n))))
	      (< (abs (/ term (+ a n))) 1.0E-7))
       sum)
    (declare (float n term sum))
    ))

(defun gamma* (a z)
  (/ (gamma*aux a z) (gamma a)))

;;; NBS 6.5.4 implies
;;;   \gamma(a,x) = {\Gamma(a) \over x^{-a}} \gamma^*(a,x)
;;;   \gamma(a,x) =  \Gamma(a)       x^{ a}  \gamma^*(a,x)
;;; NBS 6.5.22
;;;   \gamma(a+1,x) = a \gamma(a,x) - x^{a} e^{-x}
;;;     \gamma(a,x) = (\gamma(a+1,x)  + x^{a} e^{-x}) / a
(defun gamma-incomplete (a x)
  "
;; Examples:
 (title \"Plot of incomplete gamma Functions\")
 (plot
   (lambda(x) (gamma-incomplete .5 x))
   (lambda(x) (gamma-incomplete 1 x))
   (lambda(x) (gamma-incomplete 3 x))
;;   (lambda(x) (gamma-incomplete 10 x))
   0 15)
"
  (if (< a 1.0)
      (/ (+ (gamma-incomplete (+ a 1.0) x)
	    (* (expt x a) (exp (- x))))
	 a)
    (* (expt x a) (gamma*aux a x))))

|#


;;;; Mathematical and Physical Constants

;;;  (c) Copyright Gerald Roylance 1980, 1985
;;;      All Rights Reserved.
;;;  This file may be distributed noncommercially provided
;;;  that this notice is not removed.

;;; Bugs and Fixes
;;;   Better treatment of physical constants
;;;     in TX:/tx/glr/funct/sym/unit.lisp

#|
(defmacro defcon (name symbol number . dimensions)
  `(progn 'compile
	  (declare (special ,name)
		   (float ,name))
	  (setq ,name ,number)))
|#
;; ael: I defined defcon this way, instead:
(defmacro defcon (name symbol number &rest dimensions)
  "
;; Description: macro defcon (name symbol number &rest dimensions)

	Useful for defining physical constants, with a symbol name and
dimensions.

;;;; Physical Units

;;; base units of Si:
;;;      meter, kilogram, second, ampere, Kelvin, Candela

;;; volt
;;; ohm
;;; coulomb
;;; farad
;;; henry   (inductance that produces 1V emf when di/dt=1 A per sec)
;;; Weber   (mag. flux producing 1V emf in one turn as the flux
;;;          goes uniformly to zero in 1 second)
;;; Tesla   (flux density of 1 Weber/square meter)

;;;				MKS		EMU
;;; magnetizing force	H	A-turns/meter	=Oersteds * 1e3 / (4 pi)
;;; flux		phi	Weber		=Maxwells * 1e-8
;;; flux density	B	Weber/sq meter	=Gauss * 1e-4
;;; permeability	mu	H/m		=Gauss/Oersted *4e-7 * pi

;; Examples
 (defcon stuff	c	2.9979250e8 meters per second)
 stuff
 (get 'stuff 'symbol)
 (get 'stuff 'dimensions)
"
  (eval 
   `(let ()
      (setq ,name ,number)
      (putprop ',name ',symbol 'symbol)
      (putprop ',name ',dimensions 'dimensions)
      ,number
      )))
#|
	The constants themselves have been moved to the compprob.lsp file.
|#
