;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
#|
	The next buncha macros came out of my work on the traffic problem for
my modelling course at Ripon College.

	I automagically build the usual suspects (methods) into a prototype,
including a list of instances (under the slot 'active-list) and the total
number of objects defined (under the slot 'number).

	I also create a number of "accessor" functions, which allow me to send
messages using macros, rather than functions. Thus, to give a specific example,
rather than 

	(send andy :height)

one types

	(ask andy height)

I find this "logo-esque" style more attractive. The downside is that, when
one uses macros, one risks to break things when compiled. Thus, any compiled
functions should probably avoid these macros.


							Andy Long,
							October 30, 1997
|#
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                uber-proto
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmacro uber-proto (a b c &optional d doc)
  "
;; Description: (uber-proto name instance-slots class-slots
		 	&optional parent doc)

	A prototype a is defined, using defproto, and a number of standard
methods are installed:

	:number
	:active-list
	:isnew
	:slots
	:save
	:dispose

	Each proto comes with a new :isnew method, and a :dispose method to go
with it. This permits protos to keep track of their instances, in the slot
'active-list. The slot 'number keeps track of the sheer number of instances.

	I rigged up a kludgy :save method for intances of your proto. Other
methods you may want to add are :describe and :print.

	Methods for accessing each of the instance slots you gave are also
created, but these methods are accessed with quotes, rather than colons, as
shown in the examples below.

	You should also be able to take advantage of the functions and macros
ask, show, show-slots, change, and save-object.

	If you create a proto with uber-proto which has a parent proto created
by uber-proto, then by creating and disposing of instances of the sub-proto you
will also be registering their appearance/disappearance with the parent (this
is achieved via the use of the call-next-method command).

;; See also: new, dispose, ask, show, show-slots, change

;; Examples:
 (uber-proto longs (hio hoi) (thisn))

 (uber-proto tchapo (hio hoi thth ththht) (thisn) longs)
 (show tchapo)

 (uber-proto anna (hio hoi pthth ht) ()  longs)
 (show anna)
 (def annaroo (send anna :new 1 2))
 (show annaroo)
 (show anna)
 (save-object annaroo)

 (uber-proto andy (hio hoi) (thisn) longs \"Andy object\")
 (def andyroo (send andy :new 1 2))
 (show andyroo)
 (show andy)
 (save-object andyroo)

 (ask andy number)
 (ask andy thisn)
 (change andy thisn 'howdy)
 (ask andy thisn)

 ;; Methods are automatically constructed of this form: note the use of the
 ;; quote rather than the colon:
 (send andyroo 'thisn)
 (ask andyroo thisn)
"
  (let* (
	 (cc (append `,c '(number active-list)))
	 vals
	 )
    
    ;; Define the proto
    (eval `(defproto ,a ',b ',cc ,d ,doc))
    
    ;; create the instance-variable methods:
    (dolist
     (i `,b)
     (eval `(defmeth ,a ,i (&optional new)
	      (when new (setf (slot-value ',i) new))
	      (slot-value ',i)
	      )
	   )
     )
    ;; then the class-variable methods:
    (dolist
     (i `,cc)
     (eval `(defmeth ,a ,i (&optional new)
	      (when new (setf (slot-value ',i) new))
	      (slot-value ',i)
	      )
	   )
     )
    ;; the :number method:
    (eval `(defmeth ,a :number (&optional new)
	     (when new (setf (slot-value 'number) new))
	     (length (slot-value 'active-list))
	     )
	  )

    ;; the :active-list method:
    (eval `(defmeth ,a :active-list (&optional new)
	     (when new (setf (slot-value 'active-list) new))
	     (slot-value 'active-list)
	     )
	  )

    (eval `(defmeth ,a :slots ()
	     "Slots method, so that one can use the show-slots command."
	     (mapcar (lambda(x) (slot-value x)) (send self :slot-names))
	     ))
    
    ;; Needs an isnew:
    (eval `(defmeth ,a :isnew (&rest e)
	     ;; The following use of call-next-method should have the effect of
	     ;; updating the active-list and number only!
	     (call-next-method)
	     (mapcar (lambda(x y) (setf (slot-value x) y)) ',b e)
	     (send ,a :slot-value 'number (1+ (ask ,a number)))
	     (send ,a :slot-value 'active-list 
		   (append (ask ,a active-list) (list self)))
	     self
	     ))

    (setq
     vals (eval `(send ,a :slot-value 'proto-name))
     vals (format nil "	This is the :dispose method for ~a objects.
The argument to :dispose should be an object of type ~a or a list of such
objects." vals vals))
    (eval `(defmeth ,a :dispose (&rest e)
	     ,vals
	     (apply #'call-next-method e)
	     (if (listp (first e))
		 (mapcar #'(lambda(x) (send ,a :dispose x)) (first e))
	       (let ((bong (reverse (set-difference (ask ,a active-list) e))))
		 (send ,a :slot-value 'number (length bong))
		 (send ,a :slot-value 'active-list bong)
		 )
	       )
	     )
	  )

    ;; Needs a save method: this will save instances of prototypes, but it's
    ;; really kludgy! It relies on the :save-self method I created for
    ;; *objects*, defined in obmeths.lsp in the math subdirectory.
    (eval `(defmeth ,a :save (&optional name)
	     (send self :save-self 
		   (if name name (send self :slot-value 'proto-name)))))
    
    (eval `(change ,a number 0))
    (eval `(change ,a active-list nil))
    nil
    )
  )
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                new
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmacro new (thingy &rest something)
  "
Creates a new thingy using the usual instance slots.

;; Examples:
 (uber-proto longs (body) (aches))
 (uber-proto andy (arms legs) (pains) longs)
 (def andyroo (new andy))
 (show andyroo)
 (def andyroo (new andy 1 2))
 (show andyroo)
"
  (if something 
      `(send ,thingy :new ,@something)
    `(send ,thingy :new)
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                show
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmacro show (thingy) 
  "
;; Description:

	(show object)

This is just a shorthand for (send object :show).

;; Examples:
 (uber-proto longs (body) (aches))
 (uber-proto andy (arms legs) (pains) longs)
 (def andyroo (new andy))
 (show andyroo)
 (show andy)
 (show (list andy andyroo))
"
  (if (listp thingy)
      `(mapcar #'(lambda(x) (send x :show)) ,thingy)
    `(send ,thingy :show)
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                ask
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmacro ask (thingy something) 
  "
;; Examples:
 (uber-proto longs (body) (aches))
 (uber-proto andy (arms legs) (pains) longs)
 (def andyroo (new andy))
 (show andyroo)
 (show andy)
 (ask andy number) ;; how many andys?
"
  `(send ,thingy :slot-value ',something)
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                change
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmacro change (f1 f2 f3)
  "
;; Description: macro
 (change object slot new-value)

;; See also: change2

;; Examples:
 (uber-proto longs (body) (aches))
 (uber-proto andy (arms legs) (pains) longs)
 (def andyroo (new andy))
 (ask andyroo arms)
 (change andyroo arms 1)
 (ask andyroo arms)
"
  (eval `(change2 ,f1 ',f2 ,f3))
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                change2
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun change2 (thingy something to-this)
  "
 (change2 object slot new-value)
"
  `(send ,thingy :slot-value ',something ',to-this)
  )
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                show-slots
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun show-slots (obj)
  "
;; Description: Function
	show-slots shows the slots and all their values for an object.

;; Examples:
 (uber-proto longs (body) (aches))
 (uber-proto andy (arms legs) (pains) longs)
 (def andyroo (new andy))
 (ask andyroo arms)
 (change andyroo arms 1)
 (show-slots andyroo)
"
  (let* (
	 (own-slots (reverse (remove-duplicates (send obj :slot-names))))
	 (slot-vals
	  (mapcar (lambda(x) (send obj :slot-value x)) own-slots))
	 )
    (mapcar (lambda(x y) (format t "~t~a: ~s~%" x y))
	    own-slots slot-vals)
    )
  nil
  )
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 save-object
;;
(defun save-object(object)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          save-object Doc:
;;
;; Author: Andy Long
;; Created: Sat Nov 1 09:47:13 1997
;;
;; Description/Arguments:

       (save-object object)
  
where object is an object or a list of objects.

	Produces a command sufficient to reconstruct the given instance of an
object, provided a :save method is defined for the function. This is really
just a short-hand for (send object :save).

;; Notes:

;; See Also: the *object* method :save-self

;; Examples:
 (uber-proto longs (hio hoi) (thisn))
 (uber-proto andy (hio hoi) (thisn) longs \"Andy object\")
 (def andyroo (send andy :new 1 2))
 ;; uber-proto automagically creates a :save method for an instance of an
 ;; object:
 (save-object andyroo)
"
  (if (listp object) (mapcar #'(lambda(x) (send x :save)) object)
    (send object :save))
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 dispose
;;
(defun dispose(object instance)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          dispose Doc:
;;
;; Author: Andy Long
;; Created: Sat Nov 1 09:47:13 1997
;;
;; Description/Arguments:

       (dispose object instance)
  
	object - an object or list of objects.
	instance - an instance or a list of instances

	If both object and instance are lists, they should have the same
length.

	Dispose removes an instance from the active-list slot of the object,
provided their is such a slot.

;; Notes:

;; See Also: the *object* method :dispose

;; Examples:
 (uber-proto longs (hio hoi) (thisn))
 (uber-proto andy (hio hoi) (thisn) longs)
 (def andyroo (send andy :new 1 2))
 ;; uber-proto automagically creates a :dispose method for an instance of an
 ;; object:
 (dispose andy andyroo)
 (ask andy number)
 (def andyroo (new andy))
 (def andyroo2 (new andy))
 (ask andy number)
 (dispose andy (list andyroo andyroo2))
 (ask andy number)
"
  (cond
   ((and (listp object) (listp instance))
    (mapcar #'(lambda(x y) (send x :dispose y)) object instance)
    )
   ((listp instance)
    (mapcar #'(lambda(y) (send object :dispose y)) instance)
    )
   ((listp object)
    (mapcar #'(lambda(y) (send x :dispose instance)) object)
    )
   (t
    (send object :dispose instance)
    )
   )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 tack-colon-onto-symbol
;;
(defun tack-colon-onto-symbol(x)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          tack-colon-onto-symbol Doc:
;;
;; Author: Andy Long
;; Created: Thu Oct 30 23:45:55 1997
;;
;; Description/Arguments:

       (tack-colon-onto-symbol x)
  
;; Notes:

;; See Also:

;; Examples:
 (tack-colon-onto-symbol 'hi)
 (tack-colon-onto-symbol '(ho hi))
"
  (if (listp x) (mapcar #'tack-colon-onto-symbol x)
    (symbol (strcat ":" (symbol-name x))))
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 new-accessor-method
;;
(defmacro new-accessor-method (proto name &optional name2)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          new-accessor-method Doc:
;;
;; Author: Andy Long
;; Created: Thu Nov 6 20:42:49 1997
;;
;; Description/Arguments: macro (proto name &optional name2)

	Writes the usual accessor method, which either sets a slot by the given
name if given an argument, or returns its value.

;; Notes:

;; See Also:

;; Examples:
 (uber-proto neat (layers) nil)
 (new-accessor-method neat :layers 'layers)
 (send neat :layers)
 (send neat :layers 4)
 (send neat :layers)
"
  (let ((name2 (if name2 name2 name)))
    (eval
     `(defmeth ,proto ,name (&optional new)
	(when new (setf (slot-value ,name2) new))
	(slot-value ,name2)))))
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
