;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                try-file
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun try-file(filename &key redoc)
  "
;; Description: (try-file filename)

	Tries a file (that is, it runs try-out on all functions and macros in a
file containing help information for files that I've written. This verifies
that all help is actually helpful. If you use the key :redoc, then redoc will
also be run, correcting mistakes that may be lurking in any old help files,
created before bugs were discovered.

;; Examples:
 (try-file (strcat *ANDY-DEFAULT-PATH* \"math/math.lsp\"))
"
  (let (function-name (source (read-file filename)))
    (dolist
     (i source)
     (let ((n (length i)))
       (setq
	function-name
	(cond
	 ((> n 8)
	  (if (equalp (subseq i 1 9) "defmacro")
	      (read-from-string (select i (iseq 10 (1- (length i)))))
	    (if (equalp (subseq i 1 6) "defun")
		 (read-from-string (select i (iseq 7 (1- (length i)))))))
	  )
	 ((> n 5)
	  (if (equalp (subseq i 1 6) "defun")
	      (read-from-string (select i (iseq 7 (1- (length i)))))))
	 (t nil)
	 )
	)
       (if function-name
	   (let ()
	     (helplass function-name)
	     (if redoc (redoc (symbol-name function-name)))
	     ))))))
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 rough-diff-dates
;;
(defun rough-diff-dates(x y)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          rough-diff-dates Doc:
;;
;; Author: Andy Long
;; Created: Fri Jun 27 16:21:23 CDT 1997
;;
;; Description/Arguments:
 
       (rough-diff-dates date1 date2)
 
	gives the rough number of days between the two dates should be correct
to within a few days. I'm just multiplying the difference in years by 365.25,
and using the usual month lengths (feb=28), so I'm going to miss a day or two
now and then.

;; See Also:
 
;; Examples:
 (rough-diff-dates '(6 27 1997) '(2 4 1961))
 (rough-diff-dates '(6 27 1997) '(6 7 1997))
 (rough-diff-dates '(12 31 1997) '(1 1 1997))
 (rough-diff-dates '(1 1 1997) '(1 1 1996))
"
  (let* (
         (months '(0 31 28 31 30 31 30 31 31 30 31 30 31))
;;       (names '(jan feb mar apr may jun jul aug sep oct nov dec))
;;       (mprint (bind-columns months names))
         (diff (- x y))
         (day-diff (second diff))
         (year-diff (third diff))
         (months-to-days
          (-
           (sum (select months (iseq (first x))))
           (sum (select months (iseq (first y))))
           ))
         (days (+ months-to-days day-diff))
         )
    (if (< year-diff 0) (rough-diff-dates y x)
      (round (+ (* year-diff 365.25) days))
      )
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 load-data
;;
(defmacro load-data(filename &optional dir)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          load-data Doc:
;;
;; Author: Andy Long
;; Created: Tue Jul 1 16:12:48 CDT 1997
;; Modified: Tue Oct 14 03:27:25 CDT 1997
;;
;; Description/Arguments:

        (load-data file-name-string &optional dir-name-string)
  
	Formerly:

LOAD-DATA                                                         [function-doc]
Args: (file)
Read in data file from the data examples library.

	The file-name-string can be a symbol, if preferred. A lower-case string
will be formed from the symbol-name.
	The optional dir-name-string should be the name of a fully-qualified
directory (that is, its full pathname).

;; Note: I changed this function because the Data files did not come in with
the standard DOS distribution. I added them to my own examples directory, but
thought that it would be a good idea to establish some paths. The default path
is *data-path*, where

 (setq *data-path*
   (list
     (strcat *default-path* separator \"Data\" separator)
     (strcat *andy-default-path* separator \"examples\" separator)
     ))

;; See also: load-example (aliased to this function)

;; Examples:
 (strcat *default-path* separator \"Data\" separator)
 (strcat *andy-default-path* separator \"examples\" separator)
 (load-data \"absorbtion\")
"
  (let* (
	 (file
	  (eval (if (stringp `,filename) `,filename 
		  (nstring-downcase (format nil "~a" `,filename)))))
	 (dir
	  (eval (if (stringp `,dir) `,dir 
		  (nstring-downcase (format nil "~a" `,dir)))))
	 (dirs (if dir (combine dir *data-path*) *data-path*))
	 (suffix (strcat file ".lsp"))
	 )
    (setq file
	  (dolist
	   (d dirs)
	   (if (probe-file (strcat d suffix))
	       (return (strcat d suffix)))))
    (if file (load file))
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          replace-char-by-char
;;
(defun replace-char-by-char(string char1 char2)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          replace-char-by-char Doc:
;;
;; Author: Andy Long
;; Created: Thu Jul 3 12:48:32 CDT 1997
;;
;; Description/Arguments:

       (replace-char-by-char string char1 char2)
  
;; Notes: grep-replace actually does this fine; this might be a little faster.

;; See Also: grep-replace

;; Examples:
 (replace-char-by-char \"This stuff\" #\\s #\\v)
 (replace-char-by-char \"This stuff\" \"s\" \"v\")

 (grep-replace '(#\\s) '(#\\v) \"This stuff\")
 (grep-replace \"s\" \"v\" \"This stuff\")
"
  (let (
	(char1 (if (stringp char1) (elt char1 0) char1))
	(char2 (if (stringp char2) (elt char2 0) char2))
	)
    (dotimes
     (i (length string))
     (if (equal (elt string i) char1) (setf (elt string i) char2))
     )
    string
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 intersections
;;
(defun intersections(&rest sets)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          intersections Doc:
;;
;; Author: Andy Long
;; Created: Wed Jul 9 12:19:10 CDT 1997
;;
;; Description/Arguments:
 
       (intersections &rest sets)
 
;; Notes: extends intersection to allow for more than three sets.
 
;; See Also:
 
;; Examples:
 (intersections  '(1 2) '(2 3))
 (intersections  '(1 2) '(2 3) '(4 5))
 (intersections  '(1 2) '(2 3) '(2 6))
"
  (let ((result (first sets)))
    (dolist
     (set (rest sets))
     (setq result (intersection result set))
     )
    result
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 unions
;;
(defun unions(&rest sets)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          unions Doc:
;;
;; Author: Andy Long
;; Created: Wed Jul 9 12:19:10 CDT 1997
;;
;; Description/Arguments:
 
       (unions &rest sets)
 
;; Notes: extends union to allow for more than three sets.
 
;; See Also:
 
;; Examples:
 (unions  '(1 2) '(2 3))
 (unions  '(1 2) '(2 3) '(4 5))
 (unions  '(1 2) '(2 3) '(2 6))
"
  (let ((result (first sets)))
    (dolist
     (set (rest sets))
     (setq result (union result set))
     )
    result
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 francais
;;
(defun francais()
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          francais Doc:
;;
;; Author: Andy Long
;; Created: 10/4/1997
;;
;; Description/Arguments:

  Loads in some translations (from the file french.lsp in the etc directory)
  for those who prefer programming in French.

       (francais)
  
;; Notes:

;; See Also:

;; Examples:
 (francais)
;; French, now loaded, permits one to write the function choses as:
 (fonction
  choses (n m)
  \"
 ;; Examples:
  (choses 20 10)
  \"
  (soit
   ((v 0) r)
   (emballer
    (imprimer v)
    (faites-liste
     (i (iseq n))
     (imprimer i)
     (si (> i m) (aller fini))
     )
    fini
    )
   )
  )
 (choses 20 10)
"
  (load (strcat *andy-default-path* "etc" separator "french.lsp"))
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 primep
;;
(defun primep(n)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          primep Doc:
;;
;; Author: Andy Long
;; Created: 10/4/1997
;;
;; Description/Arguments:

       (primep n)
  
;; Notes:

;; See Also:

;; Examples:
 (primep 5)
 (primep 4)

 (mapcar #'primep (iseq 5))
 ;; The following will cause an error:
 ; (primep 'a)
"
  (if (listp n) (mapcar #'primep n)
    (if (integerp n)
	(if (> n 1)
	    (if (= 0
		   (sum
		    (mapcar
		     (lambda(i) (1- (gcd n i)))
		     (iseq 1 (ceiling (/ n 2))))))
		t nil)
	  nil
	  )
      (error 
       (format nil "Only integers may be tested, whereas you tried testing ~s."
	       n))))
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; v-or Doc:    Creation Date:5/22/95        Author: Ory Long
;;
;;
(defun v-or (n)
  "
;; Description/Arguments: (v-or list-o-statements)

 vectorized or
  
;; Notes:

;; See Also:

;; Examples:
 (v-or '(t t nil)) ;; true
 (v-or '(t (= .5 (/ 1 2)) (< 2 3))) ;; true
"
   (eval `(or ,@n))
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 system-to-string
;;
(defun system-to-string(string)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          system-to-string Doc:
;;
;; Author: Andy Long
;; Created: 10/6/1997
;;
;; Description/Arguments:

       (system-to-string string)
  
  Returns the output of your system command in a string. This in response to a
submission to the xlispstat users' list:
From majordom@stat.umn.edu Mon Sep 15 11:11 CDT 1997 

Rick Riolo wrote:
 > 
 > 
 > Is there a way to get the unix process number of
 > the running xlispstat into an list variable in that process?
 > (For example, perl has a special variable with that value.)
 > 
 > Or more generally is there a way to capture the stdout
 > stream from a job started with 
 >   (system <command >)
 > For example, capture each line as the value of a 
 > entry in a list of strings, one string for each output line.
 > That would open up a lot of possibilites.
 > thanks.
 > 
 >  - r
 > 
 > Rick Riolo                       rlriolo@umich.edu
 > Program for Study of Complex Systems (PSCS)
 > 4068 Randall Lab     University of Michigan
 > Ann Arbor MI 48109-1120
 > http://pscs.physics.lsa.umich.edu/PEOPLE/rlr-home.html
 > 
 > 

An undocumented feature of the current system command is that it takes
a stream to use as output for its second argument.  So or example,

 > (with-output-to-string (s) (system \"date\" s))
 \"Mon Sep 15 10:15:54 CDT 1997
 \"

The feature is undocumented because it may change once I figure out a
better way to do it, so rely on this with caution.

luke

-- 
Luke Tierney
University of Minnesota                      Phone:           612-625-7843
School of Statistics                         Fax:             612-624-8868
206 Church Street                            email:      luke@stat.umn.edu
Minneapolis, MN 55455 USA                    WWW:  http://www.stat.umn.edu

;; Notes:

;; See Also:
      
;; Examples:
 (system-to-string \"date\")
"
  (with-output-to-string (s) (system string s))
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun unix-file-to-dos-file(infile outfile)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          unix-file-to-dos-file Doc:
;;
;; Author: Andy Long
;; Created: 10/29/1997
;;
;; Description/Arguments:

       (unix-file-to-dos-file infile outfile)
  
	This function slaps a control-M to the end of every line of a file,
  which should make the resulting output file more useful under dos, maybe the
  Mac too. Anytime you gotta have more control-M characters!

;; Notes: infile and outfile can be the same.
	  This function really just calls put-control-m.

;; See Also: put-control-m

;; Examples:
"
  (put-control-m infile :overwrite nil :outfile outfile)
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                         put-control-m
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun put-control-m(file &key (n 13) (overwrite t) outfile)
  "
;; Description: (put-control-m file &key (n 13) (overwrite t) outfile)

	Puts control-M characters (ASCII character 13) at the end of each line
of file.

	Overwrites the existing file by default.

	If the file is very large, this may be painfully slow!
"
  (let (
	(r (read-file file))
	(s (string (code-char n))) ;; 
 if n is 13
	)
    (setq file
	  (cond
	   (overwrite file)
	   (outfile outfile)
	   (t
	    (format t "~%What is the output filename? ")
	    (read-line) ;; the first just flushes buffer...
	    (read-line)
	    )
	   ))
    (if (/= 0 (length file))
	(with-open-file
	 (f file :direction :output)
	 (dolist
	  (line r)
	  (format f "~a~a~%" line s)
	  )
	 )
      (format t "~%File unchanged: no output file given.~%")
      )
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 member-if-not
;;
(defun member-if-not(fn x)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          member-if-not Doc:
;;
;; Author: Andy Long
;; Created: Thu Dec 4 12:13:56 1997
;;
;; Description/Arguments:

	(member-if-not fn x)
  
;; Returns: (member-if #'(lambda(x) (not (funcall fn x))) x))

;; Notes: f2cl needed this common-lisp function, and I think that this was
;; enough of a kludgy fix for it; I make no claims about the true
;; common-lisp-ness of this function, however! 

;; See Also: member-if

;; Examples:
 (member-if-not 'numberp (iseq 2))
 (member-if-not 'null (iseq 2))
"
  (member-if #'(lambda(x) (not (funcall fn x))) x)
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                   These from f2cl - list-split is too useful to bury!
;;
;;       append1 (a helper)
;;       tail-chop
;;       head-chop
;;       list-split
;;
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
(defun append1 (l x)
  "
 Helper for list-split
"
  (append l (list x)))
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                tail-chop
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun tail-chop (beta lis) 
  "
 function tail-chop (beta lis) 

 Helper for list-split; cuts off everything after the first occurance of the
 first argument beta in lis. If beta is not found in lis, lis is returned.

;; Examples:
 (tail-chop 14 (iseq 10))
 (tail-chop 4 (iseq 10))
 (tail-chop #\\\a (coerce \"this is a test with a string\" 'list))
"
  (let (retlist)
    (do
     ()
     ((or (equal lis nil) (equal (car lis) beta)) retlist)
     (setq retlist (append1 retlist (car lis))
	   lis (cdr lis))
     )))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                head-chop
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun head-chop(beta lis)
  "
 function head-chop (beta lis) 

 Helper for list-split; cuts off everything before the first occurance of the
 first argument beta in lis. If beta is not found in lis, nil is returned.

;; Examples:
 (head-chop 14 (iseq 10))
 (head-chop 4 (iseq 10))
 (head-chop #\\\a (coerce \"this is a test with a string\" 'list))
"
  (cdr (member beta lis)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                list-split
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun list-split (beta lis)
  "
;; list-split:

;; From the f2cl folks. This one is useful for, of all things, splitting a list
;; according to the presence of a certain indicator and throwing the indicator
;; away.

;; Examples:
 (list-split 4 (iseq 10))
 (list-split #\\\a (coerce \"this is a test with a string\" 'list))
"
  (cond
   ((equal (member beta lis)  nil)
    (list lis))
   (t 
    `(,(tail-chop beta lis) ,@(list-split beta (head-chop beta lis))))
   ))
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;				repeater
;;
(defun repeater(object n-times)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;			repeater Doc:
;;
;; Author: Andy Long
;; Created: Sat Feb 7 09:09:29 1998
;;
;; Description/Arguments:

	(repeater object n-times)
  
;; Returns: a list of n-times copies of object.

;; Notes: my idea of repeat is different from the creater of repeat's idea: I
;; want a list of n objects.

;; See Also: repeat

;; Examples:
 (repeater #2a((1 2) (3 4)) 2)
 (repeater (list 1 2) 2)
"
  (let ((list (repeat 0 n-times)))
    (dotimes
     (i n-times)
     (setf (elt list i) object)
     )
    list
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;				make-all-tuples
;;
(defun make-all-tuples(&rest lists)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;			make-all-tuples Doc:
;;
;; Author: Andy Long
;; Created: Tue Mar 3 00:36:28 1998
;;
;; Description/Arguments: make-all-tuples
  
;; Arguments: &rest lists

	Lists is n lists of indices, from which all possible n-tuples will be
formed. This is useful for indexing arrays, especially arrays more complicated
than matrices.

;; Returns: a list of tuples

;; Notes: Created as a helper for t-mult (tensor-multiplication).

;; See Also: t-mult

;; Examples:
 (make-all-tuples (iseq 2))
 (make-all-tuples (iseq 2) (iseq 3))
 (make-all-tuples (iseq 2) (iseq 3) (iseq 2))

 ;; They don't have to be standard indices:
 (make-all-tuples '(a b) '(1 2))
"
  (flet
   ((combiner(x y) (mapcar (lambda(z) (combine x z)) y)))
   (let ((new-tuples (mapcar #'list (first lists))))
     (do*
      (
       (whats-left (rest lists) (rest whats-left))
       (next-up (first whats-left) (first whats-left))
       (tuples new-tuples new-tuples)
       )
      ((not next-up) new-tuples)
      (setq
       new-tuples
       (smoosh
	(mapcar (lambda(uple) (funcall #'combiner uple next-up)) tuples)))
      )
     )
   )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
