;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
#|
Dad,

	Thanks for the paper: I'm not sure I understand it, but it did inspire
this little project (so that's a good thing!).

	This file documents that our flesh technique can be used to match a
grid of functions (a function grid). Let me describe the following example:

1) Define the function grid (that is, the constraint functions in the two
   dimensions). I'll do that by defining a function f of two variables, then
   creating 3 functions in each directions as slices of f. Note that I'm using
   a square function grid. It is not essential to do so.
	If we don't have a square function grid, then we could create
   additional grid points (by evaluating the functions!). Thus, if a 3x5
   function grid is given, we could simply extend it to 5x5 by evaluating
   the functions at some intermediate (or even exterior) points, then spline
   those any way we'd like:

		-------------		---+-----+---
		|     |     |		|     |     |
		-------------		---+-----+---
		|     |     |		|     |     |
	5x3:	-------------	->	---+-----+---	->
		|     |     |		|     |     |
		-------------		---+-----+---
		|     |     |		|     |     |
		-------------		---+-----+---

		---+-----+---
		|  |  |  |  |
		---+-----+---
		|  |  |  |  |
		---+-----+---
		|  |  |  |  |   : 5x5!
		---+-----+---
		|  |  |  |  |
		---+-----+---

2) Find the svd of the 3x3 matrix A, and define A = U L V' (xlispstat
puts out V, not the transpose).

3) I'm going to work with vectors of functions, so we solve for v1(y), etc., in
terms of the known functions h1(y), etc.:

	Ty(y)' = [h1(y) h2(y) h3(y)]' = U L [v1(y) v2(y) v3(y)]',
so
	[v1(y) v2(y) v3(y)]' = L^(-1) U' Ty(y)'.
And
	Tx(x) = [g1(x) g2(x) g3(x)] = [u1(x) u2(x) u3(x)] L V',
so
	[u1(x) u2(x) u3(x)] = Tx(x) V L^(-1).

4) I define vector functions Ux(x) and Vy(y) to contain our singular functions.
   Then Axy = A(x,y) = Ux(x) L Vy(y)' is our interpolator! Works like a charm.

	Notice especially that our singular functions are of the same type as
the constraint functions, so if they were originally a type of cubic spline,
then so would our singular functions be cubic splines.

	I hope that you'll be able to replicate this process in one of your
favorite packages. The only trick is keeping the transposes in the right order,
etc. It's beautifully simple in xlispstat: can you do it as easily in your code
(giving you a challenge may inspire you to try it!).

	Have fun!
						Love,
						       Andy

PS: I'll put my images into gifs, and put them onto my web page (under dad, at
the bottom). Here's the url: http://maya.medctr.luc.edu/~ael/dad/

	There are files called f.gif, a.gif, both.gif, difference.gif, and some
files u1.gif, v1.gif, etc. showing that the singular functions interpolate the
singular vectors.
|#
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
#|
	Start with some complicated function (which generates a FULL matrix a,
and it is essential that the matrix be full in the following code):
|#
(func f(x y) sin(6*ln(1+x)*y) + y*x + (exp y*(1-x)))

#|
	Here's a plot, 21x21, on [0,1]x[0,1]:
|#
(spin-domain 0 1 0 1)
(spin-points 21)
(title "The original function f")
(plot f)

#|
	Now we define the grid, and compute the svd:
|#
(def n 4)
(def A (outer-product (rseq 0 1 n) (rseq 0 1 n) #'f))
(def svd (svd a))

(def U (first svd))
(def lambdas (second svd))
(def V (third svd))

;; (def lambda (diagonal-to-full lambdas)) ;; don't need this!
(def lambda-inv (diagonal-to-full (/ lambdas)))

#|
	Define vectors of constraint functions (these are just slices of f):
|#
(defun Tx(x) (mapcar (lambda(y) (f x y)) (rseq 0 1 n)))
(defun Ty(y) (mapcar (lambda(x) (f x y)) (rseq 0 1 n)))

#|
	The singular functions:
|#
(def umat (matmult (third svd) lambda-inv))
(def vmat (matmult lambda-inv (transpose (first svd))))
(defun Ux(x) (matmult (Tx x) umat))
(defun Vy(y) (matmult vmat (Ty y)))

#|
	The interpolator: same as (matprod (ux x) lambda (vy y))
|#
(defun Axy(x y) (dot (ux x) (* lambdas (vy y))))

(title "The interpolating function a(x,y)")
(plot axy)
(title "Compare the two functions")
(plot axy f)
(title "The difference between the functions")
(plot (lambda(x y) (- (axy x y) (f x y))))

;; Here's some stuff that you don't care about, Dad, so I won't bother you with
;; it (although I'm leaving it in so that I can reproduce things):
;; Produce a tiling of all four plots, and link them (link-all-plots, lap),
;; then make the gifs:
;; (load "dad2")
