;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;				jacobi-iteration
;;
(defun jacobi-iteration
  (
   A b
     &key 
     guess
     (maxits 100)
     (eps 1e-6)
     verbose
     p)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;			jacobi-iteration Doc:
;;
;; Author: Andy Long
;; Created: Mon Feb 9 20:53:55 1998
;;
;; Description/Arguments: jacobi-iteration
  
;; Arguments: A b &key guess (maxits 100) (eps 1e-6) p verbose

	A	- a matrix
	b	- a vector (or a matrix)
	guess	- a guess for the solution (Diagonal-inv * b by default)
	maxits	- maximum number of iterations
	eps	- a stopping criterion (relative change in  norm < eps)
	p	- a norm - the usual norms are 1, 2, or 'infinity
	verbose	- let's be talkative

;; Returns: The solution vector (or matrix) to the system Ax=b.

;; Notes: Burden and Faires, 'Numerical Analysis', 5th Edition, p. 409

	Jacobi can also be used to calculate the inverse of a matrix (see the
   last example below).

;; See Also: Gauss-Seidel-iteration

;; Examples:
 (def A #2a((5 2)(2 4)))
 (eigenvalues (matmult #2a((1/5 0)(0 1/4)) (- A #2a((5 0)(0 4)))))
 (def b #(1 2))
 (def x (jacobi-iteration A b :maxits 10 :verbose t))
 (matmult A x)

 ;; If we permute this matrix, Jacobi iteration will not work: the matrix is no
 ;; longer diagonally dominant (the eigenvalues are >1 in absolute value):
 (def A #2a((2 4)(5 2)))
 (eigenvalues (matmult #2a((1/2 0)(0 1/2)) (- A #2a((2 0)(0 2)))))
 (def b #(2 1))
 (def x (jacobi-iteration A b :maxits 10 :verbose t))
 (matmult A x)

 ;; page 407, B&F:
 (def A #2a((10 -1 2 0)(-1 11 -1 3)(2 -1 10 -1)(0 3 -1 8)))
 (def b #(6 25 -11 15))
 (def x (jacobi-iteration A b :maxits 10 :verbose t))
 (matmult A x)

 ;; Jacobi can be used to find the identity (but it's a lot slower than
 ;; Gaussian elimination!):
 (def b (id 4))
 (mprint (setq x (jacobi-iteration a b :maxits 100 :p 'infinity :verbose t)))
 (mprint (inverse a))
"
  (let* (
	 (diag (diagonal A))
	 (R (- A (diagonal-to-full diag)))
	 (D-inv (diagonal-to-full (/ diag)))
	 (D-inv-R (matmult D-inv R))
	 (D-inv-b (matmult D-inv b))
	 (guess (if guess guess D-inv-b))
	 )
    (if verbose (print guess))
    (do
     (
      (i 0 (incf i))
      (old-guess guess guess)
      (guess (- D-inv-b (matmult D-inv-R guess))
	     (- D-inv-b (matmult D-inv-R guess)))
      )
     ((or (>= i maxits) (< (/ (norm (- guess old-guess) :p p) (norm guess :p p))
			   eps))
      guess
      )
     (if verbose (let () (print i) (print guess)))
     )
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;				gauss-seidel-iteration
;;
(defun gauss-seidel-iteration(A b &key guess (maxits 100) (eps 1e-6) verbose)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;			gauss-seidel-iteration Doc:
;;
;; Author: Andy Long
;; Created: Mon Feb 9 20:53:55 1998
;;
;; Description/Arguments: gauss-seidel-iteration
  
;; Arguments: A b &key guess (maxits 100) (eps 1e-6) verbose

;; Returns: The solution vector to the system Ax=b.

;; Notes: Burden and Faires, 'Numerical Analysis', 5th Edition, p. 411

;; See Also: Jacobi-iteration

;; Examples:
 (def A #2a((5 2) (2 4)))
 (def b #(1 2))
 (def x (gauss-seidel-iteration A b :maxits 10 :verbose t))
 (matmult A x)

 ;; If we permute this matrix, iteration will not work: the matrix is no
 ;; longer diagonally dominant (the eigenvalues are >1 in absolute value):
 (def A #2a((2 4)(5 2)))
 (eigenvalues (matmult #2a((1/2 0)(0 1/2)) (- A #2a((2 0)(0 2)))))
 (def b #(2 1))
 (def x (gauss-seidel-iteration A b :maxits 10 :verbose t))
 (matmult A x)

 ;; page 410, B&F:
 (def A #2a((10 -1 2 0)(-1 11 -1 3)(2 -1 10 -1)(0 3 -1 8)))
 (def b #(6 25 -11 15))
 (def x (gauss-seidel-iteration A b :maxits 5 :verbose t))
 (matmult A x)
"
  (let* (
	 (diag (diagonal A))
	 (R (- A (diagonal-to-full diag)))
	 (D-inv (/ diag))
	 (D-inv-R (matmult (diagonal-to-full D-inv) R))
	 (D-inv-b (* D-inv (coerce b 'list)))
	 (len-b (length b))
	 (guess (if guess guess (repeat 0 len-b)))
	 )
    (if verbose (print guess))
    (do*
     (
      (i 0 (incf i))
      (old-guess (copy-list guess) (copy-list guess))
      (stuff
       (dotimes
	(j len-b)
	(setf (elt guess j) (- (elt D-inv-b j) (dot (row (1+ j) D-inv-R) guess)))
	)
       (dotimes
	(j len-b)
	(setf (elt guess j) (- (elt D-inv-b j) (dot (row (1+ j) D-inv-R) guess)))
	)
       )
      )
     ((or (>= i maxits) (< (/ (norm (- guess old-guess)) (norm guess)) eps))
      guess)
     (if verbose (let () (print i) (print guess)))
     )
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
