#|
				Neural Nets

	This is file neural.lsp, created to test out back-propogation and
neural nets. This work was undertaken at the Institut de Mathematiques et de
Sciences Physiques, Porto Novo, Benin, in collaboration with Dr. Joel Hounsou.

							Andy Long
							November, 1997

;; where am i:
Institute Africain d'Informatiques
Libreville, Gabon: 12-24 Janvier
|#
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                net proto:
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(uber-proto net
	    (layers learning-rate net-type options momentum)
	    (number active-list) nil "Prototype for a neural net.")
;; added net-type, where net-type is currently one of 'back-propogation or
;; 'perceptron. 

;; establishes default learning rate of 1.
(send net 'learning-rate 1)
(send net 'net-type 'back-propogation)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                layer proto:
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(uber-proto layer
	    (nodes weights delta net-values sigmoid sigmoid-prime type
		   momentum previous-weights)
	    (number active-list) nil "Prototype of a layer for a neural net.")

;; initialize sigmoidal functions:
(defun sigmoid-1(x) "the logistic function"
  (let ((val (exp x))) (/ val (1+ val))))
(defun sigmoid-1-prime(x)
  (let ((val (exp x))) (/ val (square (1+ val)))))

(defun sigmoid-2(x) "The shifted, scaled arctan function."
  (* .5 (+ 1 (atan x))) )
(defun sigmoid-2-prime(x) "The shifted, scaled arctan function."
  (* 0.5 (/ 1 (+ 1 (^ X 2)))))

;; We'll send the standard logistic model:
(send layer 'sigmoid #'sigmoid-1)
(send layer 'sigmoid-prime #'sigmoid-1-prime)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                Net methods:
;;
;;					:isnew
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth net :isnew (layers)
  "Isnew method creates instances of nets. One should give a list of the number
of nodes in each layer to the method."
  (send self 'layers
	(mapcar #'(lambda(x) (send layer :new (repeat 0 x))) layers)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :initialize
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth net :initialize ()
  "Initialization method sets the weights to uniformly distributed random
numbers."
  (let ((layers (send self 'layers)))
    (mapcar
     #'(lambda(x y) (send x :initialize y)) (rest layers) (butlast layers))
    'okay))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :data
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth net :data (input output)
  "Data method initializes input and output layers, storing the result in the
'type slot of the output layer for future reference."
  (let ((layers (send self 'layers)))
    (send (first layers) 'nodes input)
    (send (final layers) 'nodes output)
    (send (final layers) 'type output)
    'okay))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :output
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth net :output () 
  "Output method reports current state of output nodes."
  (send (final (send self 'layers)) 'nodes))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :forward
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth net :forward ()
  "Forward method takes the net through one forward step."
  (let ((layers (send self 'layers)))
    (mapcar 
     #'(lambda(x y) (send x :forward y)) (rest layers) (butlast layers))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :backward
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth net :backward ()
  "Backward method takes the net through one backward step."
  (let ((layers (send self 'layers))
	(learning-rate (send self 'learning-rate)))
    (mapcar 
     #'(lambda(x y z) (send x :backward y z learning-rate))
     (reverse (rest layers))
     (reverse (butlast layers))
     (reverse (combine (rest (rest layers)) nil)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :describe
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth net :describe ()
  "Describe method reports on the layers making up the net."
  (let ((layers (send self 'layers)))
    (mapcar #'(lambda(x y) (send x :describe y)) layers (iseq (length layers)))
    ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :learn
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth net :learn (inputs outputs
			    &key (its 1000) (eps 1e-6) (print 10) plot)
  "
Net method :learn takes input and output data as arguments, along with keys
	(its 1000) ;; its is the number of iterations
	(eps 1e-6) ;; how close we have to be to the correct response
	(print 10) ;; print should be a positive integer, which represents how
		   ;; often the output goes to the screen

"
  (setq print (if (integerp print) print nil))
  (let (
	(weights (list (send self :weights)))
	(net-type (send self 'net-type))
	)
    (do
     ((i 0 (1+ i))
      (stuff 1e31 (mapcar 
		   #'(lambda(x y)
		       (send self :data x y)
		       (case net-type
			     ('perceptron
			      (send self :perceptron-forward
				    :adapt
				    (if (search (list 'widrow-hoff)
						(send self 'options))
					'widrow-hoff))
			      (send self :perceptron-backward)
			      )
			     (t
			      (send self :forward)
			      (send self :backward)
			      )
			     )
		       (sum (square (- (send self :output) y)))
		       )
		   inputs outputs))
      )
     ((or (> i its) (< (sum stuff) eps)))
     (if print 
	 (if (integerp (/ i print))
	     (let ()
	       (if (/= stuff 1e31) (print stuff))
	       (if plot
		   (setq weights (append weights (list (send self :weights)))))
	       ))
       (if plot (setq weights (append weights (list (send self :weights)))))
       )
     )
    (if plot 
	(let ((structure (first weights)))
	  (dotimes
	   (i (length structure))
	   (dotimes
	    (j (length (elt structure i)))
	    (title (format nil "Layer ~d, Element ~d" (+ 2 i) j))
	    (plotter
	     (apply
	      #'bind-columns
	      (mapcar #'(lambda(x) (elt (elt x i) j)) weights)))
	    )
	   )
	  )
      )
    )
  ) 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :weights
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth net :weights (&optional args)
  (if args
      (let ()
	(mapcar (lambda(x y) (send x 'weights y))
		(rest (send self 'layers))
		args
		)
	nil
	)
    (mapcar (lambda(x) (send x 'weights)) (rest (send self 'layers)))
    )
  )
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :momentum
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth net :momentum (&optional momentum)
  (if momentum
      (if (and (numberp momentum) (< momentum 1) (>= momentum 0))
	  (let ()
	    (mapcar (lambda(x) (send x 'momentum momentum)) (send self 'layers))
	    (send self 'momentum momentum)
	    )
	)
    (send self 'momentum)
    )
  )
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :test
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth net :test (&rest args)
  "Test method 

	net :test (inputs &optional (outputs (repeat 0 (length inputs)))
			  &key (print t))

takes input and output data as test arguments, and returns the
actual results for comparison. Returns actual output.
"
  (let* (
	 (inputs (first args))
	 (args (rest args))
	 (print (first (catch-keys '(:print) args)))
	 (args (remove-keys '(:print) args))
	 (set (if args t nil))
	 (outputs (if set (first args) (repeat 0 (length inputs))))
	 )
    (mapcar 
     #'(lambda(x y)
	 (send self :data x y)
	 (send self :forward)
	 (let ((actual (send self :output)))
	   (if print
	       (if set
		   (format
		    t "~%Input data: ~s~%Output data: ~s~%Actual     : ~s~%"
			   x y (send self :output))
		 (format t "~%Input data: ~s~%Actual     : ~s~%"
			 x (send self :output))
		 )
	     )
	   actual
	   )
	 )
     inputs outputs)
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :perceptron-forward
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth net :perceptron-forward (&key (adapt 'widrow-hoff))
  "perceptron-forward method takes the layer through one forward step. Requires
previous-layer as an argument."
  (let* (
	 (layers (send self 'layers))
	 ;; this started as a layer method, which accounts for the name:
	 (layer-self (final layers))
	 (previous-layer (first layers))
	 (my-weights (send layer-self 'weights))
	 (their-vals (combine (send previous-layer 'nodes) 1))
	 (net-values
	  (mapcar
	   #'(lambda(weights) (sum (* weights their-vals)))
	   my-weights)))
    (send layer-self 'nodes 
	  (if (eq adapt 'widrow-hoff) (heaviside net-values)
	    (heaviside net-values)))
    ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :perceptron-backward
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth net :perceptron-backward()
  "perceptron-backward method takes this very simple net through one backward
step. Requires previous-layer, doesn't use next-layer, and learning-rate as
arguments.
" 
  (let* (
	 (layers (send self 'layers))
	 ;; this started as a layer method, which accounts for the name:
	 (layer-self (final layers))
	 (previous-layer (first layers))
	 (weights (first (send layer-self 'weights)))
	 (inputs (combine (send previous-layer 'nodes) 1))
	 (output (first (send layer-self 'nodes)))
	 (data (send layer-self 'type))
	 (sigma (- data output))
	 (del-weights (* sigma (send self 'learning-rate) inputs))
	 )
    (if (/= sigma 0) (send layer-self 'weights (list (+ weights del-weights))))
    ))
	
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                Layer methods:
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :initialize
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth layer :initialize (previous-layer)
  "Initialization method sets the weights to uniformly distributed random
numbers."
  (send
   self 'weights
   (mapcar
    #'(lambda(x) (uniform-rand (1+ (length (send previous-layer 'nodes)))))
    (send self 'nodes))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :forward
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth layer :forward (previous-layer)
  "Forward method takes the layer through one forward step. Requires
previous-layer as an argument."
  (let* ((my-weights (send self 'weights))
	 (their-vals (combine (send previous-layer 'nodes) 1))
	 (net-values
	  (mapcar
	   #'(lambda(weights) (sum (* weights their-vals)))
	   my-weights)))
    (send self 'net-values net-values)
    (send self 'nodes (mapcar (send self 'sigmoid) net-values))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :backward
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth layer :backward (previous-layer next-layer learning-rate)
  "Backward method takes the layer through one backward step. Requires
previous-layer, next-layer, and learning-rate as arguments."
  (let* ((my-weights (send self 'weights))
	 (their-vals (combine (send previous-layer 'nodes) 1))
	 (sigma
	  (if next-layer
	      (mapcar #'(lambda(x) (dot (send next-layer 'delta) x))
		      (transpose (mapcar #'butlast (send next-layer 'weights)))
		      )
	    (- (send self 'type) (send self 'nodes))))
	 (delta
	  (* sigma
	     (mapcar (send self 'sigmoid-prime) (send self 'net-values))))
	 (del-weights 
	  (row-lists (* learning-rate (outer-product delta their-vals))))
	 (momentum (send self 'momentum))
	 (previous-weights (send self 'previous-weights))
	 )
    (send self 'delta delta)
    (send self 'weights (+ my-weights del-weights
			   (if (and momentum previous-weights)
			       (* momentum (- my-weights previous-weights))
			     0)))
    (if momentum (send self 'previous-weights my-weights))
    )
  )
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :describe
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth layer :describe (&optional layer-number (stream t))
  "Describe method reports on the nodes and weights making up the layer."
  (let (
	(weights (send self 'weights))
	(nodes (send self 'nodes))
	)
    (if (integerp layer-number) (format stream "Layer: ~d~%" layer-number))
    (format stream "Nodes: ~s~%Weights: ~s~%" nodes weights)
    weights))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                Examples:
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                Example 1:
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun example1()
  "
;; Examples:
  (def nn1 (send net :new '(2 3 1)))
  (send nn1 :initialize)
  (send nn1 :learn '((1 1)) '(0) :its 100)
  (send nn1 :test '((1 1)))
  (send nn1 :test '((1 1)) '(0))
  (send nn1 :test '((1 1)) '(0) :print nil)
  
  (send nn1 'learning-rate .5)
  (send nn1 :initialize)
  (send nn1 :learn '((1 1)) '(0) :its 100)
  (send nn1 :test '((1 1)) '(0))
"
  nil
  )
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                     Example 2: exclusive or
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun example2()
  "
;; Examples:
  (def nn2 (send net :new '(2 5 1)))
  (send nn2 :initialize)
  (send nn2 :learn '((1 1) (1 0) (0 1) (0 0)) '(0 1 1 0) :print 100)
  (send nn2 :test '((1 1) (1 0) (0 1) (0 0)) '(0 1 1 0))
  
  ;; Do we really need 5 internal hidden neurons? Let's try just 2:
  (def nn2a (send net :new '(2 2 1)))
  (send nn2a :initialize)
  (def weights (send nn2a :weights))
  (send nn2a :learn '((1 1) (1 0) (0 1) (0 0)) '(0 1 1 0) 
             :print 100 :its 500 :plot t)
  (send nn2a :test '((1 1) (1 0) (0 1) (0 0)) '(0 1 1 0))

  (send nn2a :weights weights)
  (send nn2a :momentum .5)
  (send nn2a :learn '((1 1) (1 0) (0 1) (0 0)) '(0 1 1 0) :print 100 :its 500)
  (send nn2a :test '((1 1) (1 0) (0 1) (0 0)) '(0 1 1 0))

  (send nn2a :test 
	(mapcar (lambda(x) (+ x (* .25 (rand.u 2)))) '((1 1) (1 0) (0 1) (0 0)))
	'(0 1 1 0))
  (send nn2a :weights)
  
  (let* (
	 (x (rand.u 100))
	 (y (rand.u 100))
	 (z (mapcar #'first (send nn2 :test (weave x y) :print nil)))
	 )
    (lines off)
    (plot `,x `,y `,z)
    )
"
  nil
  )
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                Example 3:
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun example3()
  "
;; Examples:
  (def nn3 (send net :new '(2 2 2 1)))
  (send nn3 :initialize)
  (send nn3 :learn '((1 1)) '(0) :its 1000 :print 100)
  (send nn3 :test '((1 1)) '(0))
"
  nil
  )
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                Example 4:
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun example4()
  "
;; Examples:
  (def nn4 (send net :new '(2 5 5 1)))
  (send nn4 :initialize)
  (send nn4 :learn '((1 1)) '(0) :its 1000 :print 100)
  (send nn4 :test '((1 1)) '(0))
"
  nil
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                Example 5: Perceptron
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun example5()
  "
Perceptron example. A perceptron is a neuron, which can learn and do linear
discrimination. As an example, we create two data sets which are linearly
separated:

	I thought that I'd implement the widrow-hoff rule, but I don't see how
it's different from what I did before, so I pithed it....

;; Examples:
 (def nn5 (send net :new '(2 1)))
 (send nn5 'net-type 'perceptron)
 (send nn5 :initialize)
 (def weights (send nn5 :weights))
 (send nn5 :learn '((-1 -1) (-2 -1) (1 1) (2 1)) '(1 1 0 0))
 (send nn5 :test '((-1 -1) (-2 -1) (1 1) (2 1)) '(1 1 0 0))

 (send nn5 :weights weights)
 (send nn5 'options (list 'widrow-hoff))
 (send nn5 :learn '((-1 -1) (-2 -1) (1 1) (2 1)) '(1 1 0 0))
 (send nn5 :test '((-1 -1) (-2 -1) (1 1) (2 1)) '(1 1 0 0))

 (def xs (combine (- (rand.u 20)) (rand.u 20)))
 (def ys (combine (- (rand.u 20)) (rand.u 20)))
 (def targets (combine (repeat 1 20) (repeat 0 20)))
 (send nn5 :learn (transpose (list xs ys)) targets)
 (round (send nn5 :test (transpose (list xs ys)) targets))
 (lines off)
 (plot)
 (def weights (first (first (send nn5 :weights))))
 (def intercept (third weights))
 (def slope-x (first weights))
 (def slope-y (second weights))
 (send my-plot :add-plot-item 
               (list 'points
                 (select xs (iseq 20)) (select ys (iseq 20)) 'blue 'cross)
               (list 'points
                 (select xs (iseq 21 39)) (select ys (iseq 21 39)) 'red 'square)
               (lambda(x) (/ (- intercept (* slope-x x)) slope-y)))

 ;; The following sets are generally not quite separated:
 (def xs (combine (+ 1 (rand.n 20)) (+ -1 (rand.n 20))))
 (def ys (combine (+ 1 (rand.n 20)) (+ -1 (rand.n 20))))
 (def targets (combine (repeat 1 20) (repeat 0 20)))
 (send nn5 :learn (transpose (list xs ys)) targets)
 (send nn5 :test (transpose (list xs ys)) targets)
 (lines off)
 (plot)
 (def weights (first (first (send nn5 :weights))))
 (def intercept (third weights))
 (def slope-x (first weights))
 (def slope-y (second weights))
 (send my-plot :add-plot-item 
               (list 'points
                 (select xs (iseq 20)) (select ys (iseq 20)) 'blue 'cross)
               (list 'points
                 (select xs (iseq 21 39)) (select ys (iseq 21 39)) 'red 'square)
               (lambda(x) (/ (- intercept (* slope-x x)) slope-y)))
"
  nil)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(format t "~%
There are currently five examples available: type
         (try-out example1)
         (try-out example2)
         (try-out example3)
         (try-out example4)
         (try-out example5) ;; a perceptron
to see the examples, and see them in action.~2%")
