#| - Thu Feb 26 22:52:20 1998

	I've got a function which implements newton's method for non-linear
systems, and I want to use it to find the singular vectors of a matrix. I
consider below a couple of examples, including a 2x2 and 3x3.

	I worked up the analysis for the general matrix case, and coded it up,
and I've also considered it by AA^T. Both work, sort of, but instead of finding
the first singular vector/value, they find the last!

	I spent a good deal of time and effort trying to turn that around, but
realized that, given the way I was proceeding, I had two zeros and neither had
precedence that I could see (one is a maximum, the other is a minimum on the
unit sphere).

	So I gave up on trying to find the singular values in descending order,
and decided to change my objective function to include orthogonality to
preceding vectors. This now seems to work.

	In the case of the 3x3 matrix, the example shows the pattern that the
jacobian matrix follows - it generalizes fairly easily, so that I could create
a function which generates the svd. This is not too remarkable: I've already
once coded the tranposed qr algorithm for the svd (?robust-svd). What is
essential here is that I may have worked up a method for the calculation which
I can generalize to the tensor case.

	This is giving excellent results 

							Andy Long
							Fearless Programming
							IMSP
							February 27, 1998

|#
;; this here's our function:
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;				non-linear-system-root
;;
(defun non-linear-system-root(f jack x0 &key (eps 1e-6) (maxits 100) verbose)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;			non-linear-system-root Doc:
;;
;; Author: Andy Long
;; Created: Wed Feb 25 20:51:44 1998
;;
;; Description/Arguments: non-linear-system-root
  
;; Arguments: f jack x0 &key (eps 1e-6) (maxits 100)

;; Returns: vector

;; Notes: For DESS class at the IMSP

;; See Also: newtonmax

;; Examples:
 ;; How about trying to solve linear systems?
 (def A #2a((2 4)(5 2)))
 (def b #(2 1))
 ;; now we set up the two functions, Ax-b:
 (defun f(z)
  (let ((x (first z))
	(y (second z)))
    (list
     (- (+ (* 2 x) (* 4 y)) 2)
     (- (+ (* 5 x) (* 2 y)) 1))))
 (defun jack(z) A)
 ;; and compare solutions:
 (non-linear-system-root #'f #'jack (list 0 0) :eps 1e-12)
 (solve A b)
"
  (do*
   (
    (i 1 (1+ i))
    (x0 x0 x1)
    (x1 
     (+ x0 (solve (funcall jack x0) (- (funcall f x0))))
     (+ x0 (solve (funcall jack x0) (- (funcall f x0))))
     )
    )
   ((or (> i maxits) (< (norm (- x1 x0)) eps)) (print i) x1)
   (if verbose (print x1))
   )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Here's our example:
(def A #2a((2 4)(5 2)))
;; but as the process didn't seem to work when I tried to find u and v
;; simultaneously, I now consider M=AA':
(def M (matmult A (transpose A)))

;; Now I'm going to calculate the first singular vector of M:
(defun funcs(z)
  (let ((u (butlast z))
	(l (final z)))
    (combine
     (* 2 (- (matmult M u) (* l u)))
     (- 1 (dot u u))
     )))
(defun jack (z)
  (let ((u (butlast z))
	(l (final z)))
    (* -2 
       (combine-matrices
	(list (- (* (- l) (id (length u))) M) (bind-columns u))
	(combine u 0))
       )
    ))
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(def first-schmidt-pair
     (non-linear-system-root #'funcs #'jack 
			     (combine (normalize (rand.n 2)) (rand.n 1))
			     :eps 1e-12))
     
(mapcar #'mprint (butlast (svd M)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Even if I start close to the other it is missed - that means unstable. How
;; can I turn it around????
(non-linear-system-root #'funcs #'jack '(-0.615 -0.788 43) :eps 1e-12)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;			Now for the second:
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Rather than fight them, I should join them: I simply change all so that my
;; function is also orthogonal to the first one:
;; Now I'm going to calculate the first singular vectors of AA^T:
(def v (butlast first-schmidt-pair))
(def v0 (first v))
(def v1 (second v))
(defun n-funcs(z)
  (let ((u (select z (iseq 2)))
	(l (third z))
	(l1 (fourth z))
	)
    (combine
     (* 2 (- (matmult M u) (* l u) (* l1 v)))
     (- 1 (dot u u))
     (- (dot u v))
     )))

(defun n-jack (z)
  (let* ((u (select z (iseq 2)))
	 (l (third z))
	 (l1 (fourth z))
	 (lagrange (bind-columns u (* .5 v)))
	 )
    (* -2 
       (combine-matrices
	(list (- (* (- l) (id (length u))) M) lagrange)
	(list (transpose lagrange)  (zeros 2 2)))
       )))
    
(non-linear-system-root #'n-funcs #'n-jack '(-0.615 -0.788 43 1) :eps 1e-12)
(non-linear-system-root #'n-funcs #'n-jack
			(combine (normalize (rand.n 2)) (rand.n 2))
			:eps 1e-12
			:maxits 10
			)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;			 Here's a 3-d example:
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(def A #2a((2 4 -1)(5 6 2)(1 2 1)))
(def M (matmult A (transpose A)))

;; Now I'm going to calculate the first singular vector of M:
(defun funcs(z)
  (let ((u (butlast z))
	(l (final z)))
    (combine
     (* 2 (- (matmult M u) (* l u)))
     (- 1 (dot u u))
     )))
(defun jack (z)
  (let ((u (butlast z))
	(l (final z)))
    (* -2 
       (combine-matrices
	(list (- (* (- l) (id (length u))) M) (bind-columns u))
	(combine u 0))
       )
    ))
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(def first-schmidt-pair
     (non-linear-system-root #'funcs #'jack (rand.n 5) :eps 1e-12))
(mapcar #'mprint (butlast (svd M)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;			Now for the second:
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Rather than fight them, I should join them: I simply change all so that my
;; function is also orthogonal to the first one:
;; Now I'm going to calculate the first singular vectors of AA^T:
(def v (butlast first-schmidt-pair))
(def v0 (first v))
(def v1 (second v))
(defun n-funcs(z)
  (let ((u (butlast (butlast z)))
	(l (third z))
	(l1 (fourth z))
	)
    (combine
     (* 2 (- (matmult M u) (* l u) (* l1 v)))
     (- 1 (dot u u))
     (- (dot u v))
     )))

(defun n-jack (z)
  (let* ((u (butlast (butlast z)))
	 (l (third z))
	 (lagrange (bind-columns u (* .5 v)))
	 )
    (* -2 
       (combine-matrices
	(list (- (* (- l) (id (length u))) M) lagrange)
	(list (transpose lagrange)  (zeros 2 2)))
       )))
    
(non-linear-system-root #'n-funcs #'n-jack (rand.n 5)
			:eps 1e-12
			:maxits 10
			:verbose t
			)



















(require "nonnewt")


(def A #2a((2 4)(5 2)))
(def M (matmult A (transpose A)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;				Let's try steepest descent:
#| - Thu Feb 26 22:41:16 1998
	Steepest descents is a poor success - it doesn't take the unit vector
aspect into account.
|#
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun f(u) (/ (matmult u M u) (dot u u)))
(newtonmax #'f (normalize (rand.n 2)) :verbose 1 :count-limit 100)
;; But that doesn't work nearly as well as the non-linear-system-root.

#| - Thu Feb 26 10:05:07 1998
;; Now I'm going to calculate the first singular vectors of A:
 (def funcs
     (list
      (defun us0(u0 u1 v0 v1 l0 l1) (+ (* v0 (aref A 0 0))
				       (* v1 (aref A 0 1))
				       (- (* l0 u0))))
      (defun us1(u0 u1 v0 v1 l0 l1) (+ (* v0 (aref A 1 0))
				       (* v1 (aref A 1 1))
				       (- (* l0 u1))))
      (defun vs0(u0 u1 v0 v1 l0 l1) (+ (* u0 (aref A 0 0))
				       (* u1 (aref A 1 0))
				       (- (* l1 v0))))
      (defun vs1(u0 u1 v0 v1 l0 l1) (+ (* u0 (aref A 0 1))
				       (* u1 (aref A 1 1))
				       (- (* l1 v1))))
      (defun ls0(u0 u1 v0 v1 l0 l1) (* -.5 (- (+ (* u0 u0) (* u1 u1)) 1)))
      (defun ls1(u0 u1 v0 v1 l0 l1) (* -.5 (- (+ (* v0 v0) (* v1 v1)) 1)))
      ))
 (defun jack
  (u0 u1 v0 v1 l0 l1)
  (combine-matrices
   (list (* (- l0) (id 2)) A (bind-columns (list u0 u1) (zeros 2 1)))
   (list (transpose A) (* (- l1) (id 2)) (bind-columns (zeros 2 1) (list v0 v1)))
   (list (bind-rows (- (list u0 u1)) (zeros 1 2))
	 (bind-rows (zeros 1 2) (- (list v0 v1)))
	 (id 2))
   )
  )
 (non-linear-system-root funcs '(u0 u1 v0 v1 l0 l1)
			 :jack #'jack
			 :x0 (combine
			      (normalize (rand.n 2))
			      (normalize (rand.n 2))
			      10 10)
			 )
|#

;; Now I'm going to calculate the first singular vectors of AA^T:
(def funcs
     (list
      (defun us0(u0 u1 l) (* 2 (+ (* u0 (aref m 0 0))
				  (* u1 (aref m 0 1))
				  (- (* l u0)))))
      (defun us1(u0 u1 l) (* 2 (+ (* u0 (aref m 1 0))
				  (* u1 (aref m 1 1))
				  (- (* l u1)))))
      (defun ls0(u0 u1 l) (- 1 (+ (* u0 u0) (* u1 u1))))
      ))
(defun jack (u0 u1 l)
  (* -2 
     (combine-matrices
      (list (- (* (- l) (id 2)) M) (bind-columns (list u0 u1)))
      (combine (list u0 u1) 0))
     )
  )
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Even if I start close to the other it is missed - that means unstable. How
;; can I turn it around????
(non-linear-system-root funcs '(u0 u1 l)
			     :jack #'jack
			     :x0 '(-0.615 -0.788 43)
			     :eps 1e-12
			     )
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(def first-schmidt-pair
     (non-linear-system-root funcs '(u0 u1 l)
			     :jack #'jack
			     :x0 (combine
				  (normalize (rand.n 2))
				  10)
			     :eps 1e-12
			     )
     )
(mapcar (lambda(f) (apply f FIRST-SCHMIDT-PAIR)) funcs)
(mapcar #'mprint (butlast (svd M)))
(mapcar (lambda(f) (apply f (list -0.615412 -0.788205 43.0540))) funcs)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;			Now for the second:
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
#| - Thu Feb 26 22:15:00 1998
;; this doesn't work.
(def M (- M (* (third first-schmidt-pair) (outer-product
					   (butlast first-schmidt-pair)
					   (butlast first-schmidt-pair)))))
(defun jack (u0 u1 l)
  (* -2 
     (combine-matrices
      (list (- (* (- l) (id 2)) M) (bind-columns (list u0 u1)))
      (combine (list u0 u1) 0))
     )
  )
(def second-schmidt-pair
     (non-linear-system-root funcs '(u0 u1 l)
			     :jack #'jack
			     :x0 (combine
				  (normalize (rand.n 2))
				  10)
			     )
     )
|#

;; Rather than fight them, I should join them: I simply change all so that my
;; function is also orthogonal to the first one:
;; Now I'm going to calculate the first singular vectors of AA^T:
(def v (butlast FIRST-SCHMIDT-PAIR))
(def v0 (first v))
(def v1 (second v))
(def n-funcs
     (list
      (defun n-us0(u0 u1 l l1) (* 2 (+ (* u0 (aref m 0 0))
				       (* u1 (aref m 0 1))
				       (- (* l u0))
				       (- (* l1 v0))
				       )
				  )
	)
      (defun n-us1(u0 u1 l l1) (* 2 (+ (* u0 (aref m 1 0))
				       (* u1 (aref m 1 1))
				       (- (* l u1))
				       (- (* l1 v1))
				       )
				  )
	)
      (defun n-ls0(u0 u1 l l1) (- 1 (+ (* u0 u0) (* u1 u1))))
      (defun n-ls1(u0 u1 l l1) (- (+ (* u0 v0) (* u1 v1))))
      ))
(defun n-jack (u0 u1 l l1)
  (* -2 
     (combine-matrices
      (list (- (* (- l) (id 2)) M)
	    (bind-columns (list u0 u1) (* .5 (list v0 v1)))
	    )
      (list (bind-rows (list u0 u1) (* .5 (list v0 v1))) (zeros 2 2)))
     )
  )
(non-linear-system-root n-funcs '(u0 u1 l l1)
			     :jack #'n-jack
			     :x0 '(-0.615 -0.788 43 1)
			     :eps 1e-12
			     )
(non-linear-system-root n-funcs '(u0 u1 l l1)
			     :jack #'n-jack
			     :x0 (combine (normalize (rand.n 2)) (rand.n 2))
			     :eps 1e-12
			     :maxits 10
			     )









#| - Thu Feb 26 22:50:39 1998
;; Tried to get by without using the second lagrange multiplier, but without
;; any luck.
(def v (butlast FIRST-SCHMIDT-PAIR))
(def v0 (first v))
(def v1 (second v))
(def n-funcs
     (list
      (defun n-us0(u0 u1 l) (* 2 (+ (* u0 (aref m 0 0))
				    (* u1 (aref m 0 1))
				    (- (* l u0))
				    (- (* .5 v0))
				    )
			       )
	)
      (defun n-us1(u0 u1 l) (* 2 (+ (* u0 (aref m 1 0))
				    (* u1 (aref m 1 1))
				    (- (* l u1))
				    (- (* .5 v1))
				    )
			       )
	)
      (defun n-ls0(u0 u1 l) (- 1 (+ (* u0 u0) (* u1 u1))))
      ))
(non-linear-system-root n-funcs '(u0 u1 l l1)
			     :jack #'jack
			     :x0 '(-0.615 -0.788 43)
			     :eps 1e-12
			     )
|#
