
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                perko.lsp
#|
	This is a collection of examples and exercises from the book
"Differential Equations and Dynamical Systems", by Lawrence Perko,
Springer-Verlag, New York, 1991.

	One of the really interesting things in this book is a discussion of
how to get (approximations to) the (local) stable and unstable manifolds of
a dynamical system, and that's something that I'd like to develop in
software. The example I'm thinking of is Example 2, page 111, based on work on
the stable manifold theorem (page 107).

	Successive approximations are used to calculate (p. 108). Looks to me
like the procedure is as follows: given a hyperbolic critical point for the
system

(1)	x'(t) = f(x)

 assume that the zero of f is at 0, and replace (1) by

(2)	x'(t) = Ax + F(x),

where A=J(0), F(x)=f(x)-Ax.

	In the book, Perko then splits A so that

			[ P  0 ]
(3)	B = C^{-1}AC =  [      ]
			[ 0  Q ]

and P has negative realpart eigenvalues and Q positive; I think that this could
be replaced by letting J be the Jordan form of A,
			
(3')	B = C^{-1}AC = J

then associating P with the "negative" parts of the Jordan form and Q with the
"positive".

	Let

	y = C^{-1}x,

	G(y) = C^{-1}F(Cy),

	U(t) = exp(Pt),
and
	V(t) = exp(Qt):

Then consider the integral equation

(4)	u(t,a) = U(t)a + \int_0^t      U(t-s)G(u(s,a))ds
		       - \int_t^\infty V(t-s)G(u(s,a))ds

where a is a vector with unknown components where P is non-zero on the
diagonal. To solve this integral equation we generate the iteration sequence

(4')	u^{(j+1)}(t,a) = U(t)a  + \int_0^t      U(t-s)G(u^{(j)}(s,a))ds
				- \int_t^\infty V(t-s)G(u^{(j)}(s,a))ds

	If we start the sequence with u^0(t,a) = 0, then the first
approximation is

	u^1(t,a) = U(t)a

and the second approximation is the first interesting (non-linear) one.

	
|#
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                page 56
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(def a #2a((-2 -1 0)(1 -2 0)(0 0 -3)))
(eigenvalues a)
(defun page56(x) (matmult a x))
(orbitals #'page56
	  (* 10 (normal-rand 3))
	  1 50 10
	  :randomize 10
	  :randomizer #'normal-rand
	  :title "page 56")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                page 87
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun page87(vals)
  (let (
	(x (elt vals 0))
	(y (elt vals 1))
	(z (elt vals 2))
	)
    (list (- (/ y (square z)))
	  (/ x (square z))
	  1)))
(orbitals #'page87 (list 0 -1 (/ 1 pi)) 1 500 10 :randomize 1 :title "page 87"
	  :integrator #'rk-4)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                page99
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun page99(vals)
  (let (
	(x (elt vals 0))
	(y (elt vals 1))
	)
    (list (- x)
	  (+ y (square x)))
    ))
(phase-plane #'page99 (list .1 .1) 1 100 10
             :randomize 1
             :randomizer #'normal-rand
             :title "Page 99")

(defun page99(x y) (/ (+ y (square x)) (- x)))
(slope-field page99 -2 2 -2 2)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                page111
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun page111(vals)
  (let (
	(x (elt vals 0))
	(y (elt vals 1))
	)
    (list (- (+ x (square y)))
	  (+ y (square x)))
    ))
(defun initcond(a) (list (/ (square a) -3) a))
(phase-plane #'page111 (initcond .001) 5 1000 1 :title "Page 111")
(phase-plane #'page111 (* '(1 -1) (initcond .001)) 10 2000 1
	     :old-window my-plot)
(defun initcond(a) (list a (/ (square a) -3)))
(defun neg-page111 (vals) (- (page111 vals)))
(phase-plane #'neg-page111 (initcond .001) 5 1000 1
	     :old-window my-plot)
(phase-plane #'neg-page111 (* '(-1 1) (initcond .001)) 10 2000 1
	     :old-window my-plot)

(defun page111(x y) (/ (+ y (square x)) (- (+ x (square y)))))
(slope-field page111 -2 2 -2 2 40 40)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                page146
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun page146(vals)
  (let (
	(x (elt vals 0))
	(y (elt vals 1))
	)
    (list y
	  (- (* 4 x y) (cube x)))
    ))
(phase-plane #'page146 (list 0 0) .25 10 100
             :randomize 1
             :randomizer #'normal-rand
             :title "Page 146")

(defun page146(x y) (/ (- (* 4 x y) (cube x)) y))
(slope-field page146 -2 2 -2 2 40 40)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                page147
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun page147(x y) (/ y (square x)))
(slope-field page147 -2 2 -2 2 40 40)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                page148
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun page148(x y) (/ (square x) y))
(slope-field page148 -2 2 -2 2 40 40)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                page150
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun page150(x y) (/ (+ (half (square y)) (* x y))
		       (+ (* x (+ x y)))))
(slope-field page150 -pi 3.14 -pi 3.15 44 42)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                page177
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun page177(vals)
  (let* (
	 (x (elt vals 0))
	 (y (elt vals 1))
	 (rsq-term (- 1 (+ (square x) (square y))))
	 )
    (list
     (+ (- y) (* x rsq-term))
     (+     x (* y rsq-term))
     )))
(phase-plane #'page177 (list 1 0) .5 20 10
	     :integrator #'rk-4
             :randomize 1
             :randomizer #'normal-rand
             :title "Page 177")

(defun page177(x y)
  (let ((rsq-term (- 1 (+ (square x) (square y)))))
    (/ 
     (+     x (* y rsq-term))
     (+ (- y) (* x rsq-term))
     )
    )
  )
(slope-field page177 -2 2 -2 2 40 40)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                page217
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun page217(vals)
  (let (
	(x (elt vals 0))
	(y (elt vals 1))
	(z (elt vals 2))
	)
    (list 
     (+ (- y) (* x (square z)))
     (+     x (* y (square z)))
     (* (- z) (sum (square x y)))
     )))
(orbitals #'page217 (list 0 -1 0) 10 500 10
	  :randomizer (lambda(i) (normalize (* '(1 1 2) (rand.n i))))
	  :integrator #'rk-4)

;; Let's see how a two-d and four-d systems gets represented:
(defun trial(vals)
  (let (
	(x (elt vals 0))
	(y (elt vals 1))
	)
    (list 
     (+ x (- y))
     (+ x    y )
     )))
(orbitals #'trial (list 0 -1) 10 500 10
	  :randomizer (lambda(i) (normalize (rand.n i)))
	  :integrator #'rk-4)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                page225
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun page225(vals)
  (let (
	(x (elt vals 0))
	(y (elt vals 1))
	(z (elt vals 2))
	(w (elt vals 3))
	)
    (list 
     y
     (- x)
     w
     (* z (- z 1))
     )))
(orbitals #'page225 (list 0 -1 0 0) 10 500 1 :integrator #'rk-4)
(orbitals #'page225 (list 0 -1 1 0) 10 500 1 :integrator #'rk-4)
(orbitals #'page225 (list 0 -1 0 0) 10 500 10
	  :randomize-about-initial t
	  :randomize .01
	  :randomizer (lambda(i) (combine (normalize (rand.n (1- i))) (rand.n 1)))
	  :integrator #'rk-4)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                page240
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun page240(vals)
  (let* (
	 (x (elt vals 0))
	 (y (elt vals 1))
	 (x2 (square x))
	 )
    (list 
     (- y (* x (+ (* x2 (- (* .32 x2) 4/3)) .8)))
     (- x)
     )))
(orbitals #'page240 (list 0 -1) 10 500 10
	  :randomize 1.5
	  :randomizer #'normal-rand
	  :integrator #'rk-4
	  :title "two limit cycles: page 240"
	  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                page242
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun page242(vals &optional (eps .01))
  (let* (
	 (x (elt vals 0))
	 (y (elt vals 1))
	 (x2 (square x))
	 )
    (list 
     (+ y (* eps x (+ 72 (* x2 (+ -392/3 (* x2 (+ 224/5 (* x2 -128/35))))))))
     (- x)
     )))
(orbitals #'page242 (list 0 -1) 10 500 20
	  :randomize 1.5
	  :randomizer #'normal-rand
	  :integrator #'rk-4
	  :title "three limit cycles: page 242"
	  )



