#|

	I read in a book that Joel gave me to read ("Neural Computing: An
Introduction", by R. Beale and T. Jackson, Adam Hilger, Bristol) about MENACE
(Matchbox Educable Noughts and Crosses Engine), which learns to play the game
of "naughts and crosses" (i.e., tic-tac-toe) by the simple addition and
deletion of match sticks from matchboxes.

	Far out! As I said to Joel, this has got to be how the brain learns. So
I decided to see for myself how it works. The trouble with this model is that
it entails the creation of "Grandmother" neurons (neurons which respond to
individuals, which is clearly unrealistic given the rate at which neurons die,
and the fact that we continue to recognize Grandmothers).

	One little modification that I had to make was to prevent a cell from
"dying" (that is, as I remove unsuccessful strategies, the number of match
sticks may reach zero).	Rather than "remove matchsticks" I instead strengthen
or weaken the connection: the probability of choosing a strategy may tend to
zero, but will not equal zero: instead, each potential move starts off
equiprobable, and the numbers change as the game progresses. Thus, if the
potential moves are:

	1 2 4 5 8

then the probabilities associated with each change according to

	p(1) p(2) p(3) p(4) p(5)

summing to 1, and they are altered as follows:

	if choice 1 loses, then
p(1) = 9*p(1)/10
with the others gaining the missing p(1)/10.

	Last of all, I've discovered that I should only correct one side or the
other (and as the response moves to the opening are fewer (289, according to my
calculations, while the article says 288), it's best to correct the even
ones. Thus, the machine always plays defense. 

	The fact that I can't train both sides at the same time is obvious (in
retrospect), because if it trains against itself, and one side tends to always
make the same dumb move at some point, then MENACE will teach itself to respond
as if the opponent will always make the same dumb move. There's a lesson in
here about human learning, too!

							Andy Long
							November, 1997
|#
#|

	Preliminary stuff, to decide when two boards are equivalent in
tic-tac-toe. I want to see how many different moves there are, effectively, for
the game:
|#
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                rotate-clock-wise-matrix
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun rotate-clock-wise-matrix(a n)
  "
;; Examples:
 (setq A (matrix '(3 3) '(a b c d e f g h i)))
 (mprint a)
 (mprint (rotate-clock-wise-matrix a 1))
 (mprint (rotate-clock-wise-matrix a 2))
 (mprint (rotate-clock-wise-matrix a 3))
"
  (do
   ((i n (decf i)))
   ((= i 0) a)
   (setq a (apply #'bind-columns (reverse (row-lists a))))
   )
  )


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                equal-by-rotation
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun equal-by-rotation(a b)
  "
;; Examples:
 (mprint (setq A (matrix '(3 3) '(a b c d e f g h i))))
 (mprint (setq B (rotate-clock-wise-matrix a 1)))
 (equal-by-rotation a b)	;; t
 (equal-by-rotation a (id 3))	;; nil
"
  (or
   (matrix-equalp a b)
   (matrix-equalp (rotate-clock-wise-matrix a 1) b)
   (matrix-equalp (rotate-clock-wise-matrix a 2) b)
   (matrix-equalp (rotate-clock-wise-matrix a 3) b)
   )
  )


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                equal-board-p
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun equal-board-p (a b)
  "
;; Examples:
 (mprint (setq a (matrix '(3 3) '(x o - - - - - - -))))
 (mprint (setq b (matrix '(3 3) '(x - - o - - - - -))))
 (equal-board-p a b)
 (mprint (setq c (matrix '(3 3) '(x - - o - - - - x))))
 (equal-board-p a c)
"
  (if (or (equal-by-rotation a b)
	  (equal-by-rotation (transpose a) b))
      t
    nil))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          remove-finished-games
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun remove-finished-games(forms)
  "
 The fives are the first forms that can have completed games in them. These I
 want to remove.
"
  (let ((finished (finished-games forms)))
    (select forms (where finished nil))))
    

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                check-for-equal-boards
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun check-for-equal-boards(boards)
     (let* (
	    (some-left (copy-list boards))
	    (single (first boards))
	    (keepers (list single))
	    )
       (while
	   some-left
	 (setq
	  some-left
	  (select
	   some-left
	   (where
	    (mapcar (lambda(x) (equal-board-p single x)) some-left)
	    nil
	    )
	   )
	  )
	 (if some-left
	     (setq
	      single (first some-left)
	      keepers (append keepers (list single))))
	 )
       keepers
       )
     )


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                add-new-possibilities
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun add-new-possibilities(forms &optional (val -1))
  (apply
   #'append
   (mapcar
    (lambda(x)
      (let* (
	     (element-list (combine x))
	     (zeros (where element-list 0))
	     )
	(mapcar
	 (lambda(y) (let ((l (copy-list element-list)))
		      (setf (elt l y) val)
		      (matrix '(3 3) l)))
	 zeros
	 )
	)
      )
    forms
    )
   )
  )


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                add-future-possibilities
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun add-future-possibilities(forms)
  (apply
   #'append
   (mapcar
    (lambda(x)
      (let* (
	     (element-list (combine x))
	     (where-zero (where element-list 0))
	     (val (^ -1 (1+ (length where-zero))))
	     (matrices
	      (mapcar
	       (lambda(i)
		 (let* (
			(temp (copy-list element-list))
			(crap (setf (elt temp i) val))
			)
		   (matrix '(3 3) temp)
		   )
		 )
	       where-zero
	       )
	      )
	     (i -1)
	     (keepers
	      (mapcar 
	       (lambda(x)
		 (setf
		  (elt element-list (incf i))
		  (not (v-or (mapcar
			      (lambda(y) (equal-board-p (first x) y))
			      (rest x))))))
	       (mapcon 'list matrices)))
	     )
	(list (select where-zero (where element-list t)))
	)
      )
    forms
    )))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                finished-games
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun finished-games(forms)
  "
;; Examples:
 (finished-games (list (ones 3 3)))
 (finished-games (list (id 3)))
 (finished-games (list (- (id 3))))
 (finished-games (list (rotate-clock-wise-matrix (id 3) 1)))
 (finished-games (list (zeros 3 3)))
"
  (mapcar
   (lambda(x)
     (let* (
	    (row-sums (mapcar #'abs (mapcar #'sum (row-lists x))))
	    (col-sums (mapcar #'abs (mapcar #'sum (column-lists x))))
	    (diag-sums (abs (sum (diagonal x))))
	    (diag-sums2 (abs (sum (diagonal (rotate-clock-wise-matrix x 1)))))
	    )
       (if (search '(3) (combine row-sums col-sums diag-sums diag-sums2))
	   t
	 nil
	 )))
   forms
   )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                add-detract
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun add-detract(moves &key (fn #'evenp) (augment 1.1) (decrement .9))
  "
  add-detract(moves &key (fn #'evenp) (augment 1.1) (decrement .9))
is the correction step: it simply strengthens a node if its use in a certain
game led to victory, or weakens it if it led to defeat.

	fn:		a function (default
	augment:	the factor by which to strengthen the connection
	decrement:	the factor by which to weaken the connection

;; Examples:
 (elt (elt matchboxes 0) 1)
 (elt (elt matchboxes 1) 8)
 (add-detract '((1 8) (8 4)))
 (elt (elt matchboxes 0) 1)
 (elt (elt matchboxes 1) 8)

 (elt (elt matchboxes 0) 1)
 (elt (elt matchboxes 1) 8)
 (add-detract '((1 8) (8 4)) :fn #'evenp)
 (elt (elt matchboxes 0) 1)
 (elt (elt matchboxes 1) 8)
 (add-detract '((1 8) (8 4)) :fn #'oddp)
 (elt (elt matchboxes 0) 1)
 (elt (elt matchboxes 1) 8)
"
  (dotimes
   (i (length moves))
   ;; Only affects the even numbered weights: the others are constant, leading
   ;; to random play.
   (if (evenp i)
       (let* (
	      (move (elt moves i))
	      (board (first move))
	      (choice (second move))
	      (matchbox (elt (elt matchboxes i) board))
	      
	      (choices (first matchbox))
	      (probs (second matchbox))
	      
	      ;; choice is now the element of matchbox that will change:
	      (choice (first (where choices choice)))
	      
	      (crap
	       (setf (elt probs choice) 
		     (* (if (funcall fn i) ;; "add strength"
			    augment
			  decrement) (elt probs choice)))
	       )
	      )
	 (setf (elt (elt matchboxes i) board)
	       (list choices (/ probs (sum probs))))
	 )
     )
   )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                make-choice
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun make-choice(options interactive)
  "
;; Examples:
 (make-choice '((3 4 6 7 8) (1 1 1 1 1)) nil)
 (make-choice '((3 4 6 7 8) (1 1 1 1 1)) t)
"
  (if interactive
      (let (
	    (moves (1+ (first options)))
	    reply
	    )
	(tagbody
	 move
	 (format t "~%Your move: the choices are~{ ~d~}: " moves)
	 (setq reply (read))
	 (if (not (search (list reply) moves))
	     (let ()
	       (format t "~%Illegal choice (~s): try again!~%" reply)
	       (go move)
	       )
	   )
	 )
	(1- reply)
	)
    (let* (
	   (moves (first options))
	   (probs (second options))
	   (probs (/ (accumulate #'+ probs) (sum probs)))
	   (rand (first (rand.u 1)))
	   )
      (elt moves (first (where probs rand #'>=)))
      )
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                train
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun train(&key (print t) (interactive nil))
  "
 Interactive: 1 or 2, if you wish to play first or second; nil if you want the
computer to train itself.

;; Examples
 (train :interactive 1)
"
  (let* (
	 (val -1)
	 (interactive (if interactive (1- interactive)))
	 (openings (first moves))
	 (openings-options (first matchboxes))
	 (choice
          (if interactive
	      (if (= interactive 0)
		  (let ()
		    (setq interactive 1)
		    (format t "~%Corner (1), Side (2), or Center (3)?~%")
		    (make-choice '((0 1 2) nil) t)
		    )
		(let ()
		  (setq interactive 0)
		  (first (sample (iseq 3) 1)))
		)
	    (first (sample (iseq 3) 1)))
	  )
	 (game (matrix '(3 3) (combine (elt openings choice))))
	 actual-moves
	 finished
	 (i 0)
	 )
    ;; The first move has been made. Now for the rest:
    (mapcar
     (lambda(openings openings-options)
       (if (not finished)
	   (let* (
		  ;; which board in the next step is equalp to the game??
		  (which (first (where openings game #'equal-board-p)))
		  (newgame (elt openings which))
		  (options (elt openings-options which))
		  (crap (if print (mprint newgame)))
		  (option 
		   (if interactive
		       (if (= interactive 0)
			   (let ()
			     (setq interactive 1)
			     (make-choice options interactive)
			     )
			 (let ()
			   (setq interactive 0)
			   (make-choice options nil)
			   )
			 )
		     (make-choice options nil)
		     )
		   )
		  crap
		  )
	     (incf i)
	     (setq game (combine newgame)
		   crap (setf (elt game option) val)
		   game (matrix '(3 3) game)
		   actual-moves (append actual-moves (list (list which option)))
		   val (- val))
	     (if (first (finished-games (list game)))
		 ;; we're done! the game was won/lost, so we need to correct
		 ;; folks.
		 (let ()
		   (setq finished i)
		   (if print (format t "~%We have a winner!~%"))
		   )
	       )
	     )
	 )
       )
     moves matchboxes
     )
    ;; If finished is even, then the first player won; otherwise the second; we
    ;; add to the first, and detract from the second; otherwise, vice versa!
    ;; 	If it's a draw, then we add strength to both.
    (if print (let () (mprint game) (print actual-moves)))
    (if finished
	(if (evenp finished)
	    ;; add to odds, take from others (i.e., evens):
	    (add-detract actual-moves :fn #'oddp)
	  ;; switch this to add to evens, take from others (i.e. evens):
	  (add-detract actual-moves :fn #'evenp)
	  )
      ;; switch this to add if a number (i.e., all will be strengthened, as
      ;; it's a tie):
      (add-detract actual-moves :fn #'numberp)
      )
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                Compile for speed:
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(compile 'rotate-clock-wise-matrix)
(compile 'matrix-equalp)
(compile 'equal-by-rotation)
(compile 'equal-board-p)
(compile 'check-for-equal-boards)
(compile 'add-new-possibilities)
(compile 'add-future-possibilities)
(compile 'finished-games)
(compile 'remove-finished-games)
(compile 'add-detract)
(compile 'make-choice)
(compile 'TRAIN)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                         All possible moves:
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(def sheer-numbers
     (list
      (binomial-coef 9 1)
      (* (binomial-coef 9 1) (binomial-coef 8 1))
      (* (binomial-coef 9 2) (binomial-coef 7 1))
      (* (binomial-coef 9 2) (binomial-coef 7 2))
      (* (binomial-coef 9 3) (binomial-coef 6 2))
      (* (binomial-coef 9 3) (binomial-coef 6 3))
      (* (binomial-coef 9 4) (binomial-coef 5 3))
      (* (binomial-coef 9 4) (binomial-coef 5 4))
      ))
;; 5919 total!

;; Now we want to form equivalence classes:
(def singletons
     (mapcar (lambda(x) (matrix '(3 3) x)) (column-lists (id 9))))

#|
;;	Only need to do this once: now it's all saved to tic-tac2:

(def special-singletons (check-for-equal-boards singletons))
(mapcar #'mprint special-singletons)
;; 3 of 'em.
     
(def doubletons (add-new-possibilities special-singletons))
(def special-doubletons (check-for-equal-boards (first doubletons)))
(mapcar #'mprint special-doubletons)
;; 12 of them:
(size special-doubletons)

(def tripletons (add-new-possibilities special-doubletons 1))
(def special-tripletons (check-for-equal-boards tripletons))
(size special-tripletons)

(def quadritons (add-new-possibilities special-tripletons))
(time (def special-quadritons (check-for-equal-boards quadritons)))
(size special-quadritons)

(def quintitons (add-new-possibilities special-quadritons 1))
(size quintitons)
(time (def special-quintitons (check-for-equal-boards quintitons)))
(size special-quintitons)
(def special-quintitons (remove-finished-games special-quintitons))
(size special-quintitons)

(def sextitons (add-new-possibilities special-quintitons))
(size sextitons)
(time (def special-sextitons (check-for-equal-boards sextitons)))
(size special-sextitons)
(def special-sextitons (remove-finished-games special-sextitons))
(size special-sextitons)

(def septitons (add-new-possibilities special-sextitons 1))
(size septitons)
(time (def special-septitons (check-for-equal-boards septitons)))
(size special-septitons)
(def special-septitons (remove-finished-games special-septitons))
(size special-septitons)

(def octitons (add-new-possibilities special-septitons))
(size octitons)
(time (def special-octitons (check-for-equal-boards octitons)))
(size special-octitons)
(def special-octitons (remove-finished-games special-octitons))
(size special-octitons)

;;	Okay, once I get all these squared away, there are the total number
;; of moves to consider. I create associated lists of their prospective
;; moves, from which we will sample to get the "color" of the MENACE program.

(def moves
     (list
      special-singletons
      special-doubletons
      special-tripletons
      special-quadritons
      special-quintitons
      special-sextitons
      special-septitons
      special-octitons
      ))
(sum (mapcar #'size moves))
;; 626
;; Conclusion: there are 626 distinct non-terminal boards in tic-tac-toe,
;; distributed as follows (mapcar #'size moves):
;; 
;; (3 12 38 108 153 183 95 34)
;;
(mapcar #'sum (transpose (split-list '(3 12 38 108 153 183 95 34) 2)))
;; (289 337)
;; I get it now: they only followed the opening moves: the machine was always
;; to go first.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; According to the book, there should be only 288 moves - I count 289. Let me
;; just check this directly:
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(def defense (smoosh (first (transpose (split-list moves 2)))))
(def defense (remove-finished-games defense))
;; still 289:
(def special-defense (check-for-equal-boards defense))
(size special-defense)
;; still 289. Hmmmm.....

(def prospective-moves (mapcar #'add-future-possibilities moves))
(sum (mapcar #'size prospective-moves))
|#

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                        Store game info in tic-tac2:
;;
;; (savevar '(moves prospective-moves) "tic-tac2")
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(load "tic-tac2")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                Ready to learn!
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; fill matchboxes with equal weights:
;;
(def matchboxes
     (mapcar
      (lambda(x)
	(mapcar
	 (lambda(y) (list y (repeat 1 (length y))))
	 x
	 )
	)
      prospective-moves
      )
     )
(sum (mapcar #'size matchboxes))

;;
;; When I finally get these trained, I'm going to want to:
;;
;; (savevar '(matchboxes) "tic-tac1")
;; 
;; (while t (pprint (first matchboxes)) (train :print nil :interactive nil))
;; (while t (train :print nil :interactive nil))
;;
(load "tic-tac1")
(while t (print (first (first matchboxes))) (train :print nil :interactive nil))
