#|

			Basic Differentiation Rules

	Based on Chapter 3 of Luke Tierney's "Xlisp-Stat". Can you teach your
language differentiation this easily?

	That's been enhanced in a couple of ways: I've

1) added lots of functions;
2) vectorized the +, -, *, and / rules;
3) added functions derfunc and derivfunc which return the derivatives; and
4) included capability for handling user defined functions: that is, one can
   create a function and have the deriv function handle it by its lambda
   expression.

						Andy Long, Fall 1995
						Revised, Fall 1997
|#

;;-------------------------- Recognizing types ---------------------------
;;
;; First we define methods by which we recognize various types or functions:
;;
(defun constantp (e) (numberp e))	;; is it a constant?
(defun variablep (e) (symbolp e))	;; is it a symbol, i.e. variable?

;;
;; binary functions:
;;
(defun sump (e)				;; is it a sum?
  (and (listp e) (= (length e) 3) (eq (first e) '+)))
(defun diffp (e)				;; is it a difference?
  (and (listp e) (= (length e) 3) (eq (first e) '-)))
(defun quotientp (e)			;; is it a quotient?
  (and (listp e) (= (length e) 3) (eq (first e) '/)))
(defun productp (e)			;; is it a product?
  (and (listp e) (= (length e) 3) (eq (first e) '*)))
;;------------------------------------------------------------------------
;; Big idea: replace the criterion that the length be 3 (that makes the
;; operation binary, but why should it be when all these operators are
;; vectorized? Notice that "-" is special, as it is also a unary operator.
(defun sump (e)				;; is it a sum?
  (and (listp e) (eq (first e) '+)))
(defun diffp (e)				;; is it a difference?
  (and (listp e) (> (length e) 2) (eq (first e) '-))) ;; so it won't try unary
(defun quotientp (e)			;; is it a quotient?
  (and (listp e) (eq (first e) '/)))
(defun productp (e)			;; is it a product?
  (and (listp e) (eq (first e) '*)))
;;------------------------------------------------------------------------

(defun powerp (e)			;; is it a power?
  (and (listp e) (eq (first e) '^)))
(defun lonelylistp (e)			;; is it a "lonely list", length 1?
  (and (listp e) (= (length e) 1)))

;;
;; unary functions:
;;
(defun unaryp (e)		;; unary functions have a single argument
  (and (listp e) (= (length e) 2)))
(defun unaryp-and-closure (e)
  (let ((first-one (first `,e)))
    (eq 'closure (eval `(type-of #',first-one))))
  )
;;
;; ternary functions:
;;
(defun ifp (e)				;; is it an if?
  (and (listp e) (= (length e) 4) (eq (first e) 'if)))
(defun sum-sump (e)			;; is it a sum?
  (and (listp e) (= (length e) 2) (eq (first e) 'sum)))

;;-------------------------- Some special functions: ---------------------

(defun same-variable-p (v1 v2)		;; if 2 variables, are they the same?
  (and (variablep v1) (variablep v2) (eq v1 v2)))
(defun addend (e) (second e))		;; the first half of a sum
(defun multiplier (e) (second e))	;; the first half of a product
;;------------------------------------------------------------------------
;; This is my part of my effort to get away from binary uses of these (since
;; they're vectorized).
(defun augend (e)		;; the rest of a sum
  (if (= 2 (length (rest e))) (third e) (append '(+) (rest (rest e))))
  )
(defun multiplicand (e) 	;; the rest of a product
  (if (= 2 (length (rest e))) (third e) (append '(*) (rest (rest e))))
  )
;; note that the above are used, and work, with subtraction and division as
;; well, as the vectorized forms of these are defined as (- a all-the-rest),
;; etc.

;;------------------------------ making expressions ----------------------

;; Then we create some functions to write these functions:

(defun make-sum (a1 a2)			;; make a sum
  (cond
   ((and (numberp a1) (numberp a2)) (+ a1 a2))	;; if two numbers, just add 'em
   ((numberp a1) (if (= a1 0) a2 (list '+ a1 a2)))	;; if zeros, simplify
   ((numberp a2) (if (= a2 0) a1 (list '+ a1 a2)))	;; if zeros, simplify
   ((equal a1 a2) (list '* 2 a2))			;; if same, twice one:
   (t (list '+ a1 a2)))
  )

(defun make-diff (a1 a2) 		;; creates a difference list
  (cond
   ((and (numberp a1) (numberp a2)) (- a1 a2))	;; if two numbers, just diff 'em 
   ((numberp a1) (if (= a1 0) (list '- a2) (list '- a1 a2)))	;; if zeros, simplify
   ((numberp a2) (if (= a2 0) a1 (list '- a1 a2)))	;; if zeros, simplify
   ((equal a1 a2) 0)					;; if same, 0
   (t (list '- a1 a2)))
  )

(defun make-product (m1 m2)		;; make a product
  (cond
   ((and (numberp m1) (numberp m2)) (* m1 m2))	;; if both numbers, just mult
   ((numberp m1)			;; if one a number, consider specials
    (cond ((= m1 0) 0)
	  ((= m1 1) m2)
	  ((= m1 -1) (list '- m2))
	  (t (list '* m1 m2))))
   ((numberp m2)			;; if one a number, consider specials
    (cond ((= m2 0) 0)
	  ((= m2 1) m1)
	  (t (list '* m1 m2))))
   ((equal m1 m2) (list '^ m1 2))	;; if same, square it
   (t (list '* m1 m2))
   )
  )

(defun make-quotient (m1 m2)		;; make a quotient
  (cond
   ((and (numberp m1) (numberp m2)) (/ m1 m2))	;; if both numbers, just mult
   ((numberp m1)			;; if one a number, consider specials
    (cond ((= m1 0) 0)
	  (t (list '/ m1 m2))))
   ((numberp m2)			;; if one a number, consider specials
    (cond ((= m2 0) not-a-number)
	  ((= m2 1) m1)
	  (t (list '/ m1 m2))))
   ((equal m1 m2) 1)	;; if same, 1
   (t (list '/ m1 m2))
   )
  )

;;---------new and improved make-square---------
;;(defun make-square (a) (list '^ a 2))	;; creates a square list
(defun make-square (a)	;; creates a square list
  (cond
   ((numberp a) (^ a 2))
   (t (list '^ a 2))
   )
  )

(defun make-power (a1 a2)
  (cond
   ((and (numberp a1) (numberp a2)) (^ a1 a2))	;; if two numbers, just ^ em
   ((numberp a1)
    (cond
     ((= a1 0) 0)	;; if a1=zero, 0
     ((= a1 1) 1)	;; if a1=1, 1
     (t (list '^ a1 a2))
     )
    )
   ((numberp a2)
    (cond 
     ((= a2 1.0) a1)	;; if power=1, a1
     ((= a2 0.0) 1)	;; if power=0, 1
     (t (list '^ a1 a2))
     )
    )
   (t (list '^ a1 a2))
   )
  )

(defun make-unary (fcn arg) (list fcn arg))	;; creates a unary func.
(defun make-binary (oper x y) (list oper x y))	;; creates an equation
(defun make-if (oper x y) (list 'if oper x y))	;; creates an if
(defun make-sum-sum (oper) (list 'sum oper))	;; creates a sum-sum

;;---------------------------- Unary stuff ------------------------------

;;	Some Unary function derivatives, and derivative tools:


(def *derivatives* nil)

(defun unary-function (e) (first e))	;; it's which function?
(defun unary-argument (e) (second e))	;; and what's its argument?

(defun apply-unary-rule (entry arg)
  (funcall (second entry) arg))

(defun get-unary-rule (f)
  (let ((rule (assoc f *derivatives*)))
    (if rule
	rule
      (error
       (format
	nil 
	"In get-unary-rule, file deriv.lsp:
Sorry, can't yet differentiate this expression: ~s."
	f
	)
       )
      )
    )
  )

(defun make-unary-deriv (fcn arg)
  (apply-unary-rule (get-unary-rule fcn) arg))

(defun add-unary-rule (f rule)
  (setf *derivatives* (cons (list f rule) *derivatives*)))


;;-------------- A whole bunch of unary functions! Just add 'em in! -------

;; -:
(add-unary-rule '- #'(lambda (x) -1))

;; exp:
(add-unary-rule 'exp #'(lambda (x) (make-unary 'exp x)))

;; sin:
(add-unary-rule 'sin #'(lambda (x) (make-unary 'cos x)))

;; cos:
(add-unary-rule 'cos #'(lambda (x) (make-product -1 (make-unary 'sin x))))

;; tan:
(add-unary-rule 'tan #'(lambda (x)
			 (make-quotient 1 (make-power (make-unary 'cos x) 2))))


;; cot:
(add-unary-rule 'cot #'(lambda (x)
			 (make-quotient -1 (make-power (make-unary 'sin x) 2))))

;; asin:
(add-unary-rule
 'asin #'(lambda (x)
	   (make-quotient 1 (make-unary 'sqrt (make-diff 1 (make-power x 2))))))

;; acos:
(add-unary-rule
 'acos #'(lambda (x)
	   (make-quotient -1 (make-unary 'sqrt (make-diff 1 (make-power x 2))))))

;; atan:
(add-unary-rule
 'atan #'(lambda (x)
	   (make-quotient 1 (make-sum 1 (make-power x 2)))))

;; sinh:
(add-unary-rule 'sinh #'(lambda (x) (make-unary 'cosh x)))

;; cosh:
(add-unary-rule 'cosh #'(lambda (x) (make-unary 'sinh x)))

;; log:
(add-unary-rule 'log #'(lambda (x) (make-quotient 1 x)))

;; sqrt:
(add-unary-rule
 'sqrt #'(lambda (x) 
	   (make-quotient 1 
			  (make-product 2 (make-unary 'sqrt x)))))

;; abs:
(add-unary-rule 'abs #'(lambda (x) (make-unary 'signum x)))

;; signum:
(add-unary-rule 'signum #'(lambda (x) 0))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                Deriv
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun deriv (exp var)
  "
;; DESCRIPTION: function (deriv expression variable)

where

	expression	- the expression to be differentiated;
	variable	- the variable with which to differentiate.

	A general purpose differentiator which returns the expression, or
formula.

;; SEE ALSO: der, derfunc, derivfunc, lambda-deriv, diff, diff-n

;; Examples:
 (deriv '(exp x) 'x)
 (deriv '(sin x) 'x)
 (deriv '(exp (sin x)) 'x)
 (deriv '(^ x x) 'x)

 (deriv '(+ x 3) 'x)
 (deriv '(x) 'x)

 (deriv '(* x y) 'x)
 (deriv '(* (* x y) (+ x 3)) 'x)

 (deriv '(- x x) 'x)
 (deriv '(/ 1 x) 'x)

 (deriv '(cos x) 'x)
 (deriv '(log x) 'x)
 (deriv '(log (/ (sin x) 3)) 'x)
 (deriv '(tan x) 'x)
 (deriv '(abs x) 'x)
 (deriv '(abs (log x)) 'x)
 (deriv '(exp (sin x)) 'x)

 (deriv '(^ x x) 'x)
 (deriv '(^ x 2) 'x)

 (defun f (x) (sin x))
 (deriv #'f 'x)

 (deriv '(+ x (^ x 2) (sin x)) 'x)
"
  (cond
   ((ifp exp)
    (make-if (second exp) (deriv (third exp) var) (deriv (fourth exp) var)))
   ((sum-sump exp) (make-sum-sum (deriv (second exp) var)))
   ((constantp exp) 0)
   ((variablep exp) (if (same-variable-p exp var) 1 0))
   ((sump exp)
    (make-sum (deriv (addend exp) var)
	      (deriv (augend exp) var)))
   ((diffp exp)
    (make-diff (deriv (addend exp) var)
	      (deriv (augend exp) var)))
   ((productp exp)
    (make-sum (make-product (multiplier exp)
			    (deriv (multiplicand exp) var))
	      (make-product (deriv (multiplier exp) var)
			    (multiplicand exp))))
   ((quotientp exp)
    (make-quotient
     (make-diff (make-product (deriv (multiplier exp) var) (multiplicand exp))
		(make-product (multiplier exp) (deriv (multiplicand exp) var))
		)
     (make-square (multiplicand exp))
     ))
   ((powerp exp)
    (if (numberp (third exp))
	(make-product (deriv (second exp) var)
		      (make-product (third exp)
				    (make-power (second exp) (1- (third exp))))
		      )
      (deriv (list 'exp (make-product (third exp)
				      (make-unary 'log (second exp)))
		   )
	     var)
      )
    )
   ((unaryp exp)	;; the unary rule incorporates the chain rule!
    (if (unaryp-and-closure exp)
	(let* (
	       (first-one (first `,exp))
	       (var1
		(first
		 (second (eval `(get-lambda-expression #',first-one)))))
	       (defn (third (eval `(get-lambda-expression #',first-one))))
	       (varnew (second exp))
	       )
	  (setq exp (eval `(subst ',varnew ',var1 ',defn)))
	  (deriv exp var)
	  )
      (make-product (make-unary-deriv
		     (unary-function exp)
		     (unary-argument exp)
		     )
		    (deriv (unary-argument exp) var)
		    )
      )
    )
   ((lonelylistp exp) (deriv (first exp) var))
   #|
   ((unaryp exp)	;; the unary rule incorporates the chain rule!
    (make-product (make-unary-deriv (unary-function exp)
				    (unary-argument exp))
		  (deriv (unary-argument exp) var)))
|#
   ((functionp exp) (deriv (third (get-lambda-expression exp)) var))
   (t
    (error "I don't know a rule for differentiating this expression.")
    )
   )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;				derivfunc
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun derivfunc (thistle that var)
  "
;;
;; Description for DERIVFUNC:
;;
;; 	(derivfunc function-name expression variable)

	Derivfunc takes care of a situation that derfunc can't yet handle: it
will differentiate an arbitrary, user-defined function to return the derivative
function. 

;; Arguments:

	function-name	- the name of the derivative function (none if nil)
	expression	- the expression or formula which defines the function
	variable	- the variable with which one takes the derivative

;; See also: deriv, diff, and lambda-deriv, which differentiates without
;; specifying the function name.

;; Examples:

 (derivfunc 'fprime '(cos x) 'x)
 (plot fprime 0 4pi)

 (defun f(x) (sin x))
 (deriv #'f 'x)
 (plot (derivfunc nil #'f 'x) 0 4pi)

 (defun f(x y) (sin (* x y)))
 (derivfunc 'fprime #'f 'x)
 (gle fprime)

 (defun f(u v) (sin (* u v)))
 (derivfunc 'fprime #'f 'v)
 (gle fprime)
 (plot fprime)
" 
  (if thistle
      (let (
	    (newvar (list var))
	    bing
	    )
	(cond
	 ((eq (type-of that) 'closure)
	  (setq newvar (second (get-lambda-expression that)))
	  (setq bing (deriv that var)))
	 ((or (eq (type-of that) 'subr) (eq (type-of that) 'byte-code-closure))
	  (format t "~s: can't handle those, sorry: try der, it can!"
		  (type-of that))
	  )
	 ((eq (type-of that) 'cons)
	  (setq bing (eval `(der ,that ,var))))
	 (t
	  (print "What is that thing? Might get lucky....")
	  )
	 )
	(setq bing (eval `(defun ,thistle ,newvar ,bing)))
	`(function ,bing)
	)
    (lambda-deriv that var))
  )
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;				lambda-deriv
;;
;; Description for LAMBDA-DERIV:
;;
(defun lambda-deriv (that var)
  "
;; Documentation:

	Lambda-Deriv takes care of a situation that derfunc can't yet handle: it
will differentiate an arbitrary, user-defined function to return the derivative
function. 

;; Arguments:

	expression	- the expression or formula which defines the function
	variable	- the variable with which one takes the derivative

;; See also: derivfunc (this is a helper function for derivfunc)

;; Examples:
 (plot (lambda-deriv '(cos x) 'x) 0 4pi)

 (defun f(x) (sin x))
 (plot (lambda-deriv #'f 'x) 0 4pi)

 (defun f(x y) (sin (* x y)))
 (get-lambda-expression (lambda-deriv #'f 'x))

 (defun f(u v) (sin (* u v)))
 (get-lambda-expression (lambda-deriv #'f 'v))
 (plot (lambda-deriv #'f 'v))
" 
  (let (
	(newvar (list var))
	bing
	)
    (cond
     ((eq (type-of that) 'closure)
      (setq newvar (second (get-lambda-expression that)))
      (setq bing (deriv that var)))
     ((or (eq (type-of that) 'subr) (eq (type-of that) 'byte-code-closure))
      (format t "~s: can't handle those, sorry: try der, it can!" (type-of that))
      )
     ((eq (type-of that) 'cons)
      (setq bing (eval `(der ,that ,var))))
     (t
      (print "What is that thing? Might get lucky....")
      )
     )
    (setq bing (eval `(lambda ,newvar ,bing)))
    `,bing
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;				der
;;
;;	The macro doesn't require the quote, which is a blessing.
;;
(defmacro der (a b)
  "
;; Description: macro (der expression variable)

	A general purpose differentiator which returns the expression, or
formula. Just calls deriv.

;; SEE ALSO: derfunc, deriv, derivfunc, lambda-deriv, diff, diff-n

;; Examples:
 (der (^ x 2) x)
 (der (* x 2) x)
 (func f(x) x^2)
 (der f x)
"
  (if (symbolp a) `(deriv #',a ',b)
    `(deriv ',a ',b)
     )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; diff Doc:    Creation Date://        Author: Andy Long
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmacro diff (x y &optional z)
  "
Description:

	diff is the general purpose differentiator.

Notes: Diff back-parses the output of der to get a user-friendly form of the
	derivatives. It may be messy!

See Also: der, deriv, back-parse

;; Examples:

 (func f(x y) x ^ 2 + sin(y))
 (diff f x)
 (diff f y)
 (diff f x fx)
 (diff f y fy)

 (func f(y) sin(y))
 (diff f y)
 (diff f y fp)
 (what fp)

 (diff (func f(x) (x ^ 2)) x)
 (diff (lambda (x) (^ x 2)) x)
"
  (let ((bing (eval `(turn-it-into-a-function ,x))))
    (if z `(derivfunc ',z ,bing ',y)
      (let ((bing (eval `(deriv ,bing ',y))))
	(back-parse `',bing)))))
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; diff-n Doc:    Creation Date:5/1/96        Author: Andy Long
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmacro diff-n (x y n &optional z)
  "
Description: macro (diff-n function-name variable num-derivs &optional z)

	diff-n is a higher-order general purpose differentiator.

Notes: Diff-n back-parses the output of der to get a user-friendly form of the
	derivatives. It may, however, be messy!

See Also: der, deriv, back-parse, diff

;; Examples:
 (func f(x) x ^ 5)
 (diff-n f x 1)
 (diff-n f x 3)
 (diff-n f x 5)
 (diff-n f x 6)

 (func f(x y) x ^ 2 + sin(y))
 (diff-n f x 2)
 (diff-n f y 2)
 (diff-n f y 2 fyy) ;; second partial of f with respect to y
 (what fyy)
"
  (let (
	(n-1 (1- n))
	(bing (eval `(turn-it-into-a-function ,x)))
	)
    (if (= 1 n)
	(if z `(derivfunc ',z ,bing ',y)
	  (let ((bing (eval `(deriv ,bing ',y))))
	    (back-parse `',bing))
	  )
      `(diff-n (lambda-deriv ,bing ',y) ,y ,n-1 ,z)
      )))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;				derfunc
;;
;;	Now I'll try to make it create a derivative function:
;;
(defmacro derfunc (thistle that var)
  "
;; Description: macro (derfunc fp f var)

	Computes the derivative of a function or expression f, with respect to
the variable var, and stores it in fp. It also returns the function, so that
you can use

	(plot (derfunc fp f var))

for example, or

	(setq my-func (derfunc fp f var))

;; See also: deriv, der

;; Notes:

;; Examples:

 (derfunc try-me (^ x 3) x)
 (plot-function #'try-me (- pi) pi)

 (derfunc try-me (sin x) x)
 (plot-function #'try-me (- pi) pi)

 (derfunc try-me (cos x) x)
 (plot-function #'try-me (- pi) pi)

 (derfunc try-me (* (^ x 2) yyy) x)
 (def yyy 1)
 (plot-function #'try-me (- pi) pi)
 (undef 'yyy)

 (derfunc fff (sqrt (^ x 2)) x)
 (def points (fff (rseq 1 100 100)))
 (def myplot (plot-lines (rseq 1 100 100) points))
 (send myplot :range 1 0 2)

 (func f (x)  x ^ 2)
 (derfunc fp f x)
 (new-plot)
 (plot #'fp)

 (plot (derfunc fp sin x))

;; This provides some nice 3-d pictures to go along with derfunc:

 (defun f (X Y) (* (* 2 (SIN X)) Y))
 (derfunc fpx f x)
 (derfunc fpy f y)

 (spin-points 10)
 (spin-domain -pi pi -pi pi)
 (spin-depth nil)
 (title \"f\")
 (plot f)
 (title \"fx\")
 (plot fpx)
 (title \"fy\")
 (plot fpy)

 (defun f (X) (* (* 2 (SIN X)) (COS X)))
 (derfunc fpx f x)

 (title \"f\")
 (plot f)
 (title \"fx\")
 (plot fpx)

 (derfunc fpx (^ x 2) x)

 (title \"f\")
 (plot f)
 (title \"fx\")
 (plot fpx)
"
  (let (
	(newvar (list `,var))
	bing
	)
    (cond
     ((symbolp that)
      (cond
       ((eq (eval `(type-of #',that)) 'SUBR)
	(def bing (eval `(deriv #'(lambda (,var) (,that ,var)) ',var)))
	)
       ((eq (eval `(type-of #',that)) 'BYTE-CODE-CLOSURE)
	(def bing (eval `(deriv #'(lambda (,var) (,that ,var)) ',var)))
	)
       (t
	(setq newvar (second (eval `(get-lambda-expression #',that))))
	(def bing (eval `(deriv #',that ',var)))
	)))
     (t
      (def bing (eval `(der ,that ,var)))
      )
     )
    (def bing (eval `(defun ,thistle ,newvar ,bing)))
    `(function ,bing)
    )
  )
#|
;; The following is very strange: should be inverses, so...fff = 1
 (derfunc fff (sqrt (^ x 2)) x)
 (plot-function #'fff 1 100)
;; and this one explains it:
 (derfunc fff (- x (sqrt (^ x 2))) x)
 (plot-function #'fff 1 100)
;; says that maybe one should avoid plot-function!
|#
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;
;;			Numerical derivatives:
;;
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; num-diff Doc:    Creation Date:3/23/96        Author: Andy Long
;;
;;
(defun num-diff(x &optional (xdiff 1))
  "
;; Description/Arguments:

       (num-diff x &optional (xdiff 1))
  
	Compute a numerical derivative for a list of equally-spaced values x.
Num-diff uses a centered difference in the middle, and forward and backward
differences at the ends. 

;; Notes: if you do not specify the interval-spacing, then a value of 1 will be
;; assumed.

;; See Also:

;; Examples:
 (new-plot)
 (title \"numerical differentiation - f(x)=x^2\")
 (func f(x) x^2)
 (plot (rseq 0 3 101) (mapcar #'f (rseq 0 3 101)))
 (title \"Should look like g(x)=2x:\")
 (plot (rseq 0 3 101) (num-diff (mapcar #'f (rseq 0 3 101)) .03))

 (title \"numerical differentiation - normal cdf\")
 (plot (rseq 0 3 101) (num-diff (mapcar #'normal-cdf (rseq 0 3 101)) .03))
 (title \"should look like normal density:\")
 (plot (rseq 0 3 101) (mapcar #'normal-dens (rseq 0 3 101)))
"
  (let* (
	 (rx (rest (rest x)))
	 (bx (butlast (butlast x)))
	 (center (/ (- rx bx) 2))
	 (lend (- (second x) (first x)))
	 (rend (- (last x) (second (reverse x))))
	 (numdiff (combine lend center rend))
	 )
    (/ numdiff xdiff)
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 numerical-derivative
;;
(defmacro numerical-derivative
  (f
   &key
   (delta 1e-6)
   (diff-list '(1)))
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          numerical-derivative Doc:
;;
;; Author: Andy Long
;; Created: Thu May 1 10:47:54 CDT 1997
;;
;; Description/Arguments:

       (numerical-derivative
	  (f
	   &key
	   (delta 1e-6)
	   (diff-list '(1)
	   )
  
	Given function f (by default of one variable) compute its numerical
derivative (by default its first derivative). 

;; Notes:

;; See Also:

;; Examples:
 (func f(x) x^2)
 (numerical-derivative f)
 (plot (numerical-derivative f))

 (func f(x) x^2)
 (numerical-derivative f)
"
  (let (
	(nargs (length diff-list))
	)
    (if (= nargs 1)
	(if (equalp diff-list '(1)) 
	    `(n2diff #',f)
	  (if (= (length diff-list) 1)
	      `(n2diff #',f :n (first diff-list))
	    )
	  )
      (print "not ready yet!")
      )
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 n2diff
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun n2diff(f &key fp (dx 1e-6) (n 1))
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; n2diff Doc:    Creation Date:3/23/96        Author: Andy Long
;;
;;                Modified: 5/1/97 - to do multiple derivatives, and to use a
;;                centered difference.
;;
;; Description/Arguments:

       (n2diff f &key fp (dx 1e-6) (n 1))
  
	Compute a numerical derivative function fp of a function f, using
a forward difference.

	n - the number of derivatives desired. If you want the second
	    derivative, n=2; etc. As this gets beyond 2, you can easily run
	    into poor conditioning. See the examples below.

	dx - the 'dx' in the approximation:

			f(x+dx)-f(x-dx)
		f' =    --------------
			      2*dx

;; See Also: ndiff, num-diff

;; Examples:
 (new-plot)
 (title \"numerical differentiation\")
 (func f(x) erf x)
 (n2diff #'f :fp 'fp)
 (plot f fp)

 (func g(v) sin v)
 (n2diff #'g :fp 'gp)
 (plot g gp)
 (plot g (n2diff #'g))
 (plot g (n2diff #'g :n 2))
 (plot g (n2diff #'g :n 3 :dx .01))
 ;; I can't get a fourth derivative to look good!
 (plot g (n2diff #'g :n 4 :dx .01))
"
  (let (
	(bing (if (= n 1) f (n2diff f :n (1- n))))
	(twice-dx (* 2 dx))
	)
    (eval
     (if fp
	 `(defun ,fp (x)
	    (let (
		  (val+ (+ x ,dx))
		  (val- (- x ,dx))
		  )
	      (/ (- (funcall ,bing val+) (funcall ,bing val-)) ,twice-dx)
	      )
	    )
       `(lambda(x)
	  (let (
		(val+ (+ x ,dx))
		(val- (- x ,dx))
		)
	    (/ (- (funcall ,bing val+) (funcall ,bing val-)) ,twice-dx)
	    )
	  )
       )
     )
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; ndiff Doc:    Creation Date:3/23/96        Author: Andy Long
;;
;;
(defmacro ndiff(&rest x)
  "
;; Description/Arguments:

       (ndiff f &optional fp (dx 1e-6) &key (n 1))
  
	Compute a numerical derivative function fp of a function f.
Uses a forward difference. 

	n - the number of derivatives desired. If you want the second
	    derivative, n=2; etc. As this gets beyond 2, you can easily run
	    into poor conditioning. See the examples below.

	dx - the 'dx' in the approximation:

			f(x+dx)-f(x-dx)
		f' =    --------------
			      2*dx

;; Notes:

;; See Also: n2diff, num-diff

;; Examples:
 (new-plot)
 (title \"numerical differentiation\")
 (func f(x) erf x)
 (ndiff f fp)
 (plot f fp)
 (plot f (ndiff f))

 (func g(v) sin v)
 (ndiff g gp)
 (plot g gp)
"
  (let* (
	 (n (first (catch-keys '(:n) x)))
	 (n (if n n 1))
	 (x (remove-keys '(:n) x))
	 (bing (first x))
	 (bong (if (second x) (second x)))
	 (bang (if (third x) (third x) 1e-6))
	 )
    (if (and (symbolp bing) (not (boundp bing))) ;; it's a function!
	(cond
	 ((eq (eval `(type-of #',bing)) 'SUBR)
	  (setq bing  (eval `(function (lambda (x) (,bing x)))))
	  )
	 ((eq (eval `(type-of #',bing)) 'BYTE-CODE-CLOSURE)
	  (setq bing (eval `(function (lambda (x) (,bing x)))))
	  )
	 (t
	  (setq bing `#',bing)
	  )
	 )
      )
    (if bong
	`(n2diff ,bing :fp ',bong :dx ,bang :n ,n)
      `(n2diff ,bing :n ,n)
      )      
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; turn-it-into-a-function Doc:    Creation Date:3/24/96       Author: Andy Long
;;
;;
(defmacro turn-it-into-a-function (x)
  "
;; Description/Arguments:

        (turn-it-into-a-function f)
  
	Takes lambda-expressions, explicit functions, closures, and symbols
that name functions and turns them into explicit functions.

;; Notes: Only univariate at this time.

;; See Also:

;; Examples:
 (turn-it-into-a-function sin)
 (turn-it-into-a-function (func f(x) sin x))
 (turn-it-into-a-function f)
 (plot (turn-it-into-a-function (lambda (x) (tan x))))
"
  (let ((bing `,x))
    (if (and (symbolp bing) (not (boundp bing))) ;; it's a function!
	(cond
	 ((eq (eval `(type-of #',bing)) 'SUBR)
	  (setq bing  (eval `(function (lambda (x) (,bing x)))))
	  )
	 ((eq (eval `(type-of #',bing)) 'BYTE-CODE-CLOSURE)
	  (setq bing (eval `(function (lambda (x) (,bing x)))))
	  )
	 (t
	  (setq bing `#',bing)
	  )
	 )
      )
    `,bing
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun gradient2 (f &optional gradf)
  "
;; Description: (gradient2 f &optional gradient-function-name)

	Compute the gradient function gradf of a function f of one or many
 variables.

;; Returns: either a function whose name is the symbol supplied as the second
   argument (gradient-function-name), or a lambda-expression (anonymous
   function) which contains the definition of the gradient.

;; Note: if a function is compiled, its gradient will not compute properly.

;; See Also: gradient.

;; Example: from multivariate calc, Harvard project:
;; #17, page 205
 (func g(x y) (x + 1) ^ 4 + (y - 1) ^ 4 + 1 / (x ^ 2 * y ^ 2 + 1))
 (gradient2 #'g 'ggrad)
 (ggrad 0 1)
 ;; alternatively, if the function ggrad need not be defined,
 (funcall (gradient2 #'g) 0 1)
 ;; or
 (def grad (gradient2 #'g))
 (funcall grad 0 1)

 ;; Also works for functions of several variables:
 (new-plot)
 (title \"gradient of u*w*sin(v)\")
 (gradient2 (func g(u v w) u*w*sin(v)) 'grad)
 (def us (rseq 0 2 100))
 (def vs (rseq -pi pi 100))
 (def ws us)
 (plot (grad us vs ws))
"
  (let* (
	 ;; What are the independent variables?
	 (newvar (second (get-lambda-expression f)))
	 
	 ;; Now get the partials with respect to these variables:
	 (df (mapcar (lambda(x) (lambda-deriv f x)) newvar))
	 )

    (eval
     (if (if gradf (and (symbolp gradf)))
	 ;; if a name is given as a symbol, define a function using defun:
	 `(defun ,gradf ,newvar
	    (mapcar 
	     (lambda(gradpar) 
	       (funcall gradpar ,@newvar)) (list ,@df)))
       ;; otherwise, return an anonymous function:
       `(lambda ,newvar
	  (mapcar #'(lambda(gradpar) (funcall gradpar ,@newvar)) (list ,@df))))
     )
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; gradient Doc:    Creation Date:3/24/96        Author: Andy Long
;;
;;
(defmacro gradient(&rest x)
  "
;; Description/Arguments: macro

       (gradient f &optional gradf)
  
	Compute the gradient function gradf of a function f of one or many
 variables.

;; Returns: the gradient function gradf, if gradf is given; otherwise an
;; anonymous function (lambda-expression).

;; See Also: gradient2

;; Examples:
 (new-plot)
 (title \"gradient function\")
 (func f(x y) (cos x) + (sin y))
 (gradient f gradf)
 (plot (gradf (rseq 0 pi 100) (rseq -pi pi 100)))

 (title \"gradient of u * sin v\")
 (func g(u v) u * sin v)
 (gradient g grad)
 (plot (grad (rseq 0 2 100) (rseq -pi pi 100)))

 ;; Three dimensions:
 (title \"gradient of u*w*sin(v)\")
 (func g(u v w) u*w*sin(v))
 (gradient g grad)
 (plot (grad (rseq 0 2 100) (rseq -pi pi 100) (rseq 0 2 100)))
"
  (let (
	(bing (first x))
	(bong (second x))
	)
    (setq bing (eval `(turn-it-into-a-function ,bing)))
    `(gradient2 ,bing ',bong)
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
