;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                slope-field2
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun slope-field2 (f xmin xmax ymin ymax
		      &optional (xsubdivisions 20) (ysubdivisions 20))
  "
;; Description: 

;; Examples:
 (title \"Slope field of -x/y\")
 (func f(x y) -x/y)
 (slope-field2 #'f -2 2 -2 2)

 (title \"Slope field of 2x\")
 (func f(x y) 2*x)
 (slope-field2 #'f -2 2 -2 2)
"
  (let* (
	 
;;	 (w (send scatterplot-proto :new 2 :title *title* :show nil))
	 (w (plotter))

	 ;; What is the independent variable?
	 (newvar (second (eval `(get-lambda-expression ,f))))
	 (newx (first newvar))
	 
	 (dx (/ (- xmax xmin) xsubdivisions))
	 (dy (/ (- ymax ymin) ysubdivisions))
	 yi
	 )

    (dolist
     (i (rseq xmin xmax xsubdivisions))
     (dolist
      (j (rseq ymin ymax ysubdivisions))
      (let* (
	     (dyy (* dx (funcall f i j)))
	     (u (* dx (normalize (list dx dyy))))
	     (v (/ u 2))
	     )
	(send w :add-lines
	      (list (- i (first v)) (+ i (first v)))
	      (list (- j (second v)) (+ j (second v)))
	      )
	)
      )
     )
    (send w :adjust-to-data :draw nil)
    (if *x-axis* (send w :abline 0 0))
    (if *y-axis* (send w :add-lines '(0 0) (send w :range 1)))
    (send w :new-menu)

    (send w :add-slot 'x (/ (send w :canvas-width) 2 ))
    (send w :add-slot 'y (/ (send w :canvas-height) 2 ))
    
    (defmeth w :x (&optional (val nil set))
      (if set (setf (slot-value 'x) val))
      (slot-value 'x))
    
    (defmeth w :y (&optional (val nil set))
      (if set (setf (slot-value 'y) val))
      (slot-value 'y))
    
    (defmeth w :do-click (x y m1 m2)
      (send self :x x)
      (send self :y y)
      (let* (
	     (lb (send self :canvas-to-scaled x y))
	     (xvals (rseq (first lb)
			  (min xmax (second (send self :range 0))) 100))
	     (dx (first (/ (- (last xvals) (first xvals)) 100)))
	     (yo (second lb))
	     )
	(dolist
	 (i xvals)
	 (let ((yi (+ yo (* dx (funcall f i yo)))))
	   (send self :add-lines (list i (+ i dx)) (list yo yi))
	   (setq yo yi))
	 )
	(send self :redraw))
      )  
    
    (send w :show-window)
    w
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                slope-field
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmacro slope-field (f &rest e)
  "
;; Description: (slope-field f xmin xmax ymin ymax
		      &optional (xsubdivisions 20) (ysubdivisions 20))

        We produce the slope-field corresponding to a differential equation of
the form
                              dy/dx = f(x y)
;; Examples:
 (title \"Slope field of -x/y\")
 (func f (x y) (-  x / y))
 (slope-field f -2 2 -2 2)

 (title \"Slope field of 2x\")
 (func f (x y) 2 * x)
 (slope-field f -2 2 -2 2)

 (slope-field (func f(x y) (cos y * sin x)) -pi pi -pi pi)
"
  (let ((bing (eval `(turn-it-into-a-function ,f))))
    `(slope-field2 ,bing ,@e)
    )
  )


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
(defun euler2 (f yo xmin xmax dx &key (print t))
  "
;; Description: (euler2 f yo xmin xmax step-size &key (print t))

;; Example:
 (euler2 (func f(x y) x) 1 0 1 .1)
"
  (let* (
;;	 (w (send scatterplot-proto :new 2 :title *title* :show nil))
	 (w (plotter))
	 ;; What is the independent variable?
	 (newvar (second (eval `(get-lambda-expression ,f))))
	 (newx (first newvar))
	 (newy (if (> (length newvar) 2) (second newvar) nil))
	 (dy nil)
	 (x xmin)
	 (y yo)
	 )
    (if print (print (list x y)))
    (send w :add-points (list x) (list y))
    (dolist
     (i (iseq (ceil (/ (- xmax xmin) dx))))
     (let (
	   (ny (+ y (* dx (funcall f x y))))
	   (nx (+ x dx))
	   )
       (if print (print (list nx ny)))
       (send w :add-lines (list x nx) (list y ny))
       (setq x nx y ny)
       )
     )

    (send w :adjust-to-data :draw nil)
    (if *x-axis* (send w :abline 0 0))
    (if *y-axis* (send w :add-lines '(0 0) (send w :range 1)))
    (send w :new-menu)

    (send w :add-slot 'x (/ (send w :canvas-width) 2 ))
    (send w :add-slot 'y (/ (send w :canvas-height) 2 ))
    (send w :location 300 20)
    (send w :range 0 xmin xmax)
    
    (defmeth w :x (&optional (val nil set))
      (if set (setf (slot-value 'x) val))
      (slot-value 'x))
    
    (defmeth w :y (&optional (val nil set))
      (if set (setf (slot-value 'y) val))
      (slot-value 'y))
    
    (defmeth w :do-click (x y m1 m2)
      (send self :x x)
      (send self :y y)
      (let* (
	     (lb (send self :canvas-to-scaled x y))
	     (xvals (rseq (first lb)
			  (min xmax (second (send self :range 0))) 100))
	     (dx (first (/ (- (last xvals) (first xvals)) 100)))
	     (yo (second lb))
	     )
	(dolist
	 (i xvals)
	 (let ((yi (+ yo (* dx (funcall f i yo)))))
	   (send self :add-lines (list i (+ i dx)) (list yo yi))
	   (setq yo yi))
	 )
	(send self :redraw))
      )  
    (send w :show-window)
    w
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; euler Doc:    Creation Date:3/28/96        Author: Andy Long
;;
;;
(defmacro euler (g &rest e)
  "
;; Description/Arguments:

		(euler g yo xmin xmax nsteps)
solves  
                              dy/dx=g(x y)

	Euler computes the euler approximate solution for a given function g
representing the right hand side of a first-order differential equation.

	Note that the function g should include both variables (dependent y and
independent x) in its definition:

;; Examples:
 
 (title \"dy/dx=y(1-y)\")
 (func g(x y) y * (1 - y)) ;; for the logistic diffeq  dx/dt=x(1-x)
 (euler g .5 0 5 .1) ;; computes values of the solution from 0 to 5, steps of .1

 (title \"dy/dx=x^2-y^2\")
 (func g(x y) x ^ 2 - y ^ 2)
 (euler g 1 0 5 .05) ;; approximates solution from 0 to 5, steps of .05
"
 (let ((bing (eval `(turn-it-into-a-function ,g))))
   `(euler2 ,bing ,@e)
   )
 )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
(defun 2euler2 (f yo y1 xmin xmax dx
		   &optional (ftrue nil))
  "
;; Description: (2euler2 f yo y1 xmin xmax step-size &optional (ftrue nil))

;; Example:
 (2euler2 (func f(x y) -32) 6 11.76 0 2 .2) ;; for the trajectory of chalk
"
  (let* (
;;	 (w (send scatterplot-proto :new 2 :title *title* :show nil))
	 (w (plotter))
	 ;; What is the independent variable?
	 (newvar (second (eval `(get-lambda-expression ,f))))
	 (newx (first newvar))
	 (newy (if (> (length newvar) 2) (second newvar) nil))
	 (xold xmin)
	 (yold yo)
	 (x (+ xold dx))
	 (y y1)
	 )
    (send w :add-points (list xold x) (list yold y))
    (send w :add-lines (list xold x) (list yold y))
    (dolist
     (i (rest (iseq (ceil (/ (- xmax xmin) dx)))))
     (let (
	   (ny (+ (* 2 y) (- yold) (* dx dx (funcall f xold yold))))
	   (nx (+ x dx))
	   )
       (send w :add-lines (list x nx) (list y ny))
       (setq xold x x nx yold y y ny)
       )
     )

    (send w :adjust-to-data :draw nil)
    (if *x-axis* (send w :abline 0 0))
    (if *y-axis* (send w :add-lines '(0 0) (send w :range 1)))
    (send w :new-menu)

    (send w :add-slot 'x (/ (send w :canvas-width) 2 ))
    (send w :add-slot 'y (/ (send w :canvas-height) 2 ))
    (send w :location 300 20)
    (send w :range 0 xmin xmax)
    (if ftrue (eval `(send ,w :add-function ,ftrue ,xmin ,xmax)))
    
    (defmeth w :x (&optional (val nil set))
      (if set (setf (slot-value 'x) val))
      (slot-value 'x))
    
    (defmeth w :y (&optional (val nil set))
      (if set (setf (slot-value 'y) val))
      (slot-value 'y))
    
    (send w :show-window)
    w
    )
  )
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; 2euler Doc:    Creation Date:3/28/96        Author: Andy Long
;;
;;
(defmacro 2euler (g &rest e)
  "
;; Description/Arguments:

       (2euler g yo y1 xmin xmax step-size)
  
	2euler computes the 2euler approximate solution for a given function g
representing the right hand side of a second-order differential equation.

	Note that the function g should include both variables (dependent y and
independent x) in its definition (it may not include the first-derivative y'):

                              d^2y/dx^2=g(x y)
;; Examples:
 
 (title \"d^2y/dx^2=-32\")
 (func g(x y) -32) ;; for the trajectory of chalk, with ho=6, vo=32:
 (func f(x) -16 * x ^ 2 + 32 * x + 6)
 (2euler g 6 (f .2) 0 2 .2 (func f(x) -16 * x ^ 2 + 32 * x + 6))
 
 (title \"d^2y/dx^2=-9y\")
 (func g(x y) (-9 * y)) ;; harmonic oscillator, sin(3x)
 (2euler g 0 (sin (* 3 .2)) 0 pi .2 (func f(x) (sin (3 * x))))
 (2euler g 0 (sin (* 3 .02)) 0 pi .02 (func f(x) (sin (3 * x))))
"
 (let ((bing (eval `(turn-it-into-a-function ,g))))
   `(2euler2 ,bing ,@e)
   )
 )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 stiff-ode
;;
(defun stiff-ode
  (f
   fy 
   ynot 
   start 
   end 
   num 
   &key 
   non-autonomous 
   (maxits 100) 
   (tol 1e-6))
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          stiff-ode Doc:
;;
;; Author: Andy Long
;; Created: Fri Jan 23 18:27:54 1998
;;
;; Description/Arguments:

 (stiff-ode
   f
   fy 
   ynot 
   start 
   end 
   num 
   &key 
   non-autonomous 
   (maxits 100) 
   (tol 1e-6))
  
;; Returns:

;; Notes:

;; See Also: rk-4, rk-f, euler-nd

;; Examples:

 ;;			One-Dimensional examples:

 (title \"y'(t)=-1, where y_0=1\")
 (func f(x) -1)
 (func fy(x) 0)
 (def points (stiff-ode #'f #'fy '(1) 0 3 20))
 (def points (mapcar #'first points))
 (plot (rseq 0 3 20) points)

 (title \"y'(t)=-y, where y_0=1\")
 (func f(x) -x)
 (func fy(x) -1)
 (def points (stiff-ode #'f #'fy '(1) 0 3 20))
 (def points (mapcar #'first points))
 (plot (rseq 0 3 20) points)
 ;;				Non-autonomous:

 ;; Checks against Burden and Faires 'Numerical Analysis'....
 (title \"y'(x)=5*e^(5*t)*(y-t)^2+1, y(0)=.5\")
 (func f(x h) 5*(exp (5*h))*(x-h)^2+1)
 (func fy(x h) 10*(exp (5*h))*(x-h))

 ;; Both methods are okay here: the points are indistinguishable on a plot...
 (def points (stiff-ode #'f #'fy '(-1) 0 1 6 :non-autonomous t))
 (def points (mapcar #'first points))
 (pprint points)
 (plot (rseq 0 1 6) points)
 (def points (rk-4 #'f '(-1) 0 1 6 :non-autonomous t))
 (def points (mapcar #'first points))
 (pprint points)
 (send my-plot :add-plot-item
      (list 'points (rseq 0 1 6) points 'blue)
      (list 'function (func true-solution(x) x-(exp (-5*x))) 'red 'lines)
      )

 ;; Shows rk-4 blowing up, while implicit trapezoidal is stable....
 (def points (stiff-ode #'f #'fy '(-1) 0 1 5 :non-autonomous t))
 (pprint points)
 (def points (rk-4 #'f '(-1) 0 1 5 :non-autonomous t))
 (pprint points)
"
  (let* (
	 (h (/ (- end start) (1- num)))
	 (half-h (/ h 2))
	 (time start)
	 (yvalues (list ynot))
	 (w ynot)
	 )
    (if non-autonomous 
	(dotimes
	 (i (1- num))
	 (let (
	       (k1 (+ w (* half-h (funcall f w time))))
	       (time+ (+ time h))
	       w1
	       )
	   (do*
	    (
	     (j 1 (incf j))
	     (w0 k1 w1)
	     (num (- w0 (* half-h (funcall f w0 time+)) k1)
		  (- w0 (* half-h (funcall f w0 time+)) k1)
		  )
	     (denom (- 1 (* half-h (funcall fy w0 time+)))
		    (- 1 (* half-h (funcall fy w0 time+)))
		    )
	     )
	    () ;; keep going until a return is forced:
	    (setq w1 (- w0 (/ num denom)))
	    (if (< (norm (- w1 w0)) tol) (return nil))
	    (if (> j maxits)
		(error
		 "Maximum Number (~d) of iterations exceeded in stiff-ode."
		 maxits))
	    )
	   (setq 
	    w w1
	    time time+
	    yvalues (append yvalues (list w))
	    )
	   )
	 )
      (dotimes
       (i (1- num))
       (let (
	     (k1 (+ w (* half-h (funcall f w))))
	     w1
	     )
	 (do*
	  (
	   (j 1 (incf j))
	   (w0 k1 w1)
	   (num (- w0 (* half-h (funcall f w0)) k1)
		(- w0 (* half-h (funcall f w0)) k1)
		)
	   (denom (- 1 (* half-h (funcall fy w0)))
		  (- 1 (* half-h (funcall fy w0)))
		  )
	   )
	  () ;; keep going until a return is forced:
	  (setq w1 (- w0 (/ num denom)))
	  (if (< (norm (- w1 w0)) tol) (return nil))
	  (if (> j maxits)
	      (error
	       "Maximum Number (~d) of iterations exceeded in stiff-ode."
	       maxits))
	  )
	 (setq 
	  w w1
	  yvalues (append yvalues (list w))
	  )
	 )
       )
      )
    yvalues
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
