;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                Odds-ratio
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun odds-ratio
  (a b c d)
  "
;; Description: odds-ratio(a b c d)
computes the odds-ratio for the matrix
 		[a b]
		[c d],
i.e., (ad)/(bc).

;; Examples:
 (odds-ratio 647 2 622 27)
 ;; 14.042604501607718
 (odds 41 58 10 49)
 ;; 3.463793103448276
"
  (/ (* a d) (* b c))
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                relative-risk
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun relative-risk
  (a b c d)
  "
;; Description: relative-risk(a b c d)
computes the relative-risk for the matrix
 		[a b]
		[c d],
i.e., (a/(a+b))/(c/(c+d)).

;; Examples:
 (relative-risk 647 2 622 27) ;; 1.040192926045016
"
  (/ (/ a (+ a b)) (/ c (+ c d)))
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 2x2-diagnostic-testing
;;
(defun 2x2-diagnostic-testing(a b c d)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          2x2-diagnostic-testing Doc:
;;
;; Author: Andy Long
;; Created: Wed Apr 9 09:32:49 CDT 1997
;;
;; Description/Arguments:

       (2x2-diagnostic-testing a b c d)
  
;; Notes: returns (values sens spec ppv npv prev lr+ lr-)

;; See Also:

;; Examples:
 (2x2-diagnostic-testing 3 2 1 4)
 (2x2-diagnostic-testing 20 10 10 20)
 (2x2-diagnostic-testing 25 15 5 15)
 (2x2-diagnostic-testing 649 99 160 1671)
 (2x2-diagnostic-testing 252 52 168 528)
 (2x2-diagnostic-testing 60 81 40 869)

 (2x2-diagnostic-testing 10 6 0 4)
 (2x2-diagnostic-testing 9 3 1 7)
 (2x2-diagnostic-testing 7 1 3 9)
 (2x2-diagnostic-testing 5 0 5 10)

 (def results (cmv (2x2-diagnostic-testing 39 604 216 5762)))
"
  (let* (
	 (sens (/ a (+ a c)))
	 (spec (/ d (+ b d)))
	 (ppv (/ a (+ a b)))
	 (npv (/ d (+ c d)))
	 (lr+ (lr+ sens spec))
	 (lr- (lr- sens spec))
	 (prev (/ (+ a c) (+ a b c d)))
	 (fmt1 "~%~19t~a ~34t~a ~49t~a~%")
	 (fmt2 "~1t~a ~14t~d ~25t~d ~35t~,4g ~55t~,6g~%")
	 (fmt3 "~1tSe/Sp: ~12t~,4g ~22t~,4g~%")
	 (fmt4 "~1tPrevalence: ~15t~,4g~%")
	 )
    (format t fmt1 "Truth"  "Predictive Value" "Likelihood Ratio")
    (format t "~14t+ ~25t-~%")
    (format t fmt2 "Test +" a b ppv lr+)
    (format t fmt2 "Test -" c d npv lr-)
    (format t fmt3 sens spec)
    (format t fmt4 prev)
    (values sens spec ppv npv prev lr+ lr-)
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 probability-to-odds
;;
(defun probability-to-odds(p)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          probability-to-odds Doc:
;;
;; Author: Andy Long
;; Created: Mon May 19 13:16:16 CDT 1997
;;
;; Description/Arguments:

       (probability-to-odds p)
  
;; Notes:

;; See Also:

;; Examples:
 (probability-to-odds .75)

"
  (/ p (- 1 p))
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 odds-to-probability
;;
(defun odds-to-probability(p)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          odds-to-probability Doc:
;;
;; Author: Andy Long
;; Created: Mon May 19 13:16:16 CDT 1997
;;
;; Description/Arguments:

       (odds-to-probability p)
  
;; Notes:

;; See Also:

;; Examples:
 (odds-to-probability 3)

 (def lr+ (/ .8 (- 1 .75)))
 (odds-to-probability (* (probability-to-odds .75) lr+))
"
  (/ p (1+ p))
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 update-test-probability
;;
(defun update-test-probability(pretest-prob lr)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          update-test-probability Doc:
;;
;; Author: Andy Long
;; Created: Mon May 19 13:25:41 CDT 1997
;;
;; Description/Arguments:

       (update-test-probability pretest-prob lr)
  
;; Notes:

;; See Also:

;; Examples:
 (update-test-probability .75 (lr+ .8 .75))
 (update-test-probability .5 39)
 (update-test-probability .75 .6)
"
  (odds-to-probability (* (probability-to-odds pretest-prob) lr))
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 lr+
;;
(defun lr+(sens spec)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          lr+ Doc:
;;
;; Author: Andy Long
;; Created: Mon May 19 13:28:02 CDT 1997
;;
;; Description/Arguments:

       (lr+ sens spec)
  
;; Notes:

;; See Also:

;; Examples:
 (lr+ .8 .75)
"
  (if (= spec 1) Nan (/ sens (- 1 spec)))
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 lr-
;;
(defun lr-(sens spec)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          lr- Doc:
;;
;; Author: Andy Long
;; Created: Mon May 19 13:28:02 CDT 1997
;;
;; Description/Arguments:

       (lr- sens spec)
  
;; Notes:

;; See Also:

;; Examples:
 (lr- .8 .75)
"
  (if (= spec 0) Nan (/ (- 1 sens) spec))
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 post-test-probability
;;
(defun post-test-probability(likelihood-ratio &key title)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          post-test-probability Doc:
;;
;; Author: Andy Long
;; Created: Wed May 21 11:12:36 CDT 1997
;;
;; Description/Arguments:

       (post-test-probability likelihood-ratio &key title)
  
	This function is designed to plot the post-test probability of having
a disease, given the pre-test probability and the likelihood ratio(s) of the
test. If you supply a single likelihood-ratio, then you already know whether
the test was positive or negative, and you'll get a single function. If you
supply a pair of values, then you'll get a pair of functions (one for a
positive result and one for a negative result).

;; Notes: 

;; See Also:

;; Examples:
 (post-test-probability 9)
 (post-test-probability 1) ;; such a test tells you nothing!
 (post-test-probability '(9 .4))
"
  (setq title
	(if title title
	  (if (listp likelihood-ratio)
	      (format nil "Likelihood ratios: ~,4g and ~,4g"
		      (first likelihood-ratio) (second likelihood-ratio))
	    (format nil "Likelihood ratio: ~,4g" likelihood-ratio)
	    )))
  (if (listp likelihood-ratio)
      (plotter
       (lambda(x)
	 (if (= x 0) 0
	   (/ (1+ (/ (1- (/ x)) (first likelihood-ratio))))
	   ))
       (lambda(x)
	 (if (= x 0) 0
	   (/ (1+ (/ (1- (/ x)) (second likelihood-ratio))))
	   ))
       0 1
       )
    (plotter
     (lambda(x) (if (= x 0) 0 (/ (1+ (/ (1- (/ x)) likelihood-ratio)))))
     0 1
     )
    )    
  (mapcar
   (lambda(y) (send my-plot :add-plot-item y))
   (smoosh
    (mapcar
     (lambda(x y)
       (list
	(list 'lines (list 0 1) (list x x) 'black)
	(list 'lines (list y y) (list 0 1) 'black)
	)
       )
     (rest (rseq 0 1 *y-tics*))
     (rest (rseq 0 1 *x-tics*))
     )
    )
   )
  (send my-plot :variable-label '(0 1)
	'("Pre-test probability" "Post-test probability"))
  (send my-plot :title title)
  (make-title title)
  (send my-plot :plot-item)
  my-plot
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 kaplan-meier
;;
(defun kaplan-meier(times events &key (n 10) time-step print)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          kaplan-meier Doc:
;;
;; Author: Andy Long
;; Created: Thu Jul 3 16:20:36 CDT 1997
;;
;; Description/Arguments:

       (kaplan-meier times events &key (n 10) time-step print)
  
	times are the survival times 
	events are the end results - 1 for an event, 0 otherwise
	n is the (optional) number of subdivisions of the time scale at which
		to calculate.

;; Notes:

;; See Also:

;; Examples:
 (def deaths '(0 0 0 0 1 0 0 0 0 0 1 1 0 1 0 0 0 1 0 1 0 0 0 0 0 0 0 1 0 0 0 1
   0 0 0 0 0 0 0 0 0 0 1 0 0 0 1 1 1 0 1 1 1 0 0 1 0 0 0 1 0 0 1 0 1 0 1 1 0 0
   1 1 0 0 0 0 0 0 0 1 0 1 0 0 0 0 0 0 0 1 0 0 0 0 1 0 0 0 0 0 1))
 (def days '(771 763 751 750 529 749 748 748 745 745 511 602 743 260 742 741
    737 4 737 179 735 731 729 729 728 724 722 347 721 721 721 332 714 710 710
    709 709 707 706 703 702 700 510 694 694 692 373 467 233 685 610 350 403 674 673
    118 666 666 665 493 660 660 131 657 387 646 192 276 643 640 75 564 623 615 612
    611 608 605 603 538 601 225 584 580 580 577 577 574 573 102 560 560 559 545 533
    541 540 539 534 533 15) )
 (kaplan-meier days deaths)

 ;; From Kahn and Sempos, Statistical Methods in Epidemiology, p. 192:
 (def months (list 2 5 6 9 9 12 17 4 12 15))
 (def deaths (list 1 1 1 1 1 1 1 0 0 0))
 (kaplan-meier months deaths :n 12 :time-step 1)
 (kaplan-meier months deaths :n 18 :time-step 1 :print t)
"
  (let* (
	 (order (order times))
	 (times (select times order))
	 (events (select events order))
	 (time-step (if time-step time-step (/ (max times) n)))
	 (probs (repeat 0 n))
	 )
    (dotimes
     (i n)
     (let* (
	    (len (length times))
	    (smaller (where times (* (1+ i) time-step) #'<=))
	    (sum-events (sum (select events smaller)))
	    )
       (setf (elt probs i)
	     (*
	      (if (= i 0) 1 (elt probs (1- i)))
	      (- 1 (/ (if sum-events sum-events 0) len))))
       (if print (print (list (1+ i) (elt probs i))))
       (if (= (length smaller) (length times)) (return 0)
	 (setq
	  times (select times (iseq (length smaller) (1- len)))
	  events (select events (iseq (length smaller) (1- len)))
	  )
	 )
       )
     )
    probs
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 mantel-haenszel
;;
(defun mantel-haenszel
  (times events groups
	 &key (n 10) time-step print group-values (correction .5))
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          mantel-haenszel Doc:
;;
;; Author: Andy Long
;; Created: Tue Jul 8 10:15:51 CDT 1997
;;
;; Description/Arguments:

       (mantel-haenszel
         times events groups
	   &key (n 10) time-step print group-values (correction .5))
  
;; Notes:

;; See Also:

;; Examples:
 ;; Reproduces Page 187, Statistical Methods in Epidemiology:
 (def ar1 (list 297 97 63 41 23 9))
 (def ar2 (list 297 132 102 87 72 61))
 (def d1 (list 193 25 12 8 4 3))
 (def d2 (list 165 30 15 15 11 17))
 (def drop1 (- (butlast (- ar1 d1)) (rest ar1)))
 (def drop1 (combine drop1 (- 297 (sum d1) (sum drop1))))
 (def drop2 (- (butlast (- ar2 d2)) (rest ar2)))
 (def drop2 (combine drop2 (- 297 (sum d2) (sum drop2))))

 (def deaths
   (combine
     (mapcar (lambda(x y) (combine (repeat 1 x) (repeat 0 y))) d1 drop1)
     (mapcar (lambda(x y) (combine (repeat 1 x) (repeat 0 y))) d2 drop2)
     ))
 (def deaths (select deaths (which deaths)))
 (def days      
   (combine
     (mapcar (lambda(x y) (repeat x y)) '(2 4 6 8 10 12) (+ d1 drop1))
     (mapcar (lambda(x y) (repeat x y)) '(2 4 6 8 10 12) (+ d2 drop2))
     ))
 (def groups (combine (repeat 0 297) (repeat 1 297)))
 (mantel-haenszel days deaths groups :n 6 :time-step 2 :print t)
 (mantel-haenszel days deaths groups :n 6 :time-step 2 :group-values '(0 1))
 (mantel-haenszel days deaths groups :n 6 :time-step 2 :group-values '(1 0))

 ;; From Biostatistics, p. 550:
 (def groups (combine (repeat 0 604) (repeat 1 614)))
 (def days (combine (repeat 1 303) (repeat 2 301)
                    (repeat 1 303) (repeat 2 311)))
 (def deaths (combine
               (repeat 0 277) (repeat 1 26) (repeat 0 287) (repeat 1 14)
               (repeat 0 260) (repeat 1 43) (repeat 0 286) (repeat 1 25)))
 (mantel-haenszel days deaths groups :n 2 :time-step 1 :print t)

 ;; A made up example:
 (def deaths '(0 0 0 0 1 0 0 0 0 0 1 1 0 1 0 0 0 1 0 1 0 0 0 0 0 0 0 1 0 0 0 1
   0 0 0 0 0 0 0 0 0 0 1 0 0 0 1 1 1 0 1 1 1 0 0 1 0 0 0 1 0 0 1 0 1 0 1 1 0 0
   1 1 0 0 0 0 0 0 0 1 0 1 0 0 0 0 0 0 0 1 0 0 0 0 1 0 0 0 0 0 1))
 (def days '(771 763 751 750 529 749 748 748 745 745 511 602 743 260 742 741
    737 4 737 179 735 731 729 729 728 724 722 347 721 721 721 332 714 710 710
    709 709 707 706 703 702 700 510 694 694 692 373 467 233 685 610 350 403 674 673
    118 666 666 665 493 660 660 131 657 387 646 192 276 643 640 75 564 623 615 612
    611 608 605 603 538 601 225 584 580 580 577 577 574 573 102 560 560 559 545 533
    541 540 539 534 533 15) )
 ;; randomly assigned group should show nothing:
 (mantel-haenszel days deaths (sample '(0 1) (length days) t) :print t)

 ;; grouped by death! should see a strong effect!
 (mantel-haenszel days deaths deaths :print t)
"
  (let* (
	 (group-types (if group-values group-values (remove-duplicates groups)))
	 (ctmp (if (/= 2 (size group-types))
		   (error "Mantel-Haenszel only works for two groups!")
		 (format t "~%Groups: ~s and ~s~%"
			 (first group-types) (second group-types))
		 ))
	 (group1 (where groups (first group-types)))
	 (group2 (where groups (second group-types)))
	 (times1 (select times group1))
	 (times2 (select times group2))
	 (events1 (select events group1))
	 (events2 (select events group2))
	 (order1 (order times1))
	 (order2 (order times2))
	 (time-step (if time-step time-step (/ (max times) n)))
	 (probs (repeat 0 n))
	 (mh-odds-num 0)
	 (mh-odds-denom 0)
	 (mh-denom 0)
	 (chisq 0)
	 as bs cs ds daniel
	 )
    (if print (format t "Odds ratios by time step: these should be roughly the same.

~10t Time ~18t Odds-ratio~%"))
    (dotimes
     (i n)
     (let* (
	    (len1 (length times1))
	    (len2 (length times2))
	    (smaller1 (where times1 (* (1+ i) time-step) #'<=))
	    (smaller2 (where times2 (* (1+ i) time-step) #'<=))
	    (sum-events1 (sum (select events1 smaller1)))
	    (sum-events2 (sum (select events2 smaller2)))
	    (a (if sum-events1 sum-events1 0))
	    (b (- (if len1 len1 0) a))
	    (c (if sum-events2 sum-events2 0))
	    (d (- (if len2 len2 0) c))
	    (sum (+ len1 len2))
	    )
       (if (= i 0)
	   (setq as a bs b cs c ds d)
	 (setq as (combine as a)
	       bs (combine bs b) cs (combine cs c) ds (combine ds d))
	 )	    
       (setq
	chisq (+ chisq (- a (/ (* (+ a b) (+ a c)) sum)))
	
	mh-odds-num (+ mh-odds-num (/ (* a d) sum))
	mh-odds-denom (+ mh-odds-denom (/ (* c b) sum))
	
	mh-denom (+ mh-denom
		    (/
		     (* (+ a b) (+ c d) (+ a c) (+ b d))
		     (* (square sum) (1- sum))))
	)
       (setf (elt probs i) (odds a b c d))
       (if print
	   (format t "~10t~,6g~20t~,6g~%" (* time-step (1+ i)) (elt probs i)))
       (if (or
	    (= (length smaller1) (length times1))
	    (= (length smaller2) (length times2))
	    )
	   (return 0)
	 (setq
	  times1 (select times1 (iseq (length smaller1) (1- len1)))
	  times2 (select times2 (iseq (length smaller2) (1- len2)))
	  events1 (select events1 (iseq (length smaller1) (1- len1)))
	  events2 (select events2 (iseq (length smaller2) (1- len2)))
	  )
	 )
       )
     )
    (setq chisq (square (- (abs chisq) correction)))
    (setq daniel (catch-multiple-values (summary-mantel-haenszel as bs cs ds)))
    (if print
	(format t "
Mantel-Haenszel Odds-ratio: ~,6g
Statistical Methods X^2   : ~,6g
Probability               : ~,6f
Daniel's Biostatistics X^2: ~,6g
Probability               : ~,6f
"
		(/ mh-odds-num mh-odds-denom)
		(/ chisq mh-denom)
		(- 1 (chisq-cdf (/ chisq mh-denom) 1))
		(second daniel)
		(third daniel)
		))
    (values
     (/ mh-odds-num mh-odds-denom)
     (/ chisq mh-denom)
     (- 1 (chisq-cdf (/ chisq mh-denom) 1))
     (second daniel)
     (third daniel)
     probs)
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 summary-mantel-haenszel
;;
(defun summary-mantel-haenszel (a b c d &key print)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          summary-mantel-haenszel Doc:
;;
;; Author: Andy Long
;; Created: Tue Jul 8 10:15:51 CDT 1997
;;
;; Description/Arguments:

       (summary-mantel-haenszel a b c d &key print)
  
	Mantel-Haenszel estimator of the common odds ratio. This is described
in Daniel's Biostatistics, 6th edition, pages 550 ff.

;; Notes:

;; See Also: mantel-haenszel, summary-mh-helper

;; Examples:
 ;; From Biostatistics, p. 550:
 (summary-mantel-haenszel '(43 25) '(260 286) '(26 14) '(277 287) :print t)
"
  (let* (
	 (probs (mapcar #'summary-mh-helper a b c d))
	 (mh-odds-num 0)
	 (mh-odds-denom 0)
	 (mh-num 0)
	 (mh-denom 0)
	 as
	 chi-sq
	 )
    (if print (format t "Odds ratios by time step: these should be roughly the same.

~9tGroup ~18t Odds-ratio~%"))
    (setq
     as
     (mapcar
      (lambda(i a b c d)
	(let ((sum (+ a b c d)))
	  (setq
	   mh-odds-num (+ mh-odds-num (/ (* a d) sum))
	   mh-odds-denom (+ mh-odds-denom (/ (* c b) sum))
	   )
	  (if print (format t "~%~10t~d~20t~,6g~%" i (/ mh-odds-num mh-odds-denom)))
	  )
	)
      (iseq 1 (length a)) a b c d)
     chi-sq (/ (sum (mapcar #'first probs)) (sum (mapcar #'second probs)))
     )
    (if print
	(format t "
Mantel-Haenszel Odds-ratio: ~,6g
Daniel's Biostatistics X^2: ~,6g
Probability               : ~,6f
"
		(/ mh-odds-num mh-odds-denom)
		chi-sq
		(- 1 (chisq-cdf chi-sq 1))
		))
    (values
     (/ mh-odds-num mh-odds-denom)
     chi-sq
     (- 1 (chisq-cdf chi-sq 1))
     )
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 summary-mh-helper
;;
(defun summary-mh-helper(a b c d)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          summary-mh-helper Doc:
;;
;; Author: Andy Long
;; Created: Wed Jul 9 11:16:35 CDT 1997
;;
;; Description/Arguments:

       (summary-mh-helper a b c d)
  
;; Notes:

;; See Also: summary-mantel-haenszel

;; Examples:
 (summary-mh-helper 43 260 26 277)
"
  (let (
	(n (+ a b c d))
	)
    (list
     (square (- a (/ (* (+ a b) (+ a c)) n)))
     (/
      (* (+ a b) (+ c d) (+ a c) (+ b d))
      (* (square n) (1- n))
      )
     )
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
