;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                            vario
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun vario(data grid-size &key verbose)
  "
	(vario data grid-size &key verbose)

	This works in two dimensions ONLY. It assumes that you're
going to give it a list of triplets, and the size of the (square) grid for
which you want the variogram (it will be even, one more than any odd you
give it).

;; Examples:
 
 (def data
   (mapcar #'(lambda (x) (combine x (* (first x) (second x))))
    (weave (rand.u 10) (rand.u 10))))
 (def vardata (vario data 3))
 (plot (/ (first vardata) (1+ (second vardata))))

 (def data
   (mapcar #'(lambda (x) (combine x (first x)))
    (weave (rand.u 100) (rand.u 100))))
 (plot data)
 (def vardata (vario data 7))
 (plot (/ (first vardata) (second vardata)))

 (def data
   (mapcar #'(lambda (x) (combine x (* (first x) (second x))))
    (weave (rand.u 100) (rand.u 100))))
 (plot data)
 (def vardata (vario data 5))
 (plot (/ (first vardata) (1+ (second vardata))))

 (def data
   (mapcar #'(lambda (x)
     (combine x (+ (square (first x)) (square (second x)))))
     (weave (rand.u 100) (rand.u 100))))
 (plot data)
 (def vardata (vario data 5))
 (plot (/ (first vardata) (1+ (second vardata))))
"
  (let* (
	 (grid-size (if (oddp grid-size) grid-size (1+ grid-size)))
	 (half-grid-size (/ (1- grid-size) 2))
	 (xs (mapcar #'first data))
	 (ys (mapcar #'second data))
	 (xmin (min xs))
	 (xmax (max xs))
	 (ymin (min ys))
	 (ymax (max ys))
	 (xdiff (/ (- xmax xmin) 2))
	 (ydiff (/ (- ymax ymin) 2))
	 (xfactor (/ grid-size xdiff))
	 (yfactor (/ grid-size ydiff))
	 (len (length xs))
	 (1-len (1- len))
	 (grid (make-array (list (1+ grid-size) (1+ (* 2 grid-size)))
			   :initial-element 0))
	 (nums (make-array (size grid) :initial-element 0))
	 (i 0)
	 )
    (dolist
     (piece data)
     (if verbose (print i))
     (let (
	    (x (elt piece 0))
	    (y (elt piece 1))
	    (z (elt piece 2))
	    )
       (dolist
	(newpiece (select data (iseq i 1-len)))
	(let* (
	       (xx (elt newpiece 0))
	       (yy (elt newpiece 1))
	       (xdist (- x xx))
	       (abs-xdist (abs xdist))
	       (ydist (- y yy))
	       )
	  (if (and (< abs-xdist xdiff) (< (abs ydist) ydiff))
	      (let* (
		     (zz (elt newpiece 2))
		     (xd (round (* abs-xdist xfactor)))
		     (yd (+ grid-size
			    (round (* (if (< xdist 0) -1 1) ydist yfactor))))
		     )
		(setf (aref grid xd yd)
		      (+ (aref grid xd yd) (square (- z zz))))
		(setf (aref nums xd yd) (1+ (aref nums xd yd)))
		)
	    )
	  )
	)
       )
     (setq i (1+ i))
     )
    (list
     grid nums
     (rseq 0 xdiff (1+ grid-size))
     (rseq (- ydiff) ydiff (1+ (* 2 grid-size)))
     )
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun diffs (x)
  "Args: x a list. Computes the difference of list positions.
Useful for variogram-type calculations.

;; Examples:
 (def x (iseq 1 5))
 (diffs x)
"
  (let (
	(n (- (length x) 1))
	(xxx 0)
	)
    (dotimes
     (i n)
     (let (
	   (j (+ i 1))
	   )
       (setq xxx (combine xxx
			  (- (select x (iseq j n))
			     (select x (iseq 0 (- n j))))))
       )
     )
    (rest xxx)
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun variogram (x y)
  "
Description: (variogram x y)

	Creates and returns the variogram of y against lags from x.

;; Examples:
 (def x (iseq 1 10))
 (def y (+ (* .1 x) (normal-rand 10)))
 (def vario (variogram x y))
 (def h (first vario))
 (def vario (second vario))

 (def ker1 (kernel-smooth h vario :type 'g))
 (def ker2 (kernel-smooth h vario :type 'u))
 (def ker3 (kernel-smooth h vario :type 't))
 (def ker4 (kernel-smooth h vario :type 'b))

 (new-plot)
 (lines off)
 (plot h vario)
 (send my-plot :add-plot-item
   (list 'lines (first ker1) (second ker1) 'black)
   (list 'lines (first ker2) (second ker2) 'red)
   (list 'lines (first ker3) (second ker3) 'yellow)
   (list 'lines (first ker4) (second ker4) 'blue)
   )
 (send my-plot :update-plot)
"
  (let* (
	 (h (abs (diffs (order x))))
	 (var (diffs y))
	 )
    (list h (* .5 var var))
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun variance (x)
  "
Args: (x), a list. Calculates the sample variance.

;; See also: standard-deviation

;; Examples:
 (def z '(1 2 3 5 7))
 (variance z)
"
  (square (standard-deviation x))
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 global-krig
;;
(defun global-krig(data estimation-sites variogram &rest args)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; global-krig Doc:    Creation Date:2/3/97        Author: Andy Long
;;
;; Description/Arguments:

       (global-krig data estimation-sites variogram &key weights)
  
	Global-Krig implements global-kriging, and assumes a vectorized
variogram.

;; Notes:

;; See Also:

;; Examples:
 ;; Create some data:
 (def x (rand.u 100))
 (def y (rand.u 100))
 (def z (+ x y))

 ;; Plot the variogram:
 (plot (first (vario (weave x y z) 10)))

 ;; Do the kriging:
 (def estimates
   (global-krig
     (weave x y z)
     (make-grid 0 1 10 0 1 10)
     #'(lambda(x) (linear-variogram x 0 1))
   ))
 (plot estimates)

 ;; Demonstrates interpolation of global kriging:
 (def x (rseq 0 2 5))
 (def y (repeat 1 5))
 (def z (rand.u 5))
 (def estimates
   (global-krig
     (weave x y z)
     (weave x y)
     #'(lambda(x) (linear-variogram x 0 1))
   ))
 (lines off)
 (plot x z)
 (send my-plot :add-plot-item (list 'lines (xs estimates) (zs estimates)))

 ;; Demonstrates that a linear variogram (w/o nugget) just plays
 ;; connect-the-dots:
 (def x (rseq 0 2 5))
 (def y (repeat 1 5))
 (def z (rand.u 5))
 (def estimates
   (global-krig
     (weave x y z)
     (make-grid 0 2 50 1 1 1)
     #'(lambda(x) (linear-variogram x 0 1))
     ))
 (lines off)
 (plot x z)
 (send my-plot :add-plot-item (list 'lines (xs estimates) (zs estimates)))

 ;; Demonstrates that a linear variogram just plays connect-the-dots, with jump
 ;; discontinuities, if the linear variogram has a nugget:
 (def x (rseq 0 2 5))
 (def y (repeat 1 5))
 (def z (rand.u 5))
 (def estimates
   (global-krig
     (weave x y z)
     (make-grid 0 2 177 1 1 1)
     #'(lambda(x) (linear-variogram x 1 1))
     ))
 (lines off)
 (plot x z)
 (send my-plot :add-plot-item (list 'lines (xs estimates) (zs estimates)))

 ;; Gives a picture of the weights:
 (def nrow 4)
 (def ncol 4)
 (multiple-value-setq
  'heavy
  (global-krig
   (mapcar #'combine
          (weave
           (make-grid 0 1 nrow 0 1 ncol)
           (rand.u (* nrow ncol))
           )
          )
   (list '(.5 .5))
   #'(lambda(x) (gaussian-variogram x 1 1 3))
   :verbose t
   )
  )
 (plot (matrix (list nrow ncol) (butlast heavy)))
"
  (let* (
	 (data-sites (mapcar #'butlast data))
	 (z (mapcar #'final data))
	 (N (length z))
	 (n-1 (1- n))
	 (verbose (first (catch-keys '(:verbose) args)))
	 (weights (first (catch-keys '(:weights) args)))
	 (weights (if weights weights (repeat 1 N)))
	 (mat (make-array (list N N) :initial-element 0))
	 solution estimates
	 )
    ;; Create the upper part of the kriging matrix:
    (mapcar
     (lambda(i)
       (let* (
	      (i+1 (1+ i))
	      (x0 (elt data-sites i))
	      (differences
	       (mapcar #'(lambda(x) (- x0 x))
		       (select data-sites (iseq i+1 (1- N)))))
	      )
	 (setf (select mat i (iseq i+1 n-1))
	       (matrix (list 1 (- n i+1)) (mapcar variogram differences)))
	 )
       )
     (iseq n-1)
     )
    ;; Create the full kriging matrix:
    (setq mat
	  (bind-columns
	   (bind-rows
	    (upper-to-full mat)
	    weights
	    )
	   (combine weights 0)
	   ))
    (if verbose (plotter mat))

    ;; calculate the global weights: data * K^{-1}:
    (setq solution (solve mat (combine (* z weights) 0)))
	  
    ;; Now move through the sites at which an estimate is desired:
    (dolist
     (x0 estimation-sites)
     (if verbose (print x0))
     (let (
	   (vario
	    (mapcar
	     variogram
	     (mapcar #'(lambda(x) (- x0 x)) data-sites)
	     )
	    )
	   )
       (setq estimates
	     (append estimates 
		     (list
		      (combine x0 (matmult (combine vario 1) solution))
		      )
		     )
	     )
       (if verbose
	   (plotter
	    (mapcar #'combine
		    (weave `,data-sites
			   (butlast (solve `,mat (combine `,vario 1)))))))
       )
     )
    ;;ael    estimates
    (values estimates solution)
    )
  )
;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 local-krig
;;
(defun local-krig (data sites variogram num
			&key weights verbose (use-vario-metric t))
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; local-krig Doc:    Creation Date:2/18/97        Author: Andy Long
;;
;; Description/Arguments:

       (local-krig data sites variogram num
		   &key weights verbose (use-vario-metric t))
  
	The idea is that I give a set of sites, the data (either in the form of
an N by three matrix, or as a list of sites), and the variogram model.
local-krig will then march over the sites, pick out the number to use (num) by
finding those sites at which the variogram is lowest. Optionally one assigns
weights.

	The use-vario-metric option is unusual, but seems reasonable: if it is
set to true, then the measure of distance of neighbors will not be physical
proximity, but rather proximity as measured by the size of the variogram
between the two points.

;; Notes:

;; See Also:

;; Examples:
 (def x (rseq 0 2 5))
 (def y (repeat 1 5))
 (def z (rand.u 5))
 (def new-estimates
     (local-krig
      (weave x y z)
      (weave x y)
      #'(lambda(x) (linear-variogram x 1 1))
      5
      :verbose t
      ))
 (lines off)
 (plot x z)
 (send my-plot :add-plot-item (list 'lines (xs new-estimates) (zs
    new-estimates))) 

 (def estimates
     (global-krig
      (weave x y z)
      (weave x y)
      #'(lambda(x) (linear-variogram x 1 1))
      ))
 (lines off)
 (plot x z)
 (send my-plot :add-plot-item (list 'lines (xs estimates) (zs estimates)))


 (def new-estimates
     (local-krig
      (weave x y z)
      (make-grid 0 2 20 1 1 1)
      #'(lambda(x) (linear-variogram x 1 1))
      5
      :verbose t
      ))
 (lines off)
 (plot x z)
 (send my-plot :add-plot-item
    (list 'lines (xs new-estimates) (zs new-estimates))) 

 (def estimates
     (global-krig
      (weave x y z)
      (make-grid 0 2 20 1 1 1)
      #'(lambda(x) (linear-variogram x 1 1))
      ))
 (lines off)
 (plot x z)
 (send my-plot :add-plot-item (list 'lines (xs estimates) (zs estimates)))

 (def x (rseq 0 2 25))
 (def y (repeat 1 25))
 (def z (rand.u 25))
 (def new-estimates
     (local-krig
      (weave x y z)
      (make-grid 0 2 10 1 1 1)
      #'(lambda(x) (linear-variogram x 1 1))
      10
      :verbose t
      ))
 (lines off)
 (plot x z)
 (send my-plot :add-plot-item
    (list 'lines (xs new-estimates) (zs new-estimates))) 

 (def estimates
     (global-krig
      (weave x y z)
      (make-grid 0 2 10 1 1 1)
      #'(lambda(x) (linear-variogram x 1 1))
      ))
 (lines off)
 (plot x z)
 (send my-plot :add-plot-item (list 'lines (xs estimates) (zs estimates)))
"
  (let* (
	 (data
	  (if (matrixp data)
	      (weave (column 1 data) (column 2 data) (column 3 data))
	    data
	    )
	  )
	 (data-sites (mapcar (lambda(x) (select x (iseq 2))) data))
	 (metric (if use-vario-metric variogram #'norm))
	 )
    (if verbose (print "Beginning estimation...okay?"))
    (mapcar
     (lambda(site)
       (let* (
	      (closest
	       (select
		(order 
		 (mapcar (lambda(x) (funcall metric (- site x))) data-sites)
		 )
		(iseq num)))
	      (estimate
	       (first
		(global-krig
		 (select data closest) (list site) variogram
		 :weights (if weights (select weights closest)))))
	      )
	 (if verbose (print estimate))
	 estimate
	 )
       )
     sites)
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 estimate-variogram
;;
(defun estimate-variogram(mat x y
			    &key
			    weights
			    (type 'gaussian)
			    range
			    sill
			    nugget
			    power
			    )
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; estimate-variogram Doc:    Creation Date:2/4/97        Author: Andy Long
;;
;; Description/Arguments:

       (estimate-variogram mat x y
			    &key
			    weights
			    (type 'gaussian)
			    range
			    sill
			    nugget
			    power
			    )
  
	This is my non-linear variogram estimator for two dimensions. It uses
newtonmax to fit a variogram model to a matrix.

	Mat and weights should be matrices. x and y are the corresponding x and
y values for the mat matrix. The output of the function vario should be
appropriate for input to this.

	Options for type are 'gaussian, 'exponential, 'spherical, 'linear, and
'power. If 'power is chosen, then the power option may be specified: otherwise
a default power of 1 is used.

;; Notes:

;; See Also: *-variogram

;; Examples:
 (new-plot)
 (def x (rand.u 100))
 (def y (rand.u 100))
 (def z (+ x y))

 ;; Compute the empirical variogram:
 (def var (vario (weave x y z) 10))

 ;; Guess the theoretical variogram:
 (def guess
   (estimate-variogram (first var) (third var) (fourth var) :weights (second var)))

 ;; Display the theoretical variogram:
 (def my-plot (matrix-and-data (first var) (third var) (fourth var) '((0) (0) (0))))
 (send my-plot :add-function
   #'(lambda(x y)
       (gaussian-variogram (list x y) (first guess) (second guess) (third guess)))
   (min (third var)) (max (third var))
   (min (fourth var)) (max (fourth var)) :num-points 7)

 ;; Use the theoretical variogram to estimates points on a grid:
 (def estimates
   (global-krig
     (weave x y z)
     (make-grid 0 1 10 0 1 10)
     #'(lambda(x) (gaussian-variogram x (first guess) (second guess) (third guess))))
   )

 ;; And display the results:
 (matrix-and-data
   (matrix '(10 10) (mapcar #'third estimates))
   (rseq 0 1 10) (rseq 0 1 10) (list x y z))
"
  (let* (
	 (sites (make-grid (min x) (max x) (length x) (min y) (max y) (length y)))
	 (ln20 (ln 20))
	 (nugget (if nugget nugget 0))
	 (sill (if sill sill
		 (if weights (sum (/ (* weights mat) (sum weights))) (mean mat))))
	 (range (if range range (+ (square (max (abs x))) (square (max (abs y))))))
	 (power (if (equalp type 'power) (if power power 1) nil))
	 (guess (if power (abs (list nugget sill range power))
		  (abs (list nugget sill range))
		  )
		)
	 )
    (printall "nugget sill range:" guess nil)
    (setq
     guess
     (newtonmax
      #'(lambda(z)
	  (let* (
		 (nugget (abs (first z)))
		 (sill (abs (second z)))
		 (range (abs (third z)))
		 (power (if (fourth z) (fourth z)))
		 (tmp
		  (outer-product
		   x y 
		   #'(lambda(x y)
		       (let ((z (list x y)))
			 (cond
			  ((equalp type 'gaussian)
			   (gaussian-variogram z nugget sill range)
			   )
			  ((equalp type 'exponential)
			   (exponential-variogram z nugget sill range)
			   )
			  ((equalp type 'spherical)
			   (spherical-variogram z nugget sill range)
			   )
			  ((equalp type 'linear)
			   (linear-variogram z nugget sill :range range)
			   )
			  ((equalp type 'power)
			   (power-variogram z nugget sill range power)
			   )
			  (t
			   (nugget-variogram z nugget sill range)
			   )
			  )
			 )
		       )
		   )
		  
		  )
		 )	       
	    (- (sum
		(if weights
		    (* weights (square (- tmp mat)))
		  (square (- tmp mat))
		  )
		)
	       )
	    )
	  )
      guess
      :verbose t
      )
     )
    (abs guess)
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 gaussian-variogram-2d
;;
(defun gaussian-variogram-2d(x nugget sill range &key (angle 0) (x-y-stretch 1))
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; gaussian-variogram Doc:    Creation Date:2/4/97        Author: Andy Logn
;;
;; Description/Arguments:

       (gaussian-variogram-2d x nugget sill range
		&key (angle 0) (x-y-stretch 1)) 
  
	Evaluates the gaussian-variogram variogram model at a point x (may be a
vector or scalar: it will be turned into a distance).

;; Notes:

;; See Also:

;; Examples:
 (new-plot)
 (spin-points 21)
 (plot #'(lambda(x y) (gaussian-variogram-2d (list x y) 0 1 1)))

 (spin-domain -3 3 -3 3)
 (plot #'(lambda(x y)
   (gaussian-variogram-2d (list x y) 0 1 1 :angle (/ pi 4) :x-y-stretch 3)))
"
  (let (
	(h (norm
	    (matprod (matrix '(2 2) (list x-y-stretch 0 0 (/ x-y-stretch)))
		     (orthogonal-2-by-2 angle) 
		     x)
	    )
	   )
	)
    (+
     (nugget-variogram h nugget)
     (* sill (- 1 (exp (- (square (/ h range))))))
     )
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 gaussian-variogram
;;
(defun gaussian-variogram(x nugget sill range &key (angle 0) (x-y-stretch 1))
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; gaussian-variogram Doc:    Creation Date:2/4/97        Author: Andy Long
;;
;; Description/Arguments:

       (gaussian-variogram x nugget sill range &key (angle 0) (x-y-stretch 1))
  
	Evaluates the gaussian-variogram variogram model at a point x (may be a
vector or scalar: it will be turned into a distance).
	
	Angle and x-y-stretch only work for 2-d. The angle represents a
rotation in the counter-clockwise sense; the x-y-stretch is the square-root of
the actual ratio of the rotated coordinates (i.e., if x-y-stretch is 2, then
the rotated x-direction is weighted 4 times the rotated y-direction.

;; Notes: It is vectorized for 2-d, handles 1-d

;; See Also:

;; Examples:
 (new-plot)
 (plot {(gaussian-variogram x 0 1 1) :dash t} 0 4)

 (new-plot)
 (spin-points 21)
 (plot #'(lambda(x y) (gaussian-variogram (list x y) 0 1 1)))

 (spin-domain -3 3 -3 3)
 (plot #'(lambda(x y)
   (gaussian-variogram (list x y) 0 1 1 :angle (/ pi 4) :x-y-stretch 3)))
"
  (let (
	(h 
	 (if (numberp x) (abs x)
	   (if (listp (first x))
	       (mapcar
		#'norm
		(mapcar
		 #'(lambda(x)
		     (matprod
		      (matrix '(2 2) (list x-y-stretch 0 0 (/ x-y-stretch)))
		      (orthogonal-2-by-2 angle) 
		      x))
		 x
		 ))
	     (if (= (length x) 2)
		 (norm
		  (matprod (matrix '(2 2) (list x-y-stretch 0 0 (/ x-y-stretch)))
			   (orthogonal-2-by-2 angle) 
			   x)
		  )
	       (norm x)))))
	)
    (+
     (if (listp h)
	 (mapcar #'(lambda(x) (nugget-variogram x nugget)) h)
       (nugget-variogram h nugget)
       )
     (* sill (- 1 (exp (- (square (/ h range))))))
     )
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 power-variogram
;;
(defun power-variogram
  (x nugget sill range power
     &key (angle 0) (x-y-stretch 1))
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; power-variogram Doc:    Creation Date:2/4/97        Author: Andy Long
;;
;; Description/Arguments:

  (power-variogram
  	x nugget sill range power
	&key (angle 0) (x-y-stretch 1))
  
	Evaluates the power-variogram variogram model at a point x (may be a
vector or scalar: it will be turned into a distance).

	Angle and x-y-stretch only work for 2-d. The angle represents a
rotation in the counter-clockwise sense; the x-y-stretch is the square-root of
the actual ratio of the rotated coordinates (i.e., if x-y-stretch is 2, then
the rotated x-direction is weighted 4 times the rotated y-direction.

;; Notes:

;; See Also:

;; Examples:
 (new-plot)
 (plot (lambda(x) (power-variogram x 0 1 1 1.5)) 0 4)

 (new-plot)
 (spin-points 21)
 (plot (lambda(x y) (power-variogram (list x y) 0 1 1 1.5)))

 (spin-domain -3 3 -3 3)
 (plot (lambda(x y)
   (power-variogram (list x y) 0 1 1 1.5 :angle (/ pi 4) :x-y-stretch 3)))

"
  (let (
	(h 
	 (if (numberp x) (abs x)
	   (if (listp (first x))
	       (mapcar
		#'norm
		(mapcar
		 #'(lambda(x)
		     (matprod
		      (matrix '(2 2) (list x-y-stretch 0 0 (/ x-y-stretch)))
		      (orthogonal-2-by-2 angle) 
		      x))
		 x
		 ))
	     (if (= (length x) 2)
		 (norm
		  (matprod (matrix '(2 2) (list x-y-stretch 0 0 (/ x-y-stretch)))
			   (orthogonal-2-by-2 angle) 
			   x)
		  )
	       (norm x)))))
	)
    (+
     (nugget-variogram h nugget)
     (* sill (^ (/ h range) power))
     )
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 linear-variogram
;;
(defun linear-variogram(x nugget sill &key range (angle 0) (x-y-stretch 1))
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; linear-variogram Doc:    Creation Date:2/4/97        Author: Andy Long
;;
;; Description/Arguments:

       (linear-variogram x nugget sill &key range (angle 0) (x-y-stretch 1))
  
	Evaluates the linear-variogram variogram model at a point x (may be a
vector or scalar: it will be turned into a distance).

	Angle and x-y-stretch only work for 2-d. The angle represents a
rotation in the counter-clockwise sense; the x-y-stretch is the square-root of
the actual ratio of the rotated coordinates (i.e., if x-y-stretch is 2, then
the rotated x-direction is weighted 4 times the rotated y-direction.

;; Notes:

;; See Also:

;; Examples:
 (new-plot)
 (plot {(linear-variogram x 0 1) :dash t} 0 4)

 (new-plot)
 (spin-points 21)
 ;; with a range:
 (plot #'(lambda(x y) (linear-variogram (list x y) 0 1)))
 ;; without a range, we get a cone:
 (plot #'(lambda(x y) (linear-variogram (list x y) 0 1)))
"
  (let (
	(h 
	 (if (numberp x) (abs x)
	   (if (listp (first x))
	       (mapcar
		#'norm
		(mapcar
		 #'(lambda(x)
		     (matprod
		      (matrix '(2 2) (list x-y-stretch 0 0 (/ x-y-stretch)))
		      (orthogonal-2-by-2 angle) 
		      x))
		 x
		 ))
	     (if (= (length x) 2)
		 (norm
		  (matprod (matrix '(2 2) (list x-y-stretch 0 0 (/ x-y-stretch)))
			   (orthogonal-2-by-2 angle) 
			   x)
		  )
	       (norm x)))))
	)
    (+
     (if (listp h)
	 (mapcar #'(lambda(x) (nugget-variogram x nugget)) h)
       (nugget-variogram h nugget)
       )
     (* sill (if range (min (/ h range) 1) h)))
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 logistic-variogram
;;
(defun logistic-variogram(x nugget sill range &key (angle 0) (x-y-stretch 1))
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; logistic-variogram Doc:    Creation Date:2/4/97        Author: Andy Long
;;
;; Description/Arguments:

       (logistic-variogram x nugget sill range &key (angle 0) (x-y-stretch 1))
  
	Evaluates the logistic-variogram variogram model at a point x (may be a
vector or scalar: it will be turned into a distance).

	Angle and x-y-stretch only work for 2-d. The angle represents a
rotation in the counter-clockwise sense; the x-y-stretch is the square-root of
the actual ratio of the rotated coordinates (i.e., if x-y-stretch is 2, then
the rotated x-direction is weighted 4 times the rotated y-direction.

;; Notes:

;; See Also:

;; Examples:
 (new-plot)
 (plot {(logistic-variogram x 0 1 1) :dash t} 0 4)

 (new-plot)
 (spin-points 21)
 ;; with a range:
 (plot #'(lambda(x y) (logistic-variogram (list x y) 0 1 1)))
"
  (let (
	(h 
	 (if (numberp x) (abs x)
	   (if (listp (first x))
	       (mapcar
		#'norm
		(mapcar
		 #'(lambda(x)
		     (matprod
		      (matrix '(2 2) (list x-y-stretch 0 0 (/ x-y-stretch)))
		      (orthogonal-2-by-2 angle) 
		      x))
		 x
		 ))
	     (if (= (length x) 2)
		 (norm
		  (matprod (matrix '(2 2) (list x-y-stretch 0 0 (/ x-y-stretch)))
			   (orthogonal-2-by-2 angle) 
			   x)
		  )
	       (norm x)))))
	)
    (+
     (nugget-variogram h nugget)
     (* sill (- 1 (/ (+ 1 (exp (/ h range))))))
     )
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 nugget-variogram
;;
(defun nugget-variogram(x nugget &rest args)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; nugget-variogram Doc:    Creation Date:2/4/97        Author: Andy Long
;;
;; Description/Arguments:

       (nugget-variogram x nugget &rest args)
  
	Evaluates the nugget-variogram variogram model at a point x (may be a
vector or scalar: it will be turned into a distance).

;; Notes:

;; See Also:

;; Examples:
 (new-plot)
 (plot {(nugget-variogram x .5 1 1) :dash t} 0 4)

 (new-plot)
 (spin-points 21)
 ;; with a range:
 (plot #'(lambda(x y) (nugget-variogram (list x y) .5 1 1)))
 ;; without a range, we get a cone:
 (plot #'(lambda(x y) (nugget-variogram (list x y) .5 1)))
"
  (* nugget (heaviside (norm x)))
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 exponential-variogram
;;
(defun exponential-variogram(x nugget sill range &key (angle 0) (x-y-stretch 1))
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; exponential-variogram Doc:    Creation Date:2/4/97        Author: Andy Long
;;
;; Description/Arguments:

       (exponential-variogram x nugget sill range &key (angle 0) (x-y-stretch 1))
  
	Evaluates the exponential-variogram variogram model at a point x (may
be a vector or scalar: it will be turned into a distance).

	Angle and x-y-stretch only work for 2-d. The angle represents a
rotation in the counter-clockwise sense; the x-y-stretch is the square-root of
the actual ratio of the rotated coordinates (i.e., if x-y-stretch is 2, then
the rotated x-direction is weighted 4 times the rotated y-direction.

;; Notes:

;; See Also:

;; Examples:
 (new-plot)
 (plot (lambda(x) (exponential-variogram x 0 1 1)) 0 4)

 (new-plot)
 (spin-points 21)
 (plot (lambda(x y) (exponential-variogram (list x y) 0 1 1)))
"
  (let (
	(h 
	 (if (numberp x) (abs x)
	   (if (listp (first x))
	       (mapcar
		#'norm
		(mapcar
		 #'(lambda(x)
		     (matprod
		      (matrix '(2 2) (list x-y-stretch 0 0 (/ x-y-stretch)))
		      (orthogonal-2-by-2 angle) 
		      x))
		 x
		 ))
	     (if (= (length x) 2)
		 (norm
		  (matprod (matrix '(2 2) (list x-y-stretch 0 0 (/ x-y-stretch)))
			   (orthogonal-2-by-2 angle) 
			   x)
		  )
	       (norm x)))))
	)
    (+
     (nugget-variogram h nugget)
     (* sill (- 1 (exp (- (/ h range)))))
     )
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 spherical-variogram
;;
(defun spherical-variogram(x nugget sill range &key (angle 0) (x-y-stretch 1))
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; spherical-variogram Doc:    Creation Date:2/4/97        Author: Andy Long
;;
;; Description/Arguments:

       (spherical-variogram x nugget sill range &key (angle 0) (x-y-stretch 1))
  
	Evaluates the spherical-variogram variogram model at a point x (may be
a vector or scalar: it will be turned into a distance).

	Angle and x-y-stretch only work for 2-d. The angle represents a
rotation in the counter-clockwise sense; the x-y-stretch is the square-root of
the actual ratio of the rotated coordinates (i.e., if x-y-stretch is 2, then
the rotated x-direction is weighted 4 times the rotated y-direction.

;; Notes:

;; See Also:

;; Examples:
 (new-plot)
 (plot (lambda(x) (spherical-variogram x 0 1 1)) 0 1.5)

 (new-plot)
 (spin-points 21)
 (plot (lambda(x y) (spherical-variogram (list x y) 0 1 1)))
"
  (let* (
	(h 
	 (if (numberp x) (abs x)
	   (if (listp (first x))
	       (mapcar
		#'norm
		(mapcar
		 #'(lambda(x)
		     (matprod
		      (matrix '(2 2) (list x-y-stretch 0 0 (/ x-y-stretch)))
		      (orthogonal-2-by-2 angle) 
		      x))
		 x
		 ))
	     (if (= (length x) 2)
		 (norm
		  (matprod (matrix '(2 2) (list x-y-stretch 0 0 (/ x-y-stretch)))
			   (orthogonal-2-by-2 angle) 
			   x)
		  )
	       (norm x)))))
	 (h/r (/ h range))
	 )
    (+ (nugget-variogram h nugget)
       (* sill (if (< h range) (* .5 h/r (- 3 (* h/r h/r))) 1))
       )
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 test-geostatistics
;;
(defun test-geostatistics(&key (f #'*) (type 'gaussian))
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; test-geostatistics Doc:    Creation Date:2/5/97        Author: Andy Long
;;
;; Description/Arguments:

       (test-geostatistics &key (f #'*) (type 'gaussian))
  
	Test-geoestatistics merely shows off the geostatistical techniques of
variogram fitting and kriging on a small, non-noisy data set.

	Plots the sample variogram against the theoretical variogram, and 

;; See Also:

;; Examples:
 (test-geostatistics)
 (test-geostatistics :f #'+)
 (test-geostatistics :f #'(lambda(x y) (norm (list x y))))
 (test-geostatistics :f #'(lambda(x y) (norm (list x y))) :type 'spherical)
"
  (new-plot)
  (let* (
	 (x (rand.u 100))
	 (y (rand.u 100))
	 (z (mapcar #'(lambda(x) (apply f x)) (weave x y)))
	 var guess my-plot
	 )
    ;; Compute the empirical variogram:
    (setq var (vario (weave x y z) 10))
    
    ;; Guess the theoretical variogram:
    (setq
     guess
     (estimate-variogram
      (first var) (third var) (fourth var) :weights (second var)
      :type type
      )
     )
    
    ;; Display the theoretical variogram:
    (setq my-plot
	  (matrix-and-data
	   (first var) (third var) (fourth var) '((0) (0) (0))))
    (send my-plot :add-function
	  #'(lambda(x y)
	      (cond
	       ((eq type 'gaussian)
		(gaussian-variogram
		 (list x y) (first guess) (second guess) (third guess))
		)
	       ((eq type 'spherical)
		(spherical-variogram
		 (list x y) (first guess) (second guess) (third guess))
		)
	       ((eq type 'exponential)
		(exponential-variogram
		 (list x y) (first guess) (second guess) (third guess))
		)
	       ((eq type 'linear)
		(linear-variogram
		 (list x y) (first guess) (second guess) (third guess))
		)
	       ((eq type 'logistic)
		(logistic-variogram
		 (list x y) (first guess) (second guess) (third guess))
		)
	      ))
	  (min (third var)) (max (third var))
	  (min (fourth var)) (max (fourth var)) :num-points 7)
    
    ;; Use the theoretical variogram to estimates points on a grid:
    (setq estimates
	  (global-krig
	   (weave x y z)
	   (make-grid 0 1 10 0 1 10)
	   #'(lambda(x) (gaussian-variogram x (first guess) (second guess) (third guess))))
	  )
    
    ;; And display the results:
    (setq my-plot
	  (matrix-and-data
	   (matrix '(10 10) (mapcar #'third estimates))
	   (rseq 0 1 10) (rseq 0 1 10) (list x y z)))
    
    (send my-plot :add-function #'(lambda(x y) (apply f (list x y))) 0 1 0 1
	  :num-points 21)
    my-plot
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                 estimate-anisotropic-variogram
;;
(defun estimate-anisotropic-variogram(mat x y
					  &key
					  weights
					  (type 'gaussian)
					  range
					  sill
					  nugget
					  power
					  scale
					  (x-y-stretch 1.01)
					  (angle 0)
					  (count-limit 20)
					  )
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; estimate-anisotropic-variogram Doc:  Creation Date:2/4/97   Author: Andy Long
;;
;; Description/Arguments:

       (estimate-anisotropic-variogram mat x y
					  &key
					  weights
					  (type 'gaussian)
					  range
					  sill
					  nugget
					  power
					  scale
					  (x-y-stretch 1.01)
					  (angle 0)
					  (count-limit 20)
					  )
  
	This is my non-linear variogram estimator for two dimensions. It uses
 newtonmax to fit a variogram model to a matrix. The result is a list of

	Output is a list of the 

		angle x-y-stretch nugget sill range

 (and power, if a power model is used).


	Mat and weights should be matrices. x and y are the corresponding x and
 y values for the mat matrix. The output of the function vario should be
 appropriate for input to this.

	If you start with an x-y-stretch of 1, it will be changed to 1.01:
 otherwise, the angle has no effect, which means that newtonmax will crash (the
 numerical hessian matrix will be singular). If, in fact, the x-y-stretch is 1,
 the program will converge to 1.

	Options for type are 'gaussian, 'exponential, 'spherical, 'linear, and
 'power. If 'power is chosen, then the power option may be specified: otherwise
 a default power of 1 (linear) is used.

	If scale is true, the initial guess will be used as a scale to
 newtonmax.

;; Notes:

;; See Also: type the command (about -variogram)
   
;; Examples:
 (new-plot)
 (def x (rand.u 100))
 (def y (rand.u 100))
 (def z (* x (* .5 y)))

 ;; Compute the empirical variogram:
 (def var (vario (weave x y z) 10))
 (plot (first var))

 ;; Guess the theoretical variogram:
 (def guess
   (estimate-anisotropic-variogram
     (first var) (third var) (fourth var) :x-y-stretch 2 :count-limit 10))

 ;; Display the theoretical variogram:
 (def my-plot
   (matrix-and-data (first var) (third var) (fourth var) '((0) (0) (0))))

 (send my-plot :add-function
   #'(lambda(x y)
       (gaussian-variogram
         (list x y) (third guess) (fourth guess) (fifth guess)
                   :angle (first guess) :x-y-stretch (second guess)))
   (min (third var)) (max (third var))
   (min (fourth var)) (max (fourth var)) :num-points 7)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;     Fix the game: give a known model, and see if we find it.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 (def fixed (outer-product (third var) (fourth var)
   #'(lambda(x y)
     (gaussian-variogram
       (list x y) .1 1 1 
           :angle (/ pi 4) :x-y-stretch 2))))

 ;; Guess the theoretical variogram:
 (def guess
   (estimate-anisotropic-variogram
     fixed (third var) (fourth var) :x-y-stretch 1.8 :angle .5 :count-limit 10))

 ;; If you want to take a few more iterations with that guess, then
 (def guess
   (estimate-anisotropic-variogram
     fixed (third var) (fourth var)
       :angle (first guess)
       :x-y-stretch (second guess)
       :nugget (third guess)
       :sill (fourth guess)
       :range (fifth guess)
       :count-limit 10))

 ;; Display the theoretical variogram:
 (def my-plot
   (matrix-and-data fixed (third var) (fourth var) '((0) (0) (0))))
 (send my-plot :add-function
   #'(lambda(x y)
       (gaussian-variogram
         (list x y) (third guess) (fourth guess) (fifth guess)
                   :angle (first guess) :x-y-stretch (second guess)))
   (min (third var)) (max (third var))
   (min (fourth var)) (max (fourth var)) :num-points 7)
"
  (let* (
	 (sites
	  (make-grid (min x) (max x) (length x) (min y) (max y) (length y)))
	 (ln20 (ln 20))
	 (nugget (if nugget nugget 0))
	 (sill (if sill sill
		 (if weights
		     (sum (/ (* weights mat) (sum weights)))
		   (mean mat))))
	 (range (if range range
		  (+ (square (max (abs x))) (square (max (abs y))))))
	 (power (if (equalp type 'power) (if power power 1) nil))
	 (x-y-stretch (if (= x-y-stretch 1) (+ x-y-stretch .01) x-y-stretch))
	 (guess (if power (abs (list angle x-y-stretch nugget sill range power))
		  (abs (list angle x-y-stretch nugget sill range))
		  )
		)
	 )
    (printall "angle x-y-stretch nugget sill range:" guess nil)
    (setq
     guess
     (newtonmax
      #'(lambda(z)
	  (let* (
		 crit
		 (z (abs z))
		 (angle (elt z 0))
		 (x-y-stretch (elt z 1))
		 (nugget (elt z 2))
		 (sill (elt z 3))
		 (range (elt z 4))
		 (power (if (sixth z) (elt z 5)))
		 (tmp
		  (outer-product
		   x y
		   #'(lambda(x y)
		       (let ((z (list x y)))
			 (cond
			  ((equalp type 'gaussian)
			   (gaussian-variogram z nugget sill range
					       :angle angle
					       :x-y-stretch x-y-stretch)
			   )
			  ((equalp type 'exponential)
			   (exponential-variogram z nugget sill range
					       :angle angle
					       :x-y-stretch x-y-stretch)
			   )
			  ((equalp type 'spherical)
			   (spherical-variogram z nugget sill range
					       :angle angle
					       :x-y-stretch x-y-stretch)
			   )
			  ((equalp type 'linear)
			   (linear-variogram z nugget sill :range range
					       :angle angle
					       :x-y-stretch x-y-stretch)
			   )
			  ((equalp type 'power)
			   (power-variogram z nugget sill range power
					       :angle angle
					       :x-y-stretch x-y-stretch)
			   )
			  (t
			   (nugget-variogram z nugget sill range)
			   )
			  )
			 )
		       )
		   )
		  
		  )
		 )	       
	    (setq crit
		  (- (sum
		      (if weights
			  (* weights (square (- tmp mat)))
			(square (- tmp mat))
			)
		      )
		     )
		  )
	    (andy-print z)
	    crit
	    )
	  )
      guess
      :scale (if scale guess nil)
      :verbose 2
      :count-limit count-limit
      )
     )
    (abs guess)
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 vario-binary
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun vario-binary(data grid-size &key verbose weights xdiff ydiff)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; vario-binary Doc:    Creation Date:2/7/97        Author: Andy Long
;;
;; Description/Arguments:

	(vario-binary data grid-size &key verbose weights xdiff ydiff)

	This works in two dimensions ONLY. It assumes that you're going to give
it a list of triplets, the third of which is binary (0,1) data, and the size of
the (square) grid for which you want the variogram (it will be even, one more
than any odd you give it).

;; Notes:

;; See Also: vario, variogram

;; Examples:
 (def x (rand.u 1000))
 (def y (rand.u 1000))
 (def z (mapcar (lambda(x) (if (< (first (rand.u 1)) x) 0 1)) (* x y)))
 (mvsq (list \"x\" \"y\")
   (row-list (matmult (select (x-y-rotation (/ pi -3)) (iseq 2) (iseq 2))
			 (bind-rows x y))
		))
 (def x (coerce x 'list))
 (def y (* 3 (coerce y 'list)))
 (def vardata (vario-binary (weave x y z) 5 :verbose 100))
 (plot (/ (first vardata) (1+ (second vardata))))

 (def x (rand.u 1000))
 (def y (rand.u 1000))
 (def z (mapcar (lambda(x) (if (< (first (rand.u 1)) x) 0 1)) (* x y)))
 (def y (* 3 y))
 (def vardata (vario-binary (weave x y z) 5 :verbose 100))
 (plot (/ (first vardata) (1+ (second vardata))))

 (def x (rand.u 1000))
 (def y (rand.u 1000))
 (def z (mapcar (lambda(x) (if (< (first (rand.u 1)) x) 0 1)) (square (- x y))))
 (def vardata (vario-binary (weave x y z) 5 :verbose 100))
 (plot (/ (first vardata) (1+ (second vardata))))


 (def x (rand.u 100))
 (def y (rand.u 100))
 (def z (mapcar (lambda(x) (if (< (first (rand.u 1)) x) 0 1)) (* x y)))
 (def vardata (vario-binary (weave x y z) 5 :verbose 100))
 (plot (/ (first vardata) (1+ (second vardata))))

 (def weights (1+ (sample (iseq 5) 100 t)))
 (def z (mapcar #'(lambda(x) (first (/ (sample (iseq 1 x) 1) x))) weights))
 (def vardata (vario-binary (weave x y z) 10 :weights weights :verbose t))

 (def oldvar
     (vario
      (smoosh
       (mapcar
	#'(lambda(x1 x2 x3 w)
	   (weave
	     (repeat x1 w)
	     (repeat x2 w)
	     (append
		(repeat 0 (floor (- w (* w x3))))
		(repeat 1 (floor (* w x3)))
		)
	      )
	    )
	x y z weights
	)
       )                    
      10
      :verbose t
      )
     )

 (plot (- (/ (first oldvar) (1+ (second oldvar)))
       (/ (first vardata) (1+ (second vardata)))))
"
  (let* (
	 (grid-size (if (oddp grid-size) grid-size (1+ grid-size)))
	 (half-grid-size (/ (1- grid-size) 2))
	 (xvals (mapcar #'butlast data))
	 (xs (mapcar #'first xvals))
	 (ys (mapcar #'second xvals))
	 (zs (mapcar #'final data))
	 (len (length zs))
	 (1-len (1- len))
	 (weights (if weights weights (repeat 1 len)))
	 (xmin (min xs))
	 (xmax (max xs))
	 (ymin (min ys))
	 (ymax (max ys))
	 (xdiff (if xdiff xdiff (/ (- xmax xmin) 2)))
	 (ydiff (if ydiff ydiff (/ (- ymax ymin) 2)))
	 (xfactor (/ grid-size xdiff))
	 (yfactor (/ grid-size ydiff))
	 (grid
	  (make-array (list (1+ grid-size) (1+ (* 2 grid-size)))
		      :initial-element 0))
	 (nums (make-array (size grid) :initial-element 0))
	 (data (mapcar #'combine (weave data weights)))
	 (verbose (if verbose (if (numberp verbose) verbose 1)) nil)
	 (i 0)
	 )
    (dolist
     (piece data)
     (if verbose (let ((tmp (/ i verbose))) (if (= tmp (floor tmp)) (print i))))
     (let (
	   (x (elt piece 0))
	   (y (elt piece 1))
	   (z (elt piece 2))
	   (w1 (elt piece 3))
	   )
       (dolist
	(newpiece (select data (iseq i 1-len)))
	(let* (
	       (xx (elt newpiece 0))
	       (yy (elt newpiece 1))
	       (xdist (- x xx))
	       (abs-xdist (abs xdist))
	       (ydist (- y yy))
	       )
	  (if (and (< abs-xdist xdiff) (< (abs ydist) ydiff))
	      (let* (
		     (zz (elt newpiece 2))
		     (w (* w1 (elt newpiece 3)))
		     (xd (round (* abs-xdist xfactor)))
		     (yd (+ grid-size
			    (round (* (if (< xdist 0) -1 1) ydist yfactor))))
		     )
		(setf (aref grid xd yd)
		      (+ (aref grid xd yd) (* w (+ z zz (- (* 2 z zz))))))
		(setf (aref nums xd yd) (+ (aref nums xd yd) w))
		)
	    )
	  )
	)
       )
     (setq i (1+ i))
     )
    (list
     grid nums
     (rseq 0 xdiff (1+ grid-size))
     (rseq (- ydiff) ydiff (1+ (* 2 grid-size)))
     )
    )
  )

#|
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                vario-test
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(let ()
  (def x (rand.n 10))
  (def y (rand.n 10))
  (def z (mapcar (lambda(x) (if (< (first (rand.u 1)) x) 0 1))
		 (square (- x y))))
  
  (time (def var (vario (weave x y z) 10)))
  (mapcar #'size (list x y z))
  (time (def vardata (vario-binary (weave x y z) 10)))
  (mapcar #'size (list x y z))
  (max (abs (- (second var) (second vardata))))
  (max (abs (- (first var) (first vardata))))
  (sum (first var))
  (sum (first vardata))
  (lines on)
  (plot (/ (first var) (1+ (second var))))
  (mapcar #'size (list x y z))
  (plot (/ (first vardata) (1+ (second vardata))))
  (mapcar #'size (list x y z))
  (lines off)(plot x y z)

  (def x (rand.n 1000))
  (def y (rand.n 1000))
  (def z (mapcar (lambda(x) (if (< (first (rand.u 1)) x) 0 1))
		 (square (- x y))))
  (time (def var (vario (weave x y z) 10)))
  (mapcar #'size (list x y z))
  (time (def vardata (vario-binary (weave x y z) 10)))
  (mapcar #'size (list x y z))
  (max (abs (- (second var) (second vardata))))
  (max (abs (- (first var) (first vardata))))
  (sum (first var))
  (sum (first vardata))
  (lines on)
  (plot (/ (first var) (1+ (second var))))
  (mapcar #'size (list x y z))
  (plot (/ (first vardata) (1+ (second vardata))))
  (mapcar #'size (list x y z))
  (lines off)(plot x y z)

  (def x (rand.n 1000))
  (def y (rand.n 1000))
  (def z (mapcar (lambda(x) (if (< (first (rand.u 1)) x) 0 1))
		 (square (- x y))))
		   
  (time (def var (vario-binary (weave x y z) 10)))
  (lines on)(plot (/ (first var) (1+ (second var))))
  (lines off)(plot x y z)

  (def x (rand.u 1000))
  (def y (rand.u 1000))
  (def z (square (- x y)))
  (time (def var (vario (weave x y z) 30)))
  (lines on)(plot (/ (first var) (1+ (second var))))
  (lines off)(plot x y z)
  )
|#
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                            kernel-smoother
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun kernel-smoother(data grid function &optional weights)
  "
	(kernel-smoother data grid function &optional weights)

	This works in all dimensions, but you'll have to do a little
more work to turn your results into something reasonable. It assumes
that you're going to give it a list of tuples (although in one
dimension give just real numbers), and a list of grid locations
where you want data. Then you'll use function to smooth the data.
There is no default smoother! You need to make that choice, depending
on locations, etc.

	Be careful! This is an isotropic kernel smoother: that means that it
merely computes the norm between estimation site and data, and applies the
kernel to that distance. If you have an anisotropic model, use
kernel-smoother-with-direction.

	You can do several variables at a time, to save redoing the
calculations over and over.

;; See also: kernel-smoother-with-direction

;; Examples:
 (def trial-vals
     (list
      0.17049671033630368
      0.3350172603792924
      0.4191022411016531
      0.4204952566531957
      0.675180331972313
      0.780747818448824
      0.7939017775692289
      0.8627097570774231
      0.9525248799120555
      0.977265843668132)
     )
;; one-dimension:
 (lines off)
 (plot
  (kernel-smoother
   (weave trial-vals (mapcar #'square trial-vals))
   (rseq -1 2 10)
   #'(lambda(x) (exp (- (^ x 2))))
  )
 )
;; two-dimensions:
  (lines on)
  (plot
   (list
    (rseq -1 2 20)
    (rseq -1 2 20)
    (kernel-smoother
     (weave trial-vals trial-vals (mapcar #'square trial-vals))
     (weave (rseq -1 2 20) (rseq -1 2 20))
     #'(lambda(x) (exp (- (^ x 2))))
     )
    )
   )
;; one-dimension, and several variables at a time:
 (lines off)
 (def both-ln-and-square
  (kernel-smoother
   (weave
     trial-vals
     (weave
       (mapcar #'ln trial-vals)
       (mapcar #'square trial-vals)
      )
     ) 
   (rseq -1 2 10)
   #'(lambda(x) (exp (- (^ x 2))))
  ))
 )
 (title \"ln versus estimates\")
 (plot (ln trial-vals) (mapcar #'first both-ln-and-square))
 (send my-plot :add-plot-item #'(lambda(x) x))
 (send my-plot :update-plot)
 (title \"square versus estimates\")
 (plot (square trial-vals) (mapcar #'second both-ln-and-square))
 (send my-plot :add-plot-item #'(lambda(x) x))
 (send my-plot :update-plot)
"
  (let* (
	 (newgrid (if (listp grid) grid (matrix-to-list grid)))
	 (xvals (mapcar #'butlast data))
	 (yvals (mapcar #'final data))
	 (bunch (if (listp (first yvals)) t nil))
	 (j 0)
	 (newstuff (make-array (length newgrid)))
	 new sum-new
	 )
    (if bunch
	(dolist
	 (i newgrid) ;; go over all array (grid) points:
	 (setq new ;; these are the weights, determined by function:
	       (mapcar
		function
		(mapcar #'(lambda(x) (norm (- i x))) xvals))
	       )
	 (setq sum-new ;; this is the sum of the weights, or weighted weights:
	       (if weights (sum (* weights new)) (sum new)))
	 (setf (aref newstuff j) ;; the j-th array element to be estimated:
	       (if (/= 0 sum-new)
		   (/ (mapcar #'sum (transpose (* yvals new))) sum-new)
		 nil
		 )
	       )
	 (setq j (1+ j)) ;; step the array referencing index.
	 )
      (dolist ;; otherwise we have just one variable to do:
       (i newgrid) ;; go over all array (grid) points:
       (setq new ;; these are the weights, determined by function:
	     (mapcar
	      function
	      (mapcar #'(lambda(x) (norm (- i x))) xvals))
	     )
       (setq sum-new ;; this is the sum of the weights, or weighted weights:
	     (if weights (sum (* weights new)) (sum new)))
       (setf (aref newstuff j) ;; the j-th array element to be estimated:
	     (if (/= 0 sum-new) (/ (sum (* yvals new)) sum-new) nil)
	     )
       (setq j (1+ j)) ;; step the array referencing index.
       )
      )
    (coerce newstuff 'list)
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;               kernel-smoother-with-direction
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun kernel-smoother-with-direction
  (data grid function
	&key
	weights
	verbose
	kernel-sum-to-one
	kernel-positive
	)
  "
 (kernel-smoother-with-direction
  	data grid function
	&key
	weights
	verbose
	kernel-sum-to-one
	kernel-positive
	)

	This works in all dimensions, but you'll have to do a little
more work to turn your results into something reasonable. It assumes
that you're going to give it a list of tuples (although in one
dimension give just real numbers), and a list of grid locations
where you want data. Then you'll use function to smooth the data.
There is no default smoother! You need to make that choice, depending
on locations, etc.

	You can do several variables at a time, to save redoing the
calculations over and over.

;; Examples:
"
  (let* (
	 (newgrid (if (listp grid) grid (matrix-to-list grid)))
	 (xvals (mapcar #'butlast data))
	 (yvals (mapcar #'final data))
	 (len (length yvals))
	 (bunch (if (listp (first yvals)) t nil))
	 (j 0)
	 (newstuff (make-array (length newgrid)))
	 (newweight (make-array (length newgrid)))
	 new sum-new 
	 )

    (if bunch
	(dolist
	 (i newgrid)

	 ;; get the weights, determined by function function:
	 (setq new (mapcar function (mapcar #'(lambda(x) (- i x)) xvals)))
	 (if kernel-positive (setq new (+ new (min new))))

	 ;; this is the sum of the weights, or weighted weights:
	 (setq sum-new (if weights (sum (* weights new)) (sum new)))

	 ;; If we want the kernel weights to sum to one, then kernel-sum-to-one
	 ;; should be turned on; otherwise we'll divided the weights by the sum
	 ;; of the weights:
	 (if kernel-sum-to-one
	     (let ()
	       (setq new (+ new (/ (- 1 sum-new) len)))
	       (setf (aref newstuff j) (mapcar #'sum
					       (transpose (* yvals new))))
	       )
	   (let ()
	     (setf (aref newweight j) sum-new)
	     (setf (aref newstuff j)
		   (if (= 0 sum-new)
		       (let ()
			 (printall "Sum of weights is zero. " i)
			 nil
			 )
		     (/ (mapcar #'sum (transpose (* yvals new))) sum-new)
		     )
		   )
	     )
	   )
	 ;; Let the user know wusup:
	 (if verbose (print (combine i (aref newstuff j))))
	 
	 ;; step the array referencing index.
	 (setq j (1+ j))
	 )
      
      ;; otherwise we have just one variable to do. 
      ;; Go over all array (grid) points:
      (dolist
       (i newgrid)
       
       ;; get the weights, determined by function function:
       (setq new (mapcar function (mapcar #'(lambda(x) (- i x)) xvals)))
       (if kernel-positive (setq new (+ new (min new))))
       
       ;; this is the sum of the weights, or weighted weights:
       (setq sum-new (if weights (sum (* weights new)) (sum new)))
       
       ;; If we want the kernel weights to sum to one, then kernel-sum-to-one
       ;; should be turned on; otherwise we'll divided the weights by the sum
       ;; of the weights:
       (if kernel-sum-to-one
	   (let ()
	     (setq new (+ new (/ (- 1 sum-new) len)))
	     (setf (aref newstuff j) (sum (* yvals new)))
	     )
	 (let ()
	   (setf (aref newweight j) sum-new)
	   (setf (aref newstuff j)
		 (if (= 0 sum-new)
		     (let ()
		       (printall "Sum of weights is zero. " i)
		       nil
		       )
		   (/ (sum (* yvals new)) sum-new)
		   )
		 )
	   )
	 )
       ;; Let the user know wusup:
       (if verbose (print (combine i (aref newstuff j))))

       ;; step the array referencing index, and back we go!
       (setq j (1+ j))

       )
      )
    (if kernel-sum-to-one
	(coerce newstuff 'list)
      (list (coerce newstuff 'list) (coerce newweight 'list))
      )
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
