;; add to manual: numerical-inverse-function
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;				numerical-inverse-function
;;
(defun numerical-inverse-function
  (f &key fp (maxits 10) initial bracket)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;			numerical-inverse-function Doc:
;;
;; Author: Andy Long
;; Created: Sun Mar 29 11:31:34 1998
;;
;; Description/Arguments: numerical-inverse-function

	This function returns a function which approximates the inverse of a
function f(x) by using Newton's method (if f' - fp - is provided), or the
secant method otherwise.

	If a bracketing interval is known for the function - i.e. the domain on
which the function inverse is defined - then newton-safe will be used, which
combines newton-raphson with bisection.

;; Arguments: f &key fp (maxits 10) initial bracket

	f	- the function whose inverse is desired;
	fp	- the derivative function f'.
	maxits	- maximum number of secant iterations, if fp is not defined.
	initial	- initial is an initial value or a list of two values,
		  depending on whether one is using Newton's or the secant
	 	  method (i.e., whether fp is defined). 
	bracket	- a list of two values, indicating an interval within which the
		  solution is known to lie.

;; Returns: a function (closure) which can be invoked as the inverse of f.

;; Notes: I created this so that I could easily create quantile functions given
;; the cumulative density function (cdf) f, and the density function
;; (derivative f').

;; See Also: newton-safe, newton-zero-value, secant2

;; Examples:
 (def numinv
      (numerical-inverse-function #'sin :fp #'cos :bracket (list -pi/2 pi/2)))
 (plot numinv #'asin .1 .9)

 (defun f(x) (- (funcall numinv x) (asin x)))
 (plot f)
 ;; compare newton-safe to secant to true:
 ;; (plot numinv (numerical-inverse-function #'sin) #'asin .1 .9)
 ;; compare newton-safe to newton to true:
 ;; (plot numinv (numerical-inverse-function #'sin :fp #'cos) #'asin .1 .9)
 ;; compare newton-safe to secant to true:
 (plot (numerical-inverse-function #'sin :bracket (list -pi/2 pi/2)) #'asin .1 .9)
"
  (if fp
      (if bracket
	  (lambda(p)
	    (newton-safe (lambda(x) (- (funcall f x) p))
			 fp
			 (first bracket)
			 (second bracket)
			 :maxits maxits))
	(lambda(p)
	  (newton-zero-value (lambda(x) (- (funcall f x) p))
			     fp
			     (if initial initial p)
			     :maxits maxits
			     ))
	)
    (if bracket
	(lambda(p)
	  (regula-falsi2
	   (lambda(x) (- (funcall f x) p))
	   (first bracket) (second bracket) :maxits maxits))
      (if initial
	  (lambda(p)
	    (secant2
	     (lambda(x) (- (funcall f x) p))
	     (first initial) (second initial)
	     :maxits maxits))
	(lambda(p)
	  (secant2 (lambda(x) (- (funcall f x) p)) (- p) p :maxits maxits)))
      )
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
