;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; muller-extrema Doc:    Creation Date:11/24/96        Author: Andy Long
;;
;;
(defun muller-extrema
  (f &key
     verbose guesses guess (scale 1) (maxits 100) (tol 1e-6) (print t)
     (return 'current)
     )
  "
;; Description/Arguments: function muller-extrema

  (f &key
     verbose guesses guess (scale 1) (maxits 100) (tol 1e-6) (print t)
     (return 'current)
     )
  
	maxits: the number of times it will iterate in seeking a solution.
	tol: the tolerance for changes: small change means we're done.
	verbose: if off, will not print the iterates; otherwise it will.
	guess: a single guess for the true solution (default: (0 0))
	guesses: a bunch of guesses ((p+2)(p+1)/2 of them!)
	scale: a real number guess for the distance about the guess of the
               solution, or a list (size p) of distances appropriate for each
	       dimension. 
        print: prints out info about the maximum and maximum values found
	return: value returned. Options: 'min, or 'max.

Muller-Extrema will iteratively find an extremum of functions of many
variables. The idea of Muller-Extrema is that it uses a generalized  paraboloid
to fit your function, and so generalizes the univariate technique known as the
secant method, which uses a line.

;; See Also: muller. This is beautiful stuff! No derivatives.

;; Examples:
;; You can give some guesses (you'll need (p+2)(p+1)/2 of them!)
 (data first-guesses (0 1) (1 0) (0 0) (1 1) (1 4) (3 2))
 (muller-extrema (func f(x y) (x - 3.33) ^ 2 + (sin (y - .123)) - 1)
           :guesses first-guesses :maxits 10)
 (muller-extrema (func f(x y) (x - 3.33) ^ 2 + (sin (y - .123)) - 1)
           :guess '(1 0))

;; or the code will automatically choose some values around (0 0)
 (muller-extrema (func g(x y) (x+1)^4 + (y-1)^4 + 1/(x^2*y^2+1))) 
 ;; (-1.39104 1.3910399)

;; compare this one to the ineffective demonstration of steep-path:
 (def xvals '(1 2 3))
 (def yvals '(1 1 3))
 (func f(m b) (sum ((yvals - (b + m * xvals)) ^ 2)))
 (muller-extrema #'f)

;; Higher dimensions:
 (muller-extrema (func f(m b c) (m*m)+(b*b)+(c*c)) :guess '(1 3 2))

 (def xvals (rand.n 100))
 (def yvals (rand.n 100))
 (def zvals (+ (* 3 xvals) (* 5 yvals) (rand.n 100)))
 (func f(m a b) (sum ((zvals - (m + a * xvals + b * yvals)) ^ 2)))
 (muller-extrema #'f :tol 1e-12 :scale '(1 3 5))
;; compare to the regression model:
 (regression-model (list xvals yvals) zvals)

 (def zvals (+ (* 3 xvals) (exp (* -1.3 yvals)) (rand.n 100)))
 (func f(m a b) (sum ((zvals - (m + a * xvals + (exp (b * yvals)))) ^ 2)))
 (muller-extrema #'f :tol 1e-6 :guess '(0 3 -1.3) :scale .1 :verbose t
   :return 'min)
 ;; I think that I'm having better luck with a higher tol here: seemed to be
 ;; getting lots of 'range errors' with 1e-12.
"
  (andy-print "muller-extrema")
  (let* (
	 ;; nvar obtains the multivariate dimension from f: if the lambda
	 ;; expression of f contains &key, &optional, etc. you'll get into
	 ;; trouble right here!
	 (nvar (length (second (get-lambda-expression f))))
	 (nguesses (/ (* (+ nvar 2) (+ nvar 1)) 2))
	 (guess (if guess guess (repeat 0 nvar)))
	 (guesses
	  (if guesses guesses
	    (mapcar
	     #'(lambda(x) (+ guess (* scale (rand.n nvar))))
	     (iseq nguesses)))
	  )
	 (xnot (first guesses)) ;; the most recent guess
	 (f-of-xnot (eval `(,f ,@xnot)))
	 (ftempmax f-of-xnot)
	 (ftempmin f-of-xnot)
	 (xtempmax xnot)
	 (xtempmin xnot)
	 (fvals (mapcar #'(lambda(x) (eval `(,f ,@x))) (rest guesses)))
	 newxval iterations
	 )
    (dotimes
     (i maxits)
     (let* (
	    (a-matrix
	     (combine
	      (mapcar #'(lambda(x) 
			  (let ((xdiff (- x xnot)))
			    (combine
			     xdiff
			     (muller-extrema-squared-terms xdiff)
			     )))
		      (rest guesses)
		      )))
	    ;; If you get a "range error", it's probably here in the solve
	    ;; function: I've gotten them here before....
	    (params
	     (solve
	      (matrix (list (1- nguesses) (1- nguesses)) a-matrix)
	      (- fvals f-of-xnot)
	      )
	     )
	    (derivs (select params (iseq nvar)))
	    (derivs2 (select params (iseq nvar (1- (length params)))))
	    )
       (setq
	newxval
	(+ xnot (solve
		 (fill-symmetric-matrix-with-upper-half derivs2)
		 (- derivs)
		 )
	   )
	iterations (1+ i)
	)
       ;; If we're not changing much, bolt:
       (if (> tol (/ (norm (- xnot newxval)) (+ tol (norm xnot))))
	   (return newxval)
	 ;; otherwise, let's calculate a function value:
	 (setq
	  fvals (adjoin f-of-xnot (butlast fvals))
	  xnot newxval
	  f-of-xnot (eval `(,f ,@xnot))
	  guesses (adjoin newxval (butlast guesses))
	  )
	 )
       ;; Check if we've set any new records (mins/maxs):
       (if (> f-of-xnot ftempmax)
	   (setq ftempmax f-of-xnot xtempmax xnot)
	 (if (< f-of-xnot ftempmin)
	       (setq ftempmin f-of-xnot xtempmin xnot)
	   )
	 )
       ;; If we're supposed to be talkative, give a report:
       (if verbose
	   (format t "~%~e, after ~d iterations, with f=~e." 
		   newxval i f-of-xnot)
	 )
       )
     )
    (setq f-of-xnot (eval `(,f ,@newxval)))
    (if print
	(format t "
Extremum: ~e
after ~d iterations.~%
Current value of f obtained: ~g
at: ~g
Maximum value of f obtained: ~g
at: ~g
Minimum value of f obtained: ~g
at: ~g~%~%"
		newxval iterations f-of-xnot newxval ftempmax xtempmax ftempmin
		xtempmin)
      )
    ;; Return one of the three special values (current, min, or max):
    (case return
	  ('min (values xtempmin ftempmin))
	  ('max (values xtempmax ftempmax))
	  (t (values newxval f-of-xnot))
	  )
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Description for muller-extrema-squared-terms:
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun muller-extrema-squared-terms(x)
  "
;; Documentation:
	Helper for muller-extrema

;; Examples:
 (muller-extrema-squared-terms (iseq 3))
" 
  (combine
   (mapcar
    #'(lambda(y)
	(let (
	      (x (first y))
	      (rest (rest y))
	      )
	  (if rest (combine (* .5 x x) (* x rest)) (list (* .5 x x)))))
    (mapcon 'list x))))
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                fill-symmetric-matrix-with-upper-half
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun fill-symmetric-matrix-with-upper-half(x)
  "
;; Description: fill-symmetric-matrix-with-upper-half(x)

;; Examples:
 (fill-symmetric-matrix-with-upper-half (list 1 0 0 1 0 1)) ;; (id 3)
 (fill-symmetric-matrix-with-upper-half (list 1 0 0 0 1 0 0 1 0 1)) ;; (id 4)
"
  (let* (
	 (lenx (length x))
	 ;; lenx = n*(n-1)/2 + n = n*(n+1)/2;
	 ;; solve for n, and round to make sure it's an integer:
	 (n (round (half (1- (sqrt (1+ (* 8 lenx)))))))
	 (matrix (make-array (list n n)))
	 (k 0)
	 )
    (dotimes
     (i n)
     (dolist
      (j (iseq i (1- n)))
      (setf (aref matrix i j) (elt x k))
      (setf (aref matrix j i) (elt x k))
      (setq k (1+ k))
      )
     )
    matrix
    )
  )
