;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                orbitals
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun orbitals
  (f
   initial
   end
   nsteps
   ntimes
   &key 
   (integrator #'euler-nd) 
   randomize
   randomize-about-initial
   (randomizer #'uniform-rand)
   title 
   old-window
   )
  "
;; Description:

 (orbitals
   f				; the vector derivative function
   initial			; the initial condition
   end				; the end time
   nsteps			; the number of steps the integrator should take
   ntimes			; the number of orbits

   &key 
   (integrator #'euler-nd) 	; the integrator to use
   randomize			; a randomization multiplier
   randomize-about-initial	; randomize about the initial value given
   (randomizer #'uniform-rand)	; a randomization function
   title 			; title string for the plot
   old-window			; if you want to use an existing plot, set
				; this to the plot. Useful if you wish to plot
				; several orbits of your choosing on the same
				; plot.
   )

 Orbitals plots orbits of three-d systems (or the first three coordinates
 of higher-dimensional systems), such as the Rossler or the Lorenz equations,
 using Euler's method as the default integrator.

 It does so in a series of steps, from time 0 to time end: give it
 reasonable values of nsteps for a period from 0 to end, based on the
 performance of the integrator.

 If you use randomize, it will choose ntimes different initial conditions from
 a uniform distribution (by default) and scaled by randomize. Otherwise, it
 will continue on from the last point in the previous orbit, drawing a single
 trajectory.

 In principle, if you want randomization you should set randomize; however, if
 you change the randomizer from the uniform distribution, it is assumed that
 you want randomization!

 The last point in the orbit is represented by a circle, to indicate the
 direction of the trajectory.

;; See also: euler-nd, rk-4, phase-plane, slope-field

;; Examples:

;; Rossler equations:
 (defun rossler (vals)
   (let ((x (first vals))(y (second vals))(z (third vals)))
     (list
	(- (+ y z))
	(+ x (* .2 y))
	(+ .2 (* z (- x 5.7))))))

 (def initial (* 10 (uniform-rand 3)))

 ;; a single orbit 10 time units, in 500 steps, using runge-kutta-4 as the
 ;; integrator:
 (orbitals #'rossler initial 10 500 1 :integrator #'rk-4)

 ;; a single orbit (randomize is 0 by default). We follow the system for 50
 ;; time units:
 (orbitals #'rossler initial 1 50 10 :integrator #'rk-4)

 ;; a collection of 10 long orbits:
 (orbitals #'rossler initial 4 200 10 :randomize 10 :title \"Rossler\")

 ;; a collection of 20 short orbits:
 (orbitals #'rossler initial .1 10 20 :randomize 20)

 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; from Perko's book 'Differential Equations and Dynamical Systems':
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 (defun page217(vals)
  (let (
	(x (elt vals 0))
	(y (elt vals 1))
	(z (elt vals 2))
	)
    (list 
     (+ (- y) (* x (square z)))
     (+ x (* y (square z)))
     (* (- z) (+ (square x) (square y)))
     )))
 (orbitals #'page217 (list 0 -1 0) 10 500 1 :randomize 0 :title \"page 217\"
	   :integrator #'rk-4)
 ;; in this case, I want to add my own orbits, based on initial conditions
 ;; occurring on the unit sphere: thus, I can use the :old-window option.
 ;; Notice that plot my-plot is automatically defined by orbitals.
 (dotimes (i 9)
	  (orbitals #'page217 (normalize (* '(1 1 2) (rand.n 3))) 10 500 1
		    :integrator #'rk-4 :old-window my-plot))

 ;; alternatively, I could have defined a new randomizer, and used that option
 ;; instead: it's a little easier actually in this case.
 (orbitals #'page217 (list 0 -1 0) 10 500 10
	   :randomizer (lambda(i) (normalize (* '(1 1 2) (rand.n i))))
	   :integrator #'rk-4)
"
  (let (
	(len (length initial))
	(randomize (if randomize (if (numberp randomize) randomize 1)))
	(randomize (if (neq randomizer #'uniform-rand)
		       (if randomize randomize 1)
		     randomize))
	(nextpoint initial)
	points xvals yvals zvals
	)

    ;; for each orbit, we call our integrator:
    (dotimes
     (i ntimes)
     (setq 
      points (funcall integrator f nextpoint 0 end nsteps)
      xvals (mapcar #'first points)
      yvals (mapcar #'second points)
      zvals (mapcar #'third points)
      nextpoint (if randomize
		    (* randomize (funcall randomizer len))
		  (final points))
      nextpoint (if randomize-about-initial (+ initial nextpoint) nextpoint)
      )

     ;; if we have a two-d system, for example, then the z value is faked:
     (if (not (first zvals)) (setq zvals (repeat 0 (length zvals))))

     ;; if we haven't a window to plot in, then create my-plot:
     (if (not old-window)
	 (let()
	   (def my-plot (spin-plot (list xvals yvals zvals)))
	   (setq old-window my-plot)
	   (if title (send old-window :title title))
	   (send old-window :depth-cuing nil)
	   (send old-window :showing-axes nil)
	   (send old-window :mouse-mode 'hand-rotate)
	   )
       )

     ;; add in the current lines, and the last point:
     (send old-window :add-lines (list xvals yvals zvals))
     (send old-window :add-points (list (last xvals) (last yvals) (last zvals)))

     ;; adjust to data, so that everything is in view:
     (send old-window :adjust-to-data)
     )

    ;; return the plot:
    old-window
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                phase-plane
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun phase-plane
  (f
   initial
   end
   nsteps
   ntimes
   &key 
   show
   (integrator #'euler-nd) 
   randomize
   randomize-about-initial
   (randomizer #'uniform-rand)
   title 
   old-window
   )
  "
			Phase plane program:

 (phase-plane
   f				; right-hand side of differential equation
   initial			; initial x values
   end				; end time
   nsteps			; number of steps from 0 to end
   ntimes			; number of orbits
   &key 
   show				; adjusts after the addition of each path -
				; slow! 
   (integrator #'euler-nd) 	; type of integrator
   randomize			; a numerical multiplier of the random values
   randomize-about-initial	; adds random data to initial value
   (randomizer #'uniform-rand)	; generate this type of random data
   title 			; title string for plot
   old-window			; a plot, if adding to an existing plot
   )

Generates part of the phase plane for a differential equation using Euler's
method (by default). Points are generated randomly, or can be added to an
existing plot using the :old-window option.

This includes an option of using your own distribution for randomly selecting
points for the phase portrait. It will also generate a single orbit, if that's
what you've got in mind (by default).

;; Returns: the plot (also defines the plot my-plot, if the :old-window option
;; is not set).

;; Notes: this version of phase-plane replaces the old, in which two values
;; (xin and yin) were given rather than a list (initial). Look for backward
;; compatibility problems! This version is essentially a copy of orbitals for
;; two dimensions.

;; See also: euler-nd, rk-4, orbitals.

;; Examples:
;; Van der Pol Equations:
 (defun vpol(vals &optional (epsilon .2))
  (let ((x (first vals))(y (second vals)))
    (list (/ (- y (/ (^ x 3) 3) (- x)) epsilon) (- (* epsilon x)))
    )
  )

 ;; a collection of orbits:
 (phase-plane #'vpol (list .1 .1) 1 100 10
	     :randomize 1
	     :randomizer #'normal-rand
	     :title \"Van der Pol\")

 ;; a single orbit:
 (phase-plane #'vpol (list .1 .1) 10 1000 1 :title \"Van der Pol\")

 ;; Adding specific orbits is easy with phase-plane, using the :old-window
 ;; key. Note that my-plot is defined automagically:
 (defun page111(vals)
   (let (
	(x (elt vals 0))
	(y (elt vals 1))
	)
    (list (- (+ x (square y)))
	  (+ y (square x)))
    ))
 (defun initcond(a) (list (/ (square a) -3) a))
 (phase-plane #'page111 (initcond .001) 5 1000 1 :title \"Page 111\")
 (phase-plane #'page111 (* '(1 -1) (initcond .001)) 10 2000 1
	       :old-window my-plot)
 (defun initcond(a) (list a (/ (square a) -3)))
 (defun neg-page111 (vals) (- (page111 vals)))
 (phase-plane #'neg-page111 (initcond .001) 5 1000 1
	       :old-window my-plot)
 (phase-plane #'neg-page111 (* '(-1 1) (initcond .001)) 10 2000 1
	       :old-window my-plot)

;; Will also generate a 'phase-plane' (projection of trajectories) if the
;; equations are higher dimensional! e.g., the Rossler equations:
 (defun rossler (vals)
   (let ((x (first vals))(y (second vals))(z (third vals)))
     (list
	(- (+ y z))
	(+ x (* .2 y))
	(+ .2 (* z (- x 5.7))))))
 (def initial (* 10 (uniform-rand 3)))
 (phase-plane #'rossler initial 10 500 1 :integrator #'rk-4)
"
  (let (
	(len (length initial))
	(randomize (if randomize (if (numberp randomize) randomize 1)))
	(randomize (if (neq randomizer #'uniform-rand)
		       (if randomize randomize 1)
		     randomize))
	(nextpoint initial)
	points xvals yvals
	)

    ;; for each orbit, we call our integrator:
    (dotimes
     (i ntimes)
     (setq 
      points (funcall integrator f nextpoint 0 end nsteps)
      xvals (mapcar #'first points)
      yvals (mapcar #'second points)
      nextpoint (if randomize
		    (* randomize (funcall randomizer len))
		  (final points))
      nextpoint (if randomize-about-initial (+ initial nextpoint) nextpoint)
      )

     ;; if we haven't a window to plot in, then create my-plot:
     (if (not old-window)
	 (let()
	   (def my-plot (plotter xvals yvals))
	   (setq old-window my-plot)
	   (if title (send old-window :title title))
	   )
       )

     ;; add in the current lines, and the last point:
     (send my-plot :add (list 'lines xvals yvals))
     (send my-plot :add-points (last xvals) (last yvals))

     ;; adjust to data, so that everything is in view:
     (if show (send old-window :adjust-to-data))
     )

    (if (not show) (send old-window :adjust-to-data))
    (send old-window :plot-item)
    (send old-window :adjust-to-data)

    ;; return the plot:
    old-window
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
