(defun logistic
  (x yvals &rest args &key guess weights (prob .05) latex verbose (print t))
  "
;; Description: function
 (logistic x yvals
    &rest args &key guess weights (prob .05) latex (verbose 2) (print t))

       The world's greatest multivariate logistic regression code (I had to say
that: McGee made me).

;; Arguments: (yvals x &rest args
	&key guess weights (prob .05) latex (verbose 2) (print t))

       We regress y on x (a list, or a matrix), with initial guess guess. You
may also add any arguments relevant to newtonmax, as they will be passed to it.
Weights on points are also allowed.

If you have a guess, use it (it should be a list the size of the columns of x,
plus one). Setting the latex option will return the results as a latex table.
The verbose command will produce some diagnostics.

;; See also: newtonmax

;; Examples:
 ;;
 (def chd '(0 0 0 0 1 0 0 0 0 0 0 0 0 0 0 1 0 0 0 0 0 0 1 0 0 0 0 0 1 0 0 1 0 0
0 0 1 0 1 0 0 0 0 0 1 0 0 1 0 0 1 1 0 1 0 1 0 0 1 0 1 1 0 0 1 0 1 0 0 1 1 1 1 0
1 1 1 1 1 0 0 1 1 1 1 0 1 1 1 1 0 1 1 1 1 1 0 1 1 1))
 (def age '(20 23 24 25 25 26 26 28 28 29 30 30 30 30 30 30 32 32 33 33 34 34
34 34 34 35 35 36 36 36 37 37 37 38 38 39 39 40 40 41 41 42 42 42 42 43 43 43
44 44 44 44 45 45 46 46 47 47 47 48 48 48 49 49 49 50 50 51 52 52 53 53 54 55
55 55 56 56 56 57 57 57 57 57 57 58 58 58 59 59 60 60 61 62 62 63 64 64 65 69))
 (logistic age chd :guess '(-7 .2) :scale '(40))
 ;;
 (def faux-age (sample (iseq 20 69) (length chd) t))
 (logistic faux-age chd :scale '(40))
 (logistic faux-age chd :scale '(40) :latex t)
 (logistic faux-age chd :verbose t :latex t)
 ;;
 ;;
 ;; p. 319, Lawless's Statistical Models and Methods for Lifetime Data,
 ;; and exercise 6.11 on p. 341:
 (def yvals '(1.94 2.23 1.94 1.98 2.23 1.59 2.13 1.8 2.32 1.92 2.15 2.05 2.48
		  2.42 2.56 2.56))
 (def ages '(42 67 62 52 57 58 55 63 44 62 51 64 54 64 54 57))
 (def performance '(4 6 4 6 5 6 6 7 5 7 7 10 8 3 9 9))
 (def xvals (bind-columns ages performance))
 (logistic xvals (exp (- yvals)) :guess '(.01 .01 .01 ) :scale '(50 6))
 (logistic xvals (exp (- yvals)) :guess '(0 0 0) :scale '(50 6))
 ;;
 ;;
 ;; From page 322, Afifi and Clark, Computer-Aided Multivariate Analysis:
 ;; 183 women, 111 men; 1, depressed; 0, not depressed:
 (def xvals (combine (repeat 1 183) (repeat 0 111)))
 (def yvals (combine (repeat 1 40) (repeat 0 143) (repeat 1 10) (repeat 0 101)))
 (logistic xvals yvals :guess (list -2 1)) ;; (-2.3125445 1.0385787)
 ;;
 ;; Note that 1.039 = ln((40/143)/(10/101)):
 (calc ln((40/143)/(10/101))) ;; ~= 1.0385702
 ;; and that -2.313 = ln(10/101):
 (calc ln(10/101)) ;; ~= -2.3125354
 ;;
 ;; Note for Jay: p. 323 Afifi and Clark, Computer-Aided Multivariate Analysis:
 ;; \"When there is only a single dichotomous variable, it is not worthwhile to
 ;; perform a logistic regression analysis.\" 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
"
  (let* (
	 (ones (repeat 1 (length yvals)))
	 (xvals
	  (if (listp x)
	      (if (listp (first x))
		  (apply #'bind-columns x)
		(matrix (list (length yvals) 1) x)
		)
	    x)) ;; it must be a matrix already
	 (guess (if guess guess (repeat 0 (1+ (second (size xvals))))))
	 result
	 L0 L1 Gradient Hessian params std-errors confidence correlations
	 (probval (normal-quant (- 1 (/ prob 2))))
	 )

    (if verbose (format t "Using Newtonmax with initial guess ~g ~%" guess))
    (setq
     result
     (newtonmax
      #'(lambda (x)
	  (let* (
		 (a (first x))
		 (b (rest x))
		 (stuff (+ 1 (exp (- (+ (* a ones) (matmult xvals b))))))
		 (otherstuff
		  (+
		   (* yvals (- (ln stuff)))
		   (* (- ones yvals) (ln (- 1 (/ stuff))))
		   )
		  )
		 )
	    (sum (if weights (* weights otherstuff) otherstuff))
	    )
	  )
      guess
      :return-derivs t
      :verbose verbose
      )
     )
    (if verbose (printall "Done with Newtonmax with result:" result ""))
    (setq params (first result))
    (setq L1 (second result))
    (setq Gradient (third result))
    (setq Hessian (fourth result))
    (setq std-errors (sqrt (abs (diagonal (inverse Hessian)))))
    (setq correlations (- (inverse Hessian)))
    (setq confidence
	  (weave (exp (- params (* probval std-errors)))
		 (exp (+ params (* probval std-errors)))
		 )
	  )

    ;; log-likelihood of null hypothesis (the variables are useless)
    (let* (
           (a (sum (if weights (* weights yvals) yvals)))
           (b (sum (if weights (* weights (- 1 yvals)) (- 1 yvals))))
           (n (+ a b))
           )
      (setq L0 (+ (* a (ln (/ a n))) (* b (ln (/ b n)))))
      )
    
    (if latex
	(let ()
	  (latex-table
	   (matrix
	    '(10 2)
	    (list
	     "Parameter Values:" params
	     "Standard Errors:" std-errors
	     "Parameters/(Standard Errors):" (/ params std-errors)
             "Confidence Intervals for odds ratios:" confidence
             "Parameter Correlations:" correlations
	     "Gradient Values (expect/hope for zeros):" gradient
	     "Best Log Likelihood:" L1
	     "Null Log Likelihood:" L0
	     "Chi-square of 2*(best-null):" (* 2 (- L1 L0))
	     (strcat "Chi-square Probability ("
		     (num-to-string (second (size xvals))) " df):")
	     (- 1 (chisq-cdf (* 2 (- L1 L0)) (second (size xvals))))
	     )
	    )
	   :caption "Logistic Regression Results"
	   )
	  )
      (if print
	  (let ()
	    (format t "~%~%                 Results:~%~%Parameters : ~%")
	    (mapcar #'space-print-num params)
	    (format t "~%Standard Errors (sqrt (abs (diagonal (inv H)))) : ~%")
	    (mapcar #'space-print-num std-errors)
	    (format t "~%Parameters/(Standard Errors) (t statistic) : ~%")
	    (mapcar #'space-print-num (/ params std-errors))
	    (format t "~%Confidence Intervals for odds ratios: ~%")
	    (mapcar #'space-print-num confidence)
	    (format t "~%Parameter Correlations: ~%")
	    (mapcar #'space-print-num
		    (print-lower-diagonal correlations :print nil))
	    
	    (format t "~%Gradient Values (expect/hope for zeros)     :~%")
	    (mapcar #'space-print-num Gradient)
	    
	    (format t "~%Best Log Likelihood                         : ~g~%" L1)
	    (format t "Null Log Likelihood                         : ~g~%~%" L0)
		    
	    (format t "Chi-square of 2*(best-null)                 : ~g~%"
		    (* 2 (- L1 L0)))
	    (format t "Corresponding Chi-square Probability:       : ~,8f~%~%"
		    (- 1 (chisq-cdf (* 2 (- L1 L0)) (second (size xvals)))))
	    )
	)
      )
    (list params std-errors confidence L0 L1 Gradient Hessian)
    )
  )  

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                     geometric-regression
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun geometric-regression (x)
  "
;; Description: geometric-regression (matrix)
      Performs the regression XB=1, rather than focusing on one or the other
of the variables as dependent.

;; Examples:
 (def test-reg (rand.u 100))
 (def error (* .5 (rand.u 100)))
 (def regmod (regression-model test-reg (+ (* test-reg 3) error)))
 (def regmod (send regmod :coef-estimates))
 (setq georegmod
     (geometric-regression (bind-columns (+ (* test-reg 3) error) test-reg)))
 (setq georegmod
      (/ (combine 1 (- (rest georegmod))) (first georegmod)))

 (lines on)
 (title \"geometric regression (blue) vs linear regression (black)\")
 (plot #'(lambda(x) (+ (first regmod) (* x (second regmod)))))
 (send my-plot :range 0 0 1)
 (send my-plot :range 1 0 4)
 (send my-plot :add-plot-item (list 'points test-reg (+ (* test-reg 3) error)))
 (send my-plot :add-plot-item
   (list 'function
	 #'(lambda(x) (+ (first georegmod) (* x (second georegmod))))
	 'blue))
 (send my-plot :update-plot)
 ;; (undef '(test-reg error regmod georegmod holder))
"
  (let* (
	 (nmeans (mapcar #'sum (column-list x))) ;; n times the means
	 (inv (inverse (matmult (transpose x) x))) ;; inverse of Sigma
	 (vals (matmult inv nmeans)) ;; coefficient estimates
	 )
    (print (- (/ (combine 1 (rest vals) (first vals)))))
    (values (matmult inv nmeans) (diagonal inv))
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                            logistic-maximum-likelihood
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun logistic-maximum-likelihood(x y parameters &key (verbose t))
  "
;; Documentation:
  (logistic-maximum-likelihood x y parameters &key (verbose t))

;; Examples:
 (def chd '(0 0 0 0 1 0 0 0 0 0 0 0 0 0 0 1 0 0 0 0 0 0 1 0 0 0 0 0 1 0 0 1 0 0
0 0 1 0 1 0 0 0 0 0 1 0 0 1 0 0 1 1 0 1 0 1 0 0 1 0 1 1 0 0 1 0 1 0 0 1 1 1 1 0
1 1 1 1 1 0 0 1 1 1 1 0 1 1 1 1 0 1 1 1 1 1 0 1 1 1))
 (def age '(20 23 24 25 25 26 26 28 28 29 30 30 30 30 30 30 32 32 33 33 34 34
34 34 34 35 35 36 36 36 37 37 37 38 38 39 39 40 40 41 41 42 42 42 42 43 43 43
44 44 44 44 45 45 46 46 47 47 47 48 48 48 49 49 49 50 50 51 52 52 53 53 54 55
55 55 56 56 56 57 57 57 57 57 57 58 58 58 59 59 60 60 61 62 62 63 64 64 65 69))
 (logistic-maximum-likelihood age chd '(-5.3097 0.1109))
 (logistic-maximum-likelihood age chd '(-5.49028 0.114559))
 
 (spin-domain -6 -5 .1 .12)
 (spin-points 20)
 (plot
   #'(lambda(x y)
    (logistic-maximum-likelihood age chd (list x y) :verbose nil)))
"
  (let* (
	 (xvals
	  (if (listp x)
	      (if (listp (first x))
		  (apply #'bind-columns x)
		(matrix (list (length y) 1) x)
		)
	    x)) ;; it must be a matrix already
	 (a (first parameters))
	 (b (rest parameters))
	 (stuff (/ (+ 1 (exp (- (+ a (matmult xvals b)))))))
	 (newstuff
	  (mapcar
	   #'(lambda(z) (if (= (first z) 1) (second z) (- 1 (second z))))
	   (weave y stuff)
	   )
	  )   
	 (lnlikelihood (sum (ln newstuff)))
	 (hltest (simul-sort stuff y))
	 (expected (first hltest))
	 (observed (second hltest))
	 (tenth (floor (/ (length y) 10)))
	 (hos-lem 0)
	 )
    (if
	verbose
	(let()
	  (format t "~%~25t~a~%" "Hosmer-Lemeshow Test:")
	  (format t "~17t~a~31t~a~47t~a~%" "Total" "observed" "expected")
	  (dotimes
	   (i 9)
	   (let* (
		  (batch (iseq (* tenth i) (+ tenth (* tenth i))))
		  (e (sum (select expected batch)))
		  (o (sum (select observed batch)))
		  )
	     (format t
		     " ~dth decile: ~20t~d ~35t~d ~50t~,2f~%"
		     (1+ i) (length batch) o e)
	     (setq hos-lem (+ hos-lem (/ (square (- e o)) e)))
	     )
	   )
	  (let* (
		 (batch (iseq (* tenth 9) (1- (length y))))
		 (e (sum (select expected batch)))
		 (o (sum (select observed batch)))
		 )
	    (format t "~dth decile: ~20t~d ~35t~d ~50t~,2f~%" 10 (length batch)
		    o e)
	    (setq hos-lem (+ hos-lem (/ (square (- e o)) e)))
	    )
	  (format t "Chi-Square of ~,4g with 8 df: ~,6f~%~%"
		  hos-lem (CHISQ-cdf hos-lem 8))
	  
	  (format t "Log-likelihood: ~30t~,4g~%" lnlikelihood)
	  
	  (format t "Maximum-likelihood: ~30t~,4g~%" (exp lnlikelihood))
	  
	  (format t "Sum of Squared Errors: ~30t~,4g~%" (sum (square (- y newstuff))))
	  )
      )
    (values 
     lnlikelihood
;;     (exp lnlikelihood)
;;     (sum (square (- y newstuff)))
     )
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 minimize-sum-of-squares
;;
(defun minimize-sum-of-squares
  (y function
     &key guess (verbose 2) (count-limit 10))
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; minimize-sum-of-squares Doc:  Creation Date:2/21/97    Author: Andy Long
;;
;; Description/Arguments:

       (minimize-sum-of-squares y function
		&key guess (verbose 2) (count-limit 10)
  
;; Notes: This function minimizes the sum of squared residual errors by using
;; the Newton-Raphson procedure.

;; See Also: newtonmax

;; Examples:
 (def x (rand.u 100))
 (def y (+ (* 2 x) (* .1 (rand.u 100))))
 (minimize-sum-of-squares y #'(lambda(z) (* z x)))
 ;; should be around 2...

 (def y (+ 3 (* 2 x) (* .1 (rand.u 100))))
 (minimize-sum-of-squares y #'(lambda(w z) (+ w (* z x))))
 ;; should be around 3 and 2...

 (def y (+ 3 (* 2 x) (* .1 (rand.u 100))))
 (minimize-sum-of-squares y (func f(w z) (w+z*x)))
 ;; should be around 3 and 2...
"
  (let* (
	 ;; Dimension of the function that you use:
	 (n (length (second (eval `(get-lambda-expression ,function)))))
	 (guess (if guess guess (repeat 0 n)))
	 )
    (printall "Initial guess:" guess)
    (terpri)
    (newtonmax
     #'(lambda(x)
	 (-
	  (sum
	   (square
	    (-
	     y
	     (apply function x)
	     )
	    )
	   )
	  )
	 )
     guess
     :verbose verbose
     :count-limit count-limit
     )
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
#|
	Some regression-model-proto methods...
|#
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :missing
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(send regression-model-proto :add-slot 'missing nil)
(defmeth regression-model-proto :missing (&optional new)
  "
Sets the missing value indicator for a particular regression.
"
  (when new (setf (slot-value 'missing) new))
  (slot-value 'missing)
  )

#|
I'm trying to figure out how to get the right r-squared to use for a regression
without an intercept:

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :r-squared
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth regression-model-proto :r-squared ()
  "Message args: ()
Returns the sample squared multiple correlation coefficient, R squared, for
the regression.

	I changed this to give the proper r-squared in the event that there is
no intercept. I did this by replacing :total-sum-of-squares with :tss.
"
  (max (- 1 (/ (send self :sum-of-squares) (send self :total-sum-of-squares))) 0)
  ;;  (max (- 1 (/ (send self :sum-of-squares) (send self :tss))) 0)
  )
|#


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :plot
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth regression-model-proto :plot (&optional xmin xmax)
  "
;; Description: :plot ()

	This method allows one to plot the regression model in a 1x1
regression. 

;; Examples:
 (def xreg (rand.u 100))
 (def areg (regression-model xreg (+ (* 3 xreg) (rand.u 100))))
 (send areg :plot)

;; 	This example is from the book \"Applied Regression Analysis\", by
;; N. R. Draper and H. Smith., second edition.
 (def obs (iseq 1 25))
 (data y 10.98 11.13 12.51 8.4 9.27 8.73 6.36 8.5 7.82 9.14 8.24 12.19 11.88
      9.57 10.94 9.58 10.09 8.11 6.83 8.88 7.68 8.47 8.86 10.36 11.08)
 (data x8 35.3 29.7 30.8 58.8 61.4 71.3 74.4 76.7 70.7 57.5 46.4 28.9 28.1 39.1
     46.8 48.5 59.3 70 70 74.5 72.1 58.1 44.6 33.4 28.6)
 (def page9 (regression-model x8 y))
 (title \"Page 9, Draper and Smith\")
 (send page9 :plot 25 85)
"
  (let* (
	 (x (column 1 (send self :x)))
	 (pars (send self :coef-estimates))
	 (len (length pars))
	 (y (if (> len 2)
		(send self :adjust (iseq 1 (- len 2)))
	      (send self :y)
	      ))
	 (intercept (send self :intercept))
	 (slope (if intercept (second pars) (first pars)))
	 (intercept (if intercept (first pars) 0))
	 (myplot (send scatterplot-proto :new 2 :draw nil))
	 (n (length x))
	 (xmean (mean x))
	 (ssx (sum (square (- x xmean))))
	 (s (sqrt (send self :s-squared)))
	 )
    (send myplot :title *title*)
    (send myplot :add-slot 'andy-options '(:axes t))
    (send myplot :add-slot 'plot-item)
    (send myplot :add-plot-item
	  (list 'points (list (mean x)) (list (mean y)) 'blue 'disk)
	  (list 'points x y 'black 'dot4)
	  (list 'function #'(lambda(x) (+ intercept (* slope x))) 'black t)
	  ;; The next two functions give confidence intervals (95%) for the
	  ;; true mean value of Y:
	  (list 'function (send self :top-ci-function) 'blue t)
	  (list 'function (send self :bottom-ci-function) 'blue t)
	  :draw nil
	  )
    (if (and xmin xmax)
	(let ((range (get-nice-range (min y) (max y) 5)))
	  (send myplot :range 0 xmin xmax)
	  (send myplot :range 1 (first range) (second range))
	  (send myplot :plot-item)
	  (make-title (format nil "y=~,4g x+ ~,4g" (second pars) (first pars)))
	  )
      (let (
	    (domain (get-nice-range (min x) (max x) 5))
	    (range (get-nice-range (min y) (max y) 5))
	    )
	(send myplot :range 0 (first domain) (second domain))
	(send myplot :range 1 (first range) (second range))
	(make-title (format nil "y=~,4g x+ ~,4g" slope intercept))
	)
      )
    (send myplot :show-window)
    (send myplot :plot-item)
    myplot
    )
  )
#|
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :plot-2d
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth regression-model-proto :plot-2d (&optional xmin xmax ymin ymax)
  "
;; Description: :plot-2d ()

	This method allows one to plot the regression model in a 1x1
regression. 

;; Examples:
;; 	This example is from the book \"Applied Regression Analysis\", by
;; N. R. Draper and H. Smith., second edition.
 (def y  (list 10.98 11.13 12.51 8.4 9.27 8.73 6.36 8.5 7.82 9.14 8.24
		12.19 11.88 9.57 10.94 9.58 10.09 8.11 6.83 8.88 7.68
		8.47 8.86 10.36 11.08))
 (def x6 (list 20 20 23 20 21 22 11 23 21 20 20 21 21 19 23 20 22 22 11 23 20
		21 20 20 22))
 (def x8 (list 35.3 29.7 30.8 58.8 61.4 71.3 74.4 76.7 70.7 57.5 46.4
		28.9 28.1 39.1 46.8 48.5 59.3 70 70 74.5 72.1 58.1 44.6
		33.4 28.6))
 (title \"Page 194\")
 (def page194 (regression-model (list x6 x8) y))
 (send page194 :plot-2d)
"
  (let* (
	 (x (send self :x))
	 (y (column 2 x))
	 (x (column 1 x))
	 (z (send self :y))
	 (intercept (send self :intercept))
	 (intercept (if intercept (first (send self :coef-estimates)) 0))
	 (pars (if intercept (rest (send self :coef-estimates))
		 (send self :coef-estimates)))
	 (slope1 (first pars))
	 (slope2 (second pars))
	 (xmin (if xmin xmin (min x)))
	 (xmax (if xmax xmax (max x)))
	 (ymin (if ymin ymin (min y)))
	 (ymax (if ymax ymax (max y)))
	 (spin-domain *spin-domain*)
	 (tmp (spin-domain xmin xmax ymin ymax))
	 (my-plot (plotter (lambda (x y) (+ intercept (* slope1 x) (* slope2 y)))))
	 (tmp (funcall #'spin-domain spin-domain))
	 )
    (send my-plot :add-plot-item (list 'points x y z :color 'blue))
    (send my-plot :plot-item)
    my-plot
    )
  )
|#
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :plot-2d
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth regression-model-proto :plot-2d (&optional xmin xmax ymin ymax)
  "
;; Description: :plot-2d ()

	This method allows one to plot the regression model in a 1x1
regression. 

;; Examples:
;; 	This example is from the book \"Applied Regression Analysis\", by
;; N. R. Draper and H. Smith., second edition.
 (def y  (list 10.98 11.13 12.51 8.4 9.27 8.73 6.36 8.5 7.82 9.14 8.24
		12.19 11.88 9.57 10.94 9.58 10.09 8.11 6.83 8.88 7.68
		8.47 8.86 10.36 11.08))
 (def x6 (list 20 20 23 20 21 22 11 23 21 20 20 21 21 19 23 20 22 22 11 23 20
		21 20 20 22))
 (def x8 (list 35.3 29.7 30.8 58.8 61.4 71.3 74.4 76.7 70.7 57.5 46.4
		28.9 28.1 39.1 46.8 48.5 59.3 70 70 74.5 72.1 58.1 44.6
		33.4 28.6))
 (title \"Page 194\")
 (def page194 (regression-model (list x6 x8) y))
 (send page194 :plot-2d)
"
  (let* (
	 (x (send self :x))
	 (y (column 2 x))
	 (x (column 1 x))
	 (z (send self :y))
	 (intercept (send self :intercept))
	 (intercept (if intercept (first (send self :coef-estimates)) 0))
	 (pars (if intercept (rest (send self :coef-estimates))
		 (send self :coef-estimates)))
	 (slope1 (first pars))
	 (slope2 (second pars))
	 (myplot (spin-function
		  (lambda (x y) (+ intercept (* slope1 x) (* slope2 y)))
		  (if xmin xmin (min x))
		  (if xmax xmax (max x))
		  (if ymin ymin (min y))
		  (if ymax ymax (max y))
		  :num-points *spin-points*
		  :spline *spin-spline*
		  :color 'blue
		  )
		 )
	 )
    (send myplot :add-slot 'plot-item nil)
    (send myplot :mouse-mode 'hand-rotate)
    (if *spin-axes* (send myplot :showing-axes t)
      (send myplot :showing-axes nil))
    (if *spin-depth* (send myplot :depth-cuing t)
      (send myplot :depth-cuing nil))
    (if *real-axes* (send myplot :real-axes))
    (if *title* (send myplot :title *title*))
    (send myplot :add-slot 'andy-options (list ':axes t))
    (send myplot :add-plot-item (list 'points x y z))
    (send myplot :plot-item)
    myplot
    )
  )
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                     :sse
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth regression-model-proto :sse ()
  "
;; Description: :sse ()

	residual-sum-of-squares: sum((yhat-y)^2)

;; Examples:
 (def xreg (rand.u 100))
 (def areg (regression-model xreg (+ (* 3 xreg) (rand.u 100))))
 (send areg :sse)
"
  (send self :residual-sum-of-squares)
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :rss
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth regression-model-proto :rss ()
  "
;; Description: :rss ()

	residual-sum-of-squares: sum((yhat-y)^2)

;; Examples:
 (def xreg (rand.u 100))
 (def areg (regression-model xreg (+ (* 3 xreg) (rand.u 100))))
 (send areg :rss)
"
  (send self :residual-sum-of-squares)
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :resss
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth regression-model-proto :resss ()
  "
;; Description: :resss ()

	residual-sum-of-squares = sum((yhat-y_i)^2)

	The residual sum of square is the norm-squared of the residuals.

;; Examples:
 (def xreg (rand.u 100))
 (def areg (regression-model xreg (+ (* 3 xreg) (rand.u 100))))
 (send areg :resss)
"
  (send self :residual-sum-of-squares)
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :ss
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth regression-model-proto :ss ()
  "
;; Description: :ss ()

	s-squared: sum((yhat-y)^2)/(n-2)

;; Examples:
 (def xreg (rand.u 100))
 (def areg (regression-model xreg (+ (* 3 xreg) (rand.u 100))))
 (send areg :ss)
"
  (/ (send self :sse) (send self :n-p))
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :s-squared
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth regression-model-proto :s-squared ()
  "
;; Description: :s-squared ()

	s-squared: sum((yhat-y)^2)/(n-2)

;; Examples:
 (def xreg (rand.u 100))
 (def areg (regression-model xreg (+ (* 3 xreg) (rand.u 100))))
 (send areg :s-squared)
"
  (/ (send self :sse) (send self :n-p))
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :tss
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth regression-model-proto :tss ()
  "
;; Description: :tss ()

			total-sum-of-squares:

	sum((y_i-ymean)^2) = sum((yhat-ymean)^2) + sum((yhat-y_i)^2)

;; Notes: aka sum of squares about the mean, total corrected SS (see Draper and
;; Smith, Applied Regression Analysis, 2nd edition, p. 17 ff.

	This has been altered to give the non-mean-centered total sum of
squares when the intercept term is absent.

;; Examples:
 (def xreg (rand.u 100))
 (def areg (regression-model xreg (+ (* 3 xreg) (rand.u 100))))
 (send areg :tss)
"
  (if (send self :intercept)
      (send self :total-sum-of-squares)
    (sum (square (send self :y))))
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :regss
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth regression-model-proto :regss ()
  "
;; Description: :regss ()

	regression sum-of-squares = sum((yhat-ymean)^2)

;; Notes:
	This has been altered to use the non-mean-centered total sum of
squares when the intercept term is absent, as :tss does.

;; Examples:
 (def xreg (rand.u 100))
 (def areg (regression-model xreg (+ (* 3 xreg) (rand.u 100))))
 (send areg :regss)
"
  (- (send self :tss) (send self :sse))
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :f
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth regression-model-proto :f ()
  "
;; Description: : ()

	f statistic for regression

;; Examples:
 (def xreg (rand.u 100))
 (def areg (regression-model xreg (+ (* 3 xreg) (rand.u 100))))
 (send areg :f)
"
  (let (
	(cdf (if (send self :intercept)
		 (1- (send self :num-coefs))
	       (send self :num-coefs)))
	)
    (/ (/ (send self :regss) cdf) (send self :s-squared))
    )
  )  

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :prob-f
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth regression-model-proto :prob-f ()
  "
;; Description: : ()

	probability of f statistic for regression

;; Examples:
 (def xreg (rand.u 100))
 (def areg (regression-model xreg (+ (* 3 xreg) (rand.u 100))))
 (send areg :prob-f)
"
  (let (
	(cdf (if (send self :intercept)
		 (1- (send self :num-coefs))
	       (send self :num-coefs)))
	)
    (- 1 (f-cdf (send self :f) cdf (send self :n-p)))
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :estimates
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth regression-model-proto :estimates ()
  "
;; Description: : ()

	estimates: yhat

;; Examples:
 (def xreg (rand.u 100))
 (def areg (regression-model xreg (+ (* 3 xreg) (rand.u 100))))
 (new-plot)
 (lines off)
 (plot (column 1 (send areg :x)) (send areg :estimates))
"
  (send self :fit-values)
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :estimation-variance
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth regression-model-proto :estimation-variance(x)
  (let (
	(x0 (if (send self :intercept) (combine 1 x) x))
	)
    (* (send self :s-squared) (matmult x0 (send self :xtxinv) x0))
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :n-p
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth regression-model-proto :n-p()
  (if (send self :intercept)
      (1- (apply #'- (size (send self :x))))
    (apply #'- (size (send self :x)))
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :coef-probs
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth regression-model-proto :coef-probs (&key (prob .05) verbose)
  "
;; Description: : regression-model-proto :coef-probs (&key (prob .05) verbose)

	coef-probs: estimate/se

and the probability of this estimated t-statistic.

;; Examples:
 (def xreg (rand.u 100))
 (def areg (regression-model xreg (+ (* 3 xreg) (rand.u 100))))
 (send areg :coef-probs)
 (send areg :coef-probs)
"
  (let* (
	 (df (send self :n-p))
	 (ts
	  (/ (send self :coef-estimates) (send self :coef-standard-errors)))
	 (probs (* 2 (- 1 (t-cdf (abs ts) df))))
	 )
    (if verbose
	(let ()
	  (format
	   t "t-statistic for two-sided t-test with ~d df and probability ~,3f:~% ~g~%"
	   df prob (t-quant (- 1 (half prob)) df ))
	  (format t "t-scores for coefficients: ~g:~%" ts)
	  (format t "probabilities for coefficients: ~g.~%" probs)
	  )
      )
    probs
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :coef-confidence-intervals
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth regression-model-proto :coef-confidence-intervals
  (&key (prob .05) verbose)
  "
;; Description: : regression-model-proto :coef-confidence-intervals
			(&key (prob .05) verbose)

	coef-confidence-intervals: estimate/se

and the probability of this estimated t-statistic.

;; Examples:
 (def xreg (rand.u 100))
 (def areg (regression-model xreg (+ (* 3 xreg) (rand.u 100))))
 (send areg :coef-confidence-intervals)
"
  (let* (
	 (df (send self :n-p))
	 (coefs (send self :coef-estimates))
	 (tval (t-quant (- 1 (half prob)) df))
	 (se (send self :coef-standard-errors))
	 (ci (* tval se))
	 )
    (if verbose
	(dolist
	 (i (weave coefs ci))
	 (format t "(~13,8g,~13,8g)"
		 (- (first i) (second i))
		 (+ (first i) (second i))
		 )
	 )
      )
    (weave (- coefs ci) (+ coefs ci))
    )	  
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :xtx 
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth regression-model-proto :xtx ()
  "
;; Description: xtx returns x'x for the matrix used in the regression. If
intercept, then a column of ones is adjoined to the matrix first.
"
  (let* (
	 (x (send self :x))
	 (x (if (send self :intercept)
		(bind-columns (repeat 1 (first (size x))) x)
	      x)
	    )
	 )
    (matmult (transpose x) x)
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :joint-coef-probs
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth regression-model-proto :joint-coef-probs
  (&key (b 0) (prob .95) (print t))
  "
;; Description: : (&key (b 0) (prob .95) (print t))

	joint-coef-probs: (beta-b)'X'X(beta-b) <= ps^2F(p,v,1-alpha)

;; Examples:
 (def xreg (matrix '(7 3) (list
   -1 -1  1
    1 -1  1
   -1  1  1
    1  1  1
    0  0  0
    0  1  0
    0  2  0
  )))
 (def yreg '(1 4 8 9 3 8 9))
 (def areg (regression-model xreg yreg))
 (send areg :joint-coef-probs)
"
  (let* (
	 (beta (send self :coef-estimates))
	 (xtx (send self :xtx))
	 (diff (- beta b))
	 (val (matmult diff xtx diff))
	 (p (length beta))
	 (n (length (send self :y)))
	 (nu (- n p))
	 (statval (* p (send self :ss) (f-quant prob p nu)))
	 )
    (if print
	(let ()
	  (format t "~%f-test of all coefficients with probability ~g:~%~%"
		  prob)
	  (if (<= val statval)
	      (format t "Within confidence region: ~g <= ~g" val statval)
	    (format t "Outside confidence region: ~g > ~g" val statval)
	    )
	  ))
    val
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :joint-coef-probs-function
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth regression-model-proto :joint-coef-probs-function (&key (b 0))
  "
;; Description: regression-model-proto :joint-coef-probs-function (&key (b 0))
		  
	joint-coef-probs-function: see Draper and Smith, Applied Regression
Analysis, 2nd Edition, p. 94:

             f(b) = f-cdf[(beta-b)'X'X(beta-b) / (p*s^2), n, p]

	The joint-coef-probs-function is between 0 and 1: if you plot a contour
line at 1-alpha, then it gives the region in which the parameters will fall
1-alpha of the time. Obviously, as alpha gets smaller, the region will get
bigger.

				f(b) = 1-alpha

;; Examples:
 (data table1.1.x 35.3 29.7 30.8 58.8 61.4 71.3 74.4 76.7 70.7 57.5 46.4 28.9
      28.1 39.1 46.8 48.5 59.3 70 70 74.5 72.1 58.1 44.6 33.4 28.6)
 (data table1.1.y 10.98 11.13 12.51 8.4 9.27 8.73 6.36 8.5 7.82 9.14 8.24 12.19
      11.88 9.57 10.94 9.58 10.09 8.11 6.83 8.88 7.68 8.47 8.86 10.36 11.08)
 (def x table1.1.x)
 (def y table1.1.y)
 (def areg (regression-model x y))
 (def pf (send areg :joint-coef-probs-function))

 (spin-domain 12 15 -.11 -.04)
 ;; (The spin domain is taken about the coefficient estimates.)
 (spin-points 20)
 (contour (lambda(x y) (funcall pf (list x y)))
   :xmin 12 :xmax 15 :ymin -.11 :ymax -.04
   :levels '(.9 1)
   :labels t)
"
  (let* (
	 (beta (send self :beta))
	 (p (length beta))
	 (nu (send self :n-p))
	 (weight (/ (send self :xtx) (* p (send self :ss))))
	 )
    (lambda(x)
      (let ((val (- beta x)))
	(f-cdf (matmult val weight val) p nu)
	)
      )
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;          Make it easier to ask for coef-estimates and SE:
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :beta
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth regression-model-proto :beta () 
  "Shorthand for the :coef-estimates method."
  (send self :coef-estimates))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :se
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth regression-model-proto :se () 
  "Shorthand for the :coef-standard-errors method."
  (send self :coef-standard-errors))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;          I modified the standard display to include probabilities:
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :display
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth regression-model-proto :display (&key (fmt "~,6g"))
  "
Message args: &key (fmt \"~,6g~\")

	Prints the least squares regression summary. Variables not used in the
fit are marked as aliased.

;; Notes: ael modified the standard display to include probabilities, and to
;; consolidate things a little. Added the formating option, so that I can get
;; more presentable.

"
  (let* (
	 (coefs (coerce (send self :coef-estimates) 'list))
	 (se-s (send self :coef-standard-errors))
	 ;; se-s can be zero for a saturated model: if so, the standard errors
	 ;; will print out as nil!
	 (probs (if se-s (send self :coef-probs)))
	 (x (send self :x))
	 (p-names (send self :predictor-names))
	 (head-fmt "~%~a~26t~a ~46t~a ~62t~a~2%")
	 (real-fmt (strcat "~a~25t" fmt "~40t(" fmt ")~60t~,5f~%"))
	 (real2-fmt (strcat "~a~25t" fmt "~%"))
	 (int-fmt "~a~25t~9d~%")
	 )
    (if (send self :weights) 
        (format t head-fmt "Weighted Regression:" "Estimate" "SE" "Prob")
      (format t head-fmt "Linear Regression:" "Estimate" "SE" "Prob")
      )
    (when (send self :intercept)
          (format t real-fmt "Constant" (car coefs) (car se-s) (car probs))
          (setf coefs (cdr coefs))
          (setf probs (cdr probs))
          (setf se-s (cdr se-s)))
    (dotimes (i (array-dimension x 1)) 
             (cond 
	      ((member i (send self :basis))
	       (format t real-fmt
		       (select p-names i) (car coefs) (car se-s) (car probs))
	       (setf coefs (cdr coefs) se-s (cdr se-s) probs (cdr probs)))
	      (t (format t "~a~25taliased~%" (select p-names i)))))
    (format t "~%")
    (format t real2-fmt "R Squared:" (send self :r-squared))
    (format t real2-fmt "Sigma hat:" (send self :sigma-hat))
    (format t int-fmt "Number of cases:" (send self :num-cases))
    (if (/= (send self :num-cases) (send self :num-included))
        (format t int-fmt "Number of cases used:" (send self :num-included)))
    (format t int-fmt "Degrees of freedom:" (send self :df))
    (format t "~%")))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :anova
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth regression-model-proto :anova (&key dups)
  "
;; Description: regression-model-proto method :anova (&key dups)

	Designed to create standard anova models, including lack-of-fit and
pure-error statistics.

;; Examples:
 (title \"Exercise N, page 63\")
 (def x (* 10 (list 1 1 2 2 3 3 4 4 4 5 5 5)))
 (def y (list -2 -4 1 3 2 5 0 1 2 -2 -3 -4))

 (def exn (regression-model x y))
 (def my-plot (send exn :plot))
 (send my-plot :add-plot-item (list 'function {1.692 - x * .0546} 'red t))
 (send exn :anova)
"
  (let* (
	 (x (row-list (send self :x)))
	 (y (send self :y))
	 (n (send self :num-cases))
	 (dups (if dups dups (find-duplicates x)))
	 (non-dups (set-difference (iseq n) (combine dups)))
	 (ys (mapcar #'list (select y non-dups)))
	 (tss (send self :tss))
	 (resss (send self :resss))
	 (regss (send self :regss))
	 (r-squared (send self :r-squared))
	 (s-squared (send self :s-squared))
	 (df (send self :n-p))
	 (intercept (send self :intercept))
	 (cdf (if intercept
		  (1- (send self :num-coefs))
		(send self :num-coefs)))
	 (f (send self :f))
	 (prob-f (send self :prob-f))
	 )
    (dolist
     (i dups)
     (setq ys (append ys (list (select y i))))
     )
    (let* (
	   (group-means (mapcar #'mean ys))
	   (pure
	    (sum
	     (mapcar #'(lambda(x y) (sum (square (- x y)))) group-means ys)
	     )
	    )
	   (pure-df (- (sum (mapcar #'length ys)) (length ys)))
	   (lack-of-fit (- resss pure))
	   (ddf (- df pure-df))
	   (msl (if (= ddf 0) nan (/ lack-of-fit ddf)))
	   (se2 (if (= pure-df 0) 0 (/ pure pure-df)))
	   (f-ratio (if (= se2 0) 0 (/ msl se2)))
	   (max-r2 (/ (- tss pure) tss))
	   (fmt1 "~%  ~a ~14t~a ~29t~a ~49t~a ~64t~a~2%")
	   (fmt2 "~1t~a ~14t~d ~25t~8,8g ~45t~8,8g ~60t~8,8g~%")
	   (fmt3 "~1t~a ~14t~d ~25t~8,8g ~45t~8,8g ~60tProb(f)=~,4f~2%")
	   )    
      (format t fmt1 "Source" "df" "SS" "MS" "F ratio")
      (format t fmt2 "Regression" cdf regss (/ regss cdf) f)
      (format t fmt3 "Residual" df resss s-squared prob-f)

      (if (/= pure-df 0)
	  (let ()
	    (format t fmt2 "Lack of Fit" (- df pure-df) lack-of-fit msl f-ratio)
	    (format t fmt3 "Pure Error" pure-df pure se2
		    (- 1 (f-cdf f-ratio cdf (send self :n-p))))
	    )
	)
      
      (format t " Total ~14t~d ~25t~8,8g~2%" (- n cdf) tss)

      (format t "~%Percent of possible variance explained: ~8,8g~%"
	      (* 100 (/ r-squared max-r2)))
      )
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :pure-error
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth regression-model-proto :pure-error (&key dups)
  "
;; Description: regression-model-proto method :pure-error (&key dups)

	Returns the pure-error mean-square error.

;; Examples:
 (title \"Exercise N, page 63\")
 (def x (* 10 (list 1 1 2 2 3 3 4 4 4 5 5 5)))
 (def y (list -2 -4 1 3 2 5 0 1 2 -2 -3 -4))

 (def exn (regression-model x y))
 (def my-plot (send exn :plot))
 (send exn :pure-error)
"
  (let* (
	 (x (row-list (send self :x)))
	 (y (send self :y))
	 (n (send self :num-cases))
	 (dups (if dups dups (find-duplicates x)))
	 (non-dups (set-difference (iseq n) (combine dups)))
	 (ys (mapcar #'list (select y non-dups)))
	 (tss (send self :tss))
	 (resss (send self :resss))
	 (regss (send self :regss))
	 (r-squared (send self :r-squared))
	 (s-squared (send self :s-squared))
	 (df (send self :n-p))
	 (intercept (send self :intercept))
	 (cdf (if intercept
		  (1- (send self :num-coefs))
		(send self :num-coefs)))
	 (f (send self :f))
	 (prob-f (send self :prob-f))
	 )
    (dolist
     (i dups)
     (setq ys (append ys (list (select y i))))
     )
    (let* (
	   (group-means (mapcar #'mean ys))
	   (pure
	    (sum
	     (mapcar #'(lambda(x y) (sum (square (- x y)))) group-means ys)
	     )
	    )
	   (pure-df (- (sum (mapcar #'length ys)) (length ys)))
	   (lack-of-fit (- resss pure))
	   (ddf (- df pure-df))
	   (msl (if (= ddf 0) nan (/ lack-of-fit ddf)))
	   (se2 (if (= pure-df 0) 0 (/ pure pure-df)))
	   (f-ratio (if (= se2 0) 0 (/ msl se2)))
	   (max-r2 (/ (- tss pure) tss))
	   )    
      (if (/= pure-df 0)
	  (list
	   (list (- df pure-df) lack-of-fit msl f-ratio)
	   (list pure-df pure se2 (- 1 (f-cdf f-ratio cdf (send self :n-p))))))
      )
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :anova-sequential
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth regression-model-proto :anova-sequential
  (&key sets partial (verbose t) pure-error set-names)
  "
;; Description: regression-model-proto method :anova-sequential
  (&key sets partial (verbose t) set-names)

	Designed to create standard anova models, sequentially.

	If you use the sets key, you will give a list of lists, each containing
a set of variables to add sequentially.

	set-names should be a list of strings of the same size as the list of
sets.

;; Examples:
 (def y  (list 10.98 11.13 12.51 8.4 9.27 8.73 6.36 8.5 7.82 9.14 8.24
		12.19 11.88 9.57 10.94 9.58 10.09 8.11 6.83 8.88 7.68
		8.47 8.86 10.36 11.08))
 (def x6 (list 20 20 23 20 21 22 11 23 21 20 20 21 21 19 23 20 22 22 11 23 20
		21 20 20 22))
 (def x8 (list 35.3 29.7 30.8 58.8 61.4 71.3 74.4 76.7 70.7 57.5 46.4
		28.9 28.1 39.1 46.8 48.5 59.3 70 70 74.5 72.1 58.1 44.6
		33.4 28.6))
 (title \"Page 194\")
 (def page194 (regression-model (list x6 x8) y))
 (send page194 :anova-sequential)

 (def x1 (list 7 1 11 11 7 11 3 1 2 21 1 11 10))
 (def x2 (list 26 29 56 31 52 55 71 31 54 47 40 66 68))
 (def x3 (list 6 15 8 8 6 9 17 22 18 4 23 9 8))
 (def x4 (list 60 52 20 47 33 22 6 44 22 26 34 12 12))
 (def x5
      (/ (list 785 743 1043 876 959 1092 1027 725 931 1159 838 1133 1094) 10))
 (send (regress (list x1 x2 x4) x5 :print nil) :anova-sequential :partial t)
 (send (regress (list x1 x2 x4) x5 :print nil) :anova-sequential
     :sets (list (list x1 x2 x3) (list x4)))
"
  (let* (
	 (x (send self :x))
	 (intercept (send self :intercept))
	 (p (second (size x)))
	 (x (column-list x))
	 (n (send self :num-cases))
	 (tss (send self :tss))
	 (resss (send self :resss))
	 (regss (send self :regss))
	 (r-squared (send self :r-squared))
	 (df (send self :n-p))
	 (cdf (if (send self :intercept)
		  (1- (send self :num-coefs))
		(send self :num-coefs)))
	 (f (send self :f))
	 (prob-f (send self :prob-f))
	 (y (send self :y))
	 (name (if set-names set-names
		 (mapcar
		  (lambda(x) (strcat (if sets "Set " "Variable ")
				     (num-to-string x)))
		  (iseq (length x)))
		 ))
	 (fmt1 "~%  ~a ~14t~a ~29t~a ~44t~a ~57t~a~2%")
	 (fmt2a "~1t~a ~14t~d ~25t~8,8g ~40t~8,8g ~53t~8,8g ~64t~,4f~%")
	 (fmt2b "~1t~a ~14t~d ~25t~8,8g ~40t~8,8g~%")
	 (fmt3 "~1t~a ~14t~d ~25t~8,8g ~40t~8,8g ~53tProb(f)=~,4f~2%")
	 (k 0)
	 xx result
	 (pure-error (if pure-error (send self :pure-error)))
	 (s-squared (if pure-error (third (second pure-error))
		      (send self :s-squared)))
	 (ddf (if pure-error (first (second pure-error)) (send self :n-p)))
	 )
    (if verbose 
	(let ()
	  (format t fmt1 "Source" "df" "SS" "MS" "F ratio and prob.")
	  (format t fmt2a "Regression" cdf regss (/ regss cdf) f
		  (- 1 (f-cdf (/ (/ regss cdf) s-squared) cdf ddf))
		  )
	  ))
    (if sets
	(dolist
	 (i sets)
	 (setq xx (smoosh (list xx i)))
	 (let* (
		(reg (regression-model xx y :print nil :intercept intercept))
		(cdf (length i))
		(regss (send reg :regss))
		(f (/ (/ regss cdf) s-squared))
		(prob-f (- 1 (f-cdf f cdf ddf)))
		)
	   (if verbose
	       (format t fmt2a (elt name k) cdf regss (/ regss cdf) f prob-f))
	   (setq y (send reg :residuals))
	   (setq result (if result (combine result prob-f) (list prob-f)))
	   )
	 (incf k)
	 )
      (if partial
	  (do
	   ((i 0 (1+ i)))
	   ((= i (length x)))
	   (setq xx (select x (remove i (iseq p))))
	   (let* (
		  (reg (regression-model xx y :print nil :intercept intercept))
		  (regss2 (send reg :regss))
		  (diff (- regss regss2))
		  (f (/ diff s-squared))
		  (prob-f (- 1 (f-cdf f 1 ddf)))
		  )
	     (if verbose (format t fmt2a (elt name i) 1 diff diff f prob-f))
	     (setq result (if result (combine result prob-f) (list prob-f)))
	     )
	   )
	(do
	 ((i 0 (1+ i)))
	 ((= i (length x)))
	 (setq xx (select x (iseq (1+ i))))
	 (let* (
		(reg (regression-model xx y :print nil :intercept intercept))
		(regss (send reg :regss))
		(f (/ regss s-squared))
		(prob-f (- 1 (f-cdf f 1 ddf)))
		)
	   (if verbose (format t fmt2a (elt name i) 1 regss regss f prob-f))
	   (setq y (send reg :residuals))
	   (setq result (if result (combine result prob-f) (list prob-f)))
	   )
	 )
	)
      )
    (if verbose
	(let ()
	  (if pure-error
	      (let ((l (first pure-error))(p (second pure-error)))
		(format t fmt2a "Lack-of-fit"
			(first l) (second l) (third l) (fourth l) (fourth p))
			
		(format t fmt2b "Pure Error" (first p) (second p) (third p))
		)
	    (format t fmt3 "Residual" df resss s-squared prob-f)
	    )
	  (format t " Total ~14t~d ~25t~8,8g~2%" (- n cdf) tss)
	  (format t "R-squared: ~8,8g~%" r-squared)
	  ))
    result
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :anova-no-dups
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth regression-model-proto :anova-no-dups ()
  "
;; Description: regression-model-proto method :anova-no-dups ()

	Designed to create standard anova models, without lack-of-fit and
pure-error statistics.

;; Examples:
 (title \"Exercise N, page 63\")
 (def x (* 10 (list 1 1 2 2 3 3 4 4 4 5 5 5)))
 (def y (list -2 -4 1 3 2 5 0 1 2 -2 -3 -4))

 (def exn (regression-model x y))
 (def my-plot (send exn :plot))
 (send my-plot :add-plot-item (list 'function {1.692 - x * .0546} 'red t))
 (send exn :anova-no-dups)
"
  (let* (
	 (x (row-list (send self :x)))
	 (n (send self :num-cases))
	 (tss (send self :tss))
	 (resss (send self :resss))
	 (regss (send self :regss))
	 (r-squared (send self :r-squared))
	 (s-squared (send self :s-squared))
	 (df (send self :n-p))
	 (cdf (if (send self :intercept)
		  (1- (send self :num-coefs))
		(send self :num-coefs)))
	 (f (send self :f))
	 (prob-f (send self :prob-f))
	 )
    (let* (
	   (fmt1 "~%  ~a ~14t~a ~29t~a ~49t~a ~64t~a~2%")
	   (fmt2 "~1t~a ~14t~d ~25t~8,8g ~45t~8,8g ~60t~8,8g~%")
	   (fmt3 "~1t~a ~14t~d ~25t~8,8g ~45t~8,8g ~60tProb(f)=~,4f~2%")
	   )    
      (format t fmt1 "Source" "df" "SS" "MS" "F ratio")
      (format t fmt2 "Regression" cdf regss (/ regss cdf) f)
      (format t fmt3 "Residual" df resss s-squared prob-f)
      (format t " Total ~14t~d ~25t~8,8g~2%" (- n cdf) tss)
      (format t "R-squared: ~8,8g~%" r-squared)
      )
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                Anova
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun anova (x y f)
  "anova(x y f) is designed to allow me to test other hypothetical models than
the least squares regression model.

;; Examples:
 (title \"Exercise N, page 63\")
 (def x (* 10 (list 1 1 2 2 3 3 4 4 4 5 5 5)))
 (def y (list -2 -4 1 3 2 5 0 1 2 -2 -3 -4))

 (def exn (regression-model x y))
 (def my-plot (send exn :plot))
 (send my-plot :add-plot-item (list 'function {1.692 - x * .0546} 'red t))

 (send exn :anova)
 ;; This is the true model, according to the regression above:
 (anova x y (send exn :regression-function))

 ;; compare it to this model, which inspired this function (one other than the
 ;; actual regression model):
 (anova x y {1.692 - x * .0546})
"
  (let* (
	 (n (length y))
	 (x (if (matrixp x) (row-list x) x))
	 (dups (find-duplicates x))
	 (non-dups (set-difference (iseq n) (combine dups)))
	 (ys (mapcar #'list (select y non-dups)))
	 (mean-y (mean y))
	 (estimates (mapcar f x))
	 (tss (sum (square (- y mean-y))))
	 (regss (sum (square (- estimates mean-y))))
	 (resss (- tss regss))
	 (r-squared (/ regss tss))
	 (cdf (if (listp (first x)) (length (first x)) 1))
	 (df (- n (1+ cdf)))
	 (mse (/ regss cdf))
	 (s-squared (/ resss df))
	 (f (/ mse s-squared))
	 )
    (dolist
     (i dups)
     (setq ys (append ys (list (select y i))))
     )
    (let* (
	   (group-means (mapcar #'mean ys))
	   (pure
	    (sum
	     (mapcar #'(lambda(x y) (sum (square (- x y)))) group-means ys)
	     )
	    )
	   (pure-df (sum (mapcar (lambda(x) (1- (length x))) ys)))
	   (lack-of-fit (- resss pure))
	   (msl (/ lack-of-fit (- df pure-df)))
	   (se2 (if (= pure-df 0) 0 (/ pure pure-df)))
	   (f-ratio (if (= se2 0) 0 (/ msl se2)))
	   (max-r2 (/ (- tss pure) tss))
	   (fmt1 "~%  ~a ~14t~a ~29t~a ~49t~a ~64t~a~2%")
	   (fmt2 "~1t~a ~14t~d ~25t~8,8g ~45t~8,8g ~60t~8,8g~%")
	   (fmt3 "~1t~a ~14t~d ~25t~8,8g ~45t~8,8g ~60tProb(f)=~,4f~2%")
	   )    
      (format t fmt1 "Source" "df" "SS" "MS" "F ratio")
      (format t fmt2 "Regression" cdf regss (/ regss cdf) f)
      (format t fmt3 "Residual" df resss s-squared (- 1 (f-cdf f cdf df)))
      
      (if (/= pure-df 0)
	  (let ()
	    (format t fmt2 "Lack of Fit" (- df pure-df) lack-of-fit msl f-ratio)
	    (format t fmt3 "Pure Error" pure-df pure se2
		    (- 1 (f-cdf f-ratio cdf df)))
	    )
	)
      
      (format t " Total ~14t~d ~25t~8,8g~2%" (- n cdf) tss)
      
      (format t "~%Percent of possible variance explained: ~8,8g~%"
	      (* 100 (/ r-squared max-r2)))
      )
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 :hypothesis
;;
(defmeth regression-model-proto  :hypothesis
  (matrix
   &key verbose print (epsilon machine-epsilon))
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; :hypothesis Doc:    Creation Date:2/28/97        Author: Andy Long
;;
;; Description/Arguments:

       :hypothesis matrix &key verbose print (epsilon machine-epsilon)
  
;; Notes:

;; See Also:

;; Examples:
 (def y (list 1 4 8 9 3 8 9))
 (def x (matrix '(7 4)
   (list 1 -1 -1 1 1 1 -1 1 1 -1 1 1 1 1 1 1 1 0 0 0 1 0 1 0 1 0 2 0)))
 (def x (select-columns x '(1 2 3)))
 (def page106 (regression-model x y))

 ;; This one (and the next) assert that the third parameter is zero, and the
 ;; sum of the first two is 0:
 (send page106 :hypothesis #2a((0 0 1)(1 -1 0)))
 ;; padding with zeros is okay, but why bother?
 (send page106 :hypothesis #2a((0 0 1)(1 -1 0)(0 0 0)))
 ;; even adding in degenerate (linearly dependent) rows is okay (but why?):
 (send page106 :hypothesis #2a((0 0 1)(1 -1 0)(1 -1 1)))
 (send page106 :hypothesis #2a((0 0 1)) :verbose t)
"
  (let* (
	 (x (send self :x))
	 (y (send self :y))
	 (size (size matrix))
	 (tmp matrix)
	 (matrix
	  (if (= (second (size x)) (second size))
	      (transpose matrix)
	    (error (strcat "Dimensions are not correct.
Each row should have the same dimension as the
number of parameters (other than the intercept). In your case, this row
dimension should be " (num-to-string (second (size x))) "."))
	    )
	  )
	 (matrix (pad-to-square-with-zeros matrix))
	 (svd (svd matrix))
	 (matrix
	  (select-columns
	   (first svd)
	   (where (second svd) epsilon #'<)
	   )	   
	  )
	 (new-x (matprod x matrix))
	 (intercept (send self :intercept))
	 (new-self (regression-model new-x y :print print :intercept intercept))
	 (p (send self :num-coefs))
	 (q (send new-self :num-coefs))
	 (df (send self :n-p))
	 (ndf (- p q))
	 (self-resss (send self :resss))
	 (new-self-resss (send new-self :resss))
	 (diff (- new-self-resss self-resss))
	 (stat (/ (/ diff ndf) (/ self-resss df)))
	 (prob (- 1 (f-cdf stat ndf df)))
	 )
    (if verbose
	(let ()
	  (format t
		  "~%The hypothesis is represented by the matrix~%")
	  (mprint tmp)
	  (format t "~%~40tResidual SS
    Original model: ~40t~13,8g ~60tDenDF: ~70t~d
Hypothetical model: ~40t~13,8g
        Difference: ~40t~13,8g ~60tNumDF: ~70t~d

Test f-statistic for hypothesis: ~40t~13,8g

Probability of a more extreme value: ~40t~10,7f~2%
" self-resss df new-self-resss diff ndf stat prob)
	  )
      )
    prob
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; :regression-function Doc:    Creation Date:3/1/97        Author: Andy Long
;;
;;
(defmeth regression-model-proto  :regression-function()
  "
;; Description/Arguments:

       (defmeth regression-model-proto  :regression-function()
  
;; Notes:

;; See Also:

;; Examples:

"
  (let* (
	 (all-coefs (send self :coef-estimates))
	 (intercept (send self :intercept))
	 (coefs (if intercept (rest all-coefs) all-coefs))
	 (intercept (if intercept (first all-coefs) 0))
	 )
    (lambda(x) (+ intercept (sum (* x coefs))))
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; :ci-function Doc:    Creation Date:3/1/97        Author: Andy Long
;;
;;
(defmeth regression-model-proto  :ci-function(&key obs (p .05))
  "
;; Description/Arguments:

       (defmeth regression-model-proto  :ci-function(&key obs (p .05))
  
;; Notes: Only works for one-predictor regressions.

;; See Also: method :plot

;; Examples:
 (title \"Exercise O, page 63\")
 (def x (/ (list 133 169 199 232 263 301 426) 10))
 (def y (/ (list 35 51 48 67 60 95 81) 10))

 (def exo (regression-model x y))
 (send (send exo :plot 0 100) :range -2 30)
 (mapcar (send exo :ci-function) '(0 100))
"
  (let* (
	 (xtxinv (send self :xtxinv))
	 (n-p (send self :n-p))
	 (s (sqrt (send self :s-squared)))
	 (n (send self :num-cases))
	 (tmp1
	  (*
	   s
	   (t-quant (- 1 (/ p 2)) n-p)
	   )
	  )
	 )
    (lambda(x)
      (let* (
	     (y (funcall (send self :regression-function) x))
	     (x (if (send self :intercept) (combine 1 x) x))
	     (tmp2 (* tmp1 (sqrt (+ (if obs (/ obs) 0) (matprod x xtxinv x)))))
	     )
	(list
	 (- y tmp2)
	 (+ y tmp2)
	 )	
	)
      )
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; :top-ci-function Doc:    Creation Date:3/1/97        Author: Andy Long
;;
;;
(defmeth regression-model-proto  :top-ci-function()
  "
;; Description/Arguments:

       (defmeth regression-model-proto  :top-ci-function()
  
;; Notes: Only works for one-predictor regressions.

;; See Also: method :plot

;; Examples:
 (title \"Exercise O, page 63\")
 (def x (/ (list 133 169 199 232 263 301 426) 10))
 (def y (/ (list 35 51 48 67 60 95 81) 10))

 (def exo (regression-model x y))
 (send (send exo :plot 0 100) :range -2 30)
 (mapcar (send exo :top-ci-function) '(0 100))
"
  (let* (
	 (x (column 1 (send self :x)))
	 (all-coefs (send self :coef-estimates))
	 (intercept (send self :intercept))
	 (coefs (if intercept (rest all-coefs) all-coefs))
	 (intercept (if intercept (first all-coefs) 0))
	 (slope (first coefs))
	 (s (sqrt (send self :s-squared)))
	 (xmean (mean x))
	 (ssx (sum (square (- x xmean))))
	 (n (send self :num-cases))
	 )
    (lambda(x)
      (+
       (+ intercept (* slope x))
       (*
	s
	(t-quant .975 (- n 2))
	(sqrt
	 (+ (/ n) (/ (square (- x xmean)) ssx))
	 )
	)
       )
      )
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; :bottom-ci-function Doc:    Creation Date:3/1/97        Author: Andy Long
;;
;;
(defmeth regression-model-proto  :bottom-ci-function()
  "
;; Description/Arguments:

       (defmeth regression-model-proto  :bottom-ci-function()
  
;; Notes: Only works for one-predictor regressions.

;; See Also: method :plot

;; Examples:
 (title \"Exercise O, page 63\")
 (def x (/ (list 133 169 199 232 263 301 426) 10))
 (def y (/ (list 35 51 48 67 60 95 81) 10))

 (def exo (regression-model x y))
 (send (send exo :plot 0 100) :range -2 30)
 (mapcar (send exo :bottom-ci-function) '(0 100))
"
  (let* (
	 (x (column 1 (send self :x)))
	 (all-coefs (send self :coef-estimates))
	 (intercept (send self :intercept))
	 (coefs (if intercept (rest all-coefs) all-coefs))
	 (intercept (if intercept (first all-coefs) 0))
	 (slope (first coefs))
	 (s (sqrt (send self :s-squared)))
	 (xmean (mean x))
	 (ssx (sum (square (- x xmean))))
	 (n (send self :num-cases))
	 )
    (lambda(x)
      (-
       (+ intercept (* slope x))
       (*
	s
	(t-quant .975 (- n 2))
	(sqrt
	 (+ (/ n) (/ (square (- x xmean)) ssx))
	 )
	)
       )
      )
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                adjust
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth regression-model-proto :adjust (vars &key vals)
  "
;; Description: :adjust (vars &key vals)

	This method returns the adjusted y values given a set of variables for
which on wished to adjust. A list of vars should be passed (unless only a
single variable is to be adjusted for, in which case you may pass a single
value).

	The variables should be numbered from 0, and do not include the
 intercept term.

	If you wish to adjust your list of variables (vars) to something other
 than their means, then pass a list of the values (must be the same length as
 the vars list).

;; Notes: there is a discussion of \"adjustment\" on page 58 of Hosmer and
Lemeshow: \"The analogous situation in linear regression is called \"analysis
of covariance.\"
	Confounding and effect modification is described on pages 63ff:
Confounder: a covariate that is associated with both the outcome variable of
	interest and a primary independent variable or risk factor.
Effect Modifier: a variable that interacts with a risk factor.
Some examples are given.

;; Examples:

 ;; (data are in file absorbtion.lsp in the sample data directory/folder):
 (load-data \"absorbtion\")
 (def m (regress (list iron aluminum) absorbtion))
 (send m :coef-estimates)
 (send m :adjust '(0)) ;; adjust for iron
 (send m :adjust 0) ;; easier, since we're only doing a single variable.

 (def m (regress (list iron aluminum) absorbtion :intercept nil))
 (send m :coef-estimates)
 (send m :adjust '(0)) ;; adjust for iron
 (send m :adjust 0) ;; easier, since we're only doing a single variable.

 (def m (regress (list iron aluminum absorbtion) (iseq (length iron))))
 (send m :coef-estimates)
 (send m :adjust '(0 1)) ;; adjust for iron and aluminum
"
  (let* (
	 (y (send self :y))
	 (x (send self :x))
	 (size (size x))

	 (all-vars (iseq (second size)))
	 (vars (if (listp vars) vars (list vars)))
	 (keep-vars (sort-data (set-difference all-vars vars)))
	 
	 (vals (if vals (if (listp vals) vals (list vals))
		 (mapcar #'mean (column-list (select-columns x vars)))))
	 
	 (params (send self :coef-estimates))
	 (intercept (if (send self :intercept) (first params)))
	 (params (if intercept (rest params) params))
	 (intercept (if intercept intercept 0))

	 (adjust-params (select params vars))
	 (keep-params (select params keep-vars))

	 (s0 (sum (* vals adjust-params)))
	 (s2 (matprod (select-columns x vars) adjust-params))
	 (adjust-y (- s0 s2))
	 )

    (values (+ y adjust-y)
	    (combine
	     (+ intercept (sum (* adjust-params vals)))
	     keep-params
	     )
	    )
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :latex-table
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth regression-model-proto :latex-table ()
  "
Args: ()
Returns the estimates, standard errors, and r-squared in a latex-table.
"
  (latex-table
   (bind-columns (send self :beta) (send self :se) (send self :coef-probs))
   :caption
   (format
    nil
    "Estimates, standard errors and probabilities. N of ~d; R-squared of ~5,4f.
The probabilities given are that the estimate is significantly different from
zero at $\\\alpha=.05$, assuming that everything is nice."
    (send self :num-included) (send self :r-squared)
    )
   :row-names (combine "Intercept" (send self :predictor-names))
   :column-names (list "Estimate" "Standard Error" "Probability")
   )
  )


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 :summary-data
;;
(defmeth regression-model-proto :summary-data (numbers sd)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          :summary-data Doc:
;;
;; Author: Andy Long
;; Created: Wed Apr 23 11:48:26 CDT 1997
;;
;; Description/Arguments:

       :summary-data (numbers sd)
  
;; Notes: I'm not sure this works right: it just gets me some of the answers to
;; problem N, in Draper and Smith's Applied Regression Analysis.

;; See Also: :summary-data

;; Examples:
 (title \"Example N, p. 284\")
 (def x (iseq 16 23))
 (def x (combine x x x))
 (def wt (/ (list 10382 10539 10750 10312 10583 10750 10250 11750 9917 10767
 		10357 11219 11000 11313 11200 11321 10083 9932 10083 9639 10167
 		10000 nan 11250) 100))
 (def groups (combine (repeat 'a 8) (repeat 'b 8) (repeat 'c 8)))
 (def i1 (indicator groups (where groups 'a)))
 (def i2 (indicator groups (where groups 'b)))
 
 (def exn (regress (list x) wt :missing nan))
 (send exn :plot)
 (def exn (regress (list i1 i2 x) wt :missing nan))
 (send exn :summary-data students sd)
 
 (def students
      (list 19 19 16 8 6 6 1 1 12 29 28 32 22 8 4 7 18 33 24 18 6 4 nan 1))
 ;; In truth the following 0 sd's are missing:
 (def sd (/ (list 1270 1250 1530 768 1675 577 0 0 920 1955 1305 1760 1204
 		 1487 1118 728 1675 1450 1967 1329 1812 1521 nan 0) 100))
 (def exn2 (send exn :summary-data students sd))
 (send exn2 :anova)
 (def xmat (send exn2 :x))
 (send exn2 :anova-sequential
      :sets
      (list
       (list (column 1 xmat) (column 2 xmat))
       (list (column 3 xmat))
       ))
"
  (let* (
	 (missing (send self :missing))
	 (x (send self :x))
	 (y (send self :y))
	 (numbers (remove-missing numbers missing))
	 (sd (remove-missing sd missing))
	 (new-y (combine (mapcar (lambda(x y) (repeat x y)) y numbers)))
	 (new-x 
	  (matrix
	   (list (length new-y) (second (size x)))
	   (combine (mapcar (lambda(x y) (repeat x y)) (row-list x) numbers))
	   ))
	 (result (regress (column-list new-x) new-y :missing missing))
	 )
    (send result :add-slot 'missing nil)
    (defmeth result :missing (&optional new)
      (when new (setf (slot-value 'missing) new))
      (slot-value 'missing)
      )
    (send result :missing missing)
    result
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 regress
;;
(defun regress(&rest args &key missing)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          regress Doc:
;;
;; Author: Andy Long
;; Created: Wed Apr 23 12:32:09 CDT 1997
;;
;; Description/Arguments:

       (regress &rest x)
  
	This is just a simple interface to regression-model, which also
includes the option of prescribing a missing value so that these can be removed
on the fly. They will be removed from both the x and y values. You can also
pass an X matrix to regress.

	The usual arguments of regression-model may be passed to regress.

;; Notes:

;; See Also: regression-model

;; Examples:
 (title \"Example N, p. 284\")
 (def x (iseq 16 23))
 (def x (combine x x x))
 (def students
      (list 19 19 16 8 6 6 1 1 12 29 28 32 22 8 4 7 18 33 24 18 6 4 0 1))
 (def groups (combine (repeat 'a 8) (repeat 'b 8) (repeat 'c 8)))
 (def wt (/ (list 10382 10539 10750 10312 10583 10750 10250 11750 9917 10767
 		10357 11219 11000 11313 11200 11321 10083 9932 10083 9639 10167
 		10000 nan 11250) 100))
 (def exn (regress x wt :missing nan :predictor-names (list \"x\")))

 ;; An examples with two predictors:
 (def x1 (rand.u 20))
 (def x2 (rand.u 20))
 (def y (mapcar (lambda(x y z) (+ x (* 2 y) z)) x1 x2 (rand.u 20)))
 (def exn (regress (list x1 x2) y))

 (def x1 (list 7 1 11 11 7 11 3 1 2 21 1 11 10))
 (def x2 (list 26 29 56 31 52 55 71 31 54 47 40 66 68))
 (def x3 (list 6 15 8 8 6 9 17 22 18 4 23 9 8))
 (def x4 (list 60 52 20 47 33 22 6 44 22 26 34 12 12))
 (def x5
      (/ (list 785 743 1043 876 959 1092 1027 725 931 1159 838 1133 1094) 10))
 (def reg (regress (list x1 x2 x3 x4) x5))
 (def reg (regress (list (repeat 1 (length x1)) x1 x2 x3 x4) x5 :intercept nil))
"
  (let* (
	 (x (first args))
	 (y (second args))
	 (x (turn-it-into-a-matrix x :from 'regress))
	 (x (bind-columns x y))
	 (len (second (size x)))
	 (args (rest (rest args)))
	 result
	 )
    (setq 
     result
     (if missing (remove-missing x missing) (remove-missing x))
     )
    (setq result
	  (apply #'regression-model
		 (smoosh 
		  (list
		   (list
		    (column-list (select-columns result (iseq (1- len))))
		    )
		   (list (column len result))
		   args
		   )
		  )
		 )
	  )
    (send result :add-slot 'missing nil)
    (defmeth result :missing (&optional new)
      (when new (setf (slot-value 'missing) new))
      (slot-value 'missing)
      )
    (send result :missing missing)
    result
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                :anova-summary
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth regression-model-proto :anova-summary
  (numbers sd
	   &key 
	   (missing (send self :missing))
	   )
  "
;; Description: regression-model-proto method :anova-summary
  (numbers sd
	   &key 
	   (missing (send self :missing))
	   )

	Designed to create standard anova-summary models, including lack-of-fit
and pure-error statistics for summary data. This should be run only after a
model is created using :summary-data.

;; Examples:
 (title \"Example N, p. 284\")
 (def x (iseq 16 23))
 (def x (combine x x x))
 (def wt (/ (list 10382 10539 10750 10312 10583 10750 10250 11750 9917 10767
 		10357 11219 11000 11313 11200 11321 10083 9932 10083 9639 10167
 		10000 nan 11250) 100))
 (def groups (combine (repeat 'a 8) (repeat 'b 8) (repeat 'c 8)))
 (def i1 (indicator groups (where groups 'a)))
 (def i2 (indicator groups (where groups 'b)))
 
 (def exn (regress (list x) wt :missing nan))
 (send exn :plot)
 (def exn (regress (list i1 i2 x) wt :missing nan))
 
 (def students
      (list 19 19 16 8 6 6 1 1 12 29 28 32 22 8 4 7 18 33 24 18 6 4 nan 1))
 ;; In truth the following 0 sd's are missing:
 (def sd (/ (list 1270 1250 1530 768 1675 577 0 0 920 1955 1305 1760 1204
 		 1487 1118 728 1675 1450 1967 1329 1812 1521 nan 0) 100))
 (def exn2 (send exn :summary-data students sd))
 (send exn2 :anova)
 (def xmat (send exn2 :x))
 (send exn2 :anova-sequential
       :sets
       (list
        (list (column 1 xmat) (column 2 xmat))
        (list (column 3 xmat))
        ))
 (send exn2 :anova-summary students sd)
"
  (let* (
	 (x (row-list (send self :x)))
	 (y (send self :y))
	 (n (send self :num-cases))
	 (dups (find-duplicates x))
	 (non-dups (set-difference (iseq n) (combine dups)))
	 (ys (mapcar #'list (select y non-dups)))

	 (numbers (if missing (remove-missing numbers missing)) numbers)
	 (sd (if missing (remove-missing sd missing)) sd)
	 (sum (sum (* (1- numbers) (square sd))))
	 (tss (+ sum (send self :tss)))
	 (resss (+ sum (send self :resss)))

	 (regss (send self :regss))
	 (r-squared (send self :r-squared))
	 (s-squared (send self :s-squared))
	 (df (send self :n-p))
	 (intercept (send self :intercept))
	 (cdf (if intercept
		  (1- (send self :num-coefs))
		(send self :num-coefs)))
	 (f (send self :f))
	 (prob-f (send self :prob-f))
	 )
    (dolist
     (i dups)
     (setq ys (append ys (list (select y i))))
     )
    (let* (
	   (group-means (mapcar #'mean ys))
	   (pure
	    (sum
	     (mapcar #'(lambda(x y) (sum (square (- x y)))) group-means ys)
	     )
	    )
	   (pure (+ sum pure))
	   (pure-df (sum (mapcar (lambda(x) (1- (length x))) ys)))
	   (lack-of-fit (- resss pure))
	   (ddf (- df pure-df))
	   (msl (if (= ddf 0) nan (/ lack-of-fit ddf)))
	   (se2 (if (= pure-df 0) 0 (/ pure pure-df)))
	   (f-ratio (if (= se2 0) 0 (/ msl se2)))
	   (max-r2 (/ (- tss pure) tss))
	   (fmt1 "~%  ~a ~14t~a ~29t~a ~49t~a ~64t~a~2%")
	   (fmt2 "~1t~a ~14t~d ~25t~8,8g ~45t~8,8g ~60t~8,8g~%")
	   (fmt3 "~1t~a ~14t~d ~25t~8,8g ~45t~8,8g ~60tProb(f)=~,4f~2%")
	   )    
      (format t fmt1 "Source" "df" "SS" "MS" "F ratio")
      (format t fmt2 "Regression" cdf regss (/ regss cdf) f)
      (format t fmt3 "Residual" df resss s-squared prob-f)

      (if (/= pure-df 0)
	  (let ()
	    (format t fmt2 "Lack of Fit" (- df pure-df) lack-of-fit msl f-ratio)
	    (format t fmt3 "Pure Error" pure-df pure se2
		    (- 1 (f-cdf f-ratio cdf (send self :n-p))))
	    )
	)
      
      (format t " Total ~14t~d ~25t~8,8g~2%" (- n cdf) tss)

      (format t "~%Percent of possible variance explained: ~8,8g~%"
	      (* 100 (/ r-squared max-r2)))
      )
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 mallows
;;
(defmeth regression-model-proto :mallows
  (&key verbose r-squared)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          mallows Doc:
;;
;; Author: Andy Long
;; Created: Thu May 8 09:02:53 CDT 1997
;;
;; Description/Arguments:

       method regression-model-proto :mallows

		  (&key verbose r-squared)
  
;; Notes:

	This is described in Draper and Smith, Applied Regression Analysis,
2nd Edition, p. 299. The expected value is the number of variables in the
regression (including the intercept). It is defined as 

		C_p = ResSS_p/s^2 - (n - 2p)

where s^2 is the estimate of s-squared from the original regression (assumed to
include all variables of interest). Mallows is computed for the regressions of
all (proper) subsets. If you want to know which subset corresponds to the best
value of Mallows, then you need to see how the subsets are allocated by
all-subsets for your set of variables. To do this, I usually try it on a set of
symbols (see the examples).

	Set the r-squared key to t if you would rather see the r-squared values
for all regressions.

;; See Also: all-subsets

;; Examples:
 (def x (rand.u 10))
 (def y (rand.u 10))
 (def z (rand.u 10)) ;; z is not useful in the regression:
 (def reg (regress (list x y z) (+ x (* 2 y) (rand.u 10))))
 (send reg :mallows)

 (def stuff (cmv (send reg :mallows)))
 (def sizes (second stuff))
 (def names (third stuff))
 (point-labels names)
 (plot sizes (first stuff))
 (send my-plot :add-plot-item (list 'function {x} 'black 'lines))

 (def stuff (cmv (send reg :mallows :r-squared t)))
 (plot sizes (first stuff))
"
  (let* (
	 (x (send self :x))
	 (p (second (size x)))
	 (vars (butlast (rest (all-subsets (column-list x)))))
	 (sizes (mapcar #'size vars))
	 (sizes (if (send self :intercept) (1+ sizes) sizes))
	 (ss (send self :s-squared))
	 (y (send self :y))
	 (n (length y))
	 result
	 (names
	  (mapcar #'list-to-string (butlast (rest (all-subsets (iseq p))))))
	 )
    (print r-squared)
    (setq
     result
     (mapcar
      (lambda(x)
	(let* (
	       (reg (regress x y :print nil))
	       (resss (send reg :resss))
	       (p (length (send reg :beta)))
	       )
	  (print r-squared)
	  (if r-squared (send reg :r-squared) 
	    (- (/ resss ss) (- n (* 2 p))))
	  ))
      vars
      )
     )
    (if verbose (mprint (bind-columns names result)))
    (values result sizes names)
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 ridge-regression
;;
(defun ridge-regression(x y &key ridge intercept (factor 1))
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          ridge-regression Doc:
;;
;; Author: Andy Long
;; Created: Thu May 8 13:21:15 CDT 1997
;;
;; Description/Arguments:

       (ridge-regression x y &key ridge intercept (factor 1)))
  
	If ridge is not specified, then the suggested value of

		ridge = p(s^2)/((norm beta)^2)

 (A. E. Hoerl, et al., \"Ridge regression: some simulations,\" Communications
 in Statistics, 4, 1975, 105-123, as cited in Draper and Smith, Applied
 Regression Analysis, 2nd Edition).

;; Notes:

;; See Also:

;; Examples:

 (def x1 (rand.u 20))
 (def x2 (rand.u 20))
 (def y (mapcar (lambda(x y z) (+ x (* 2 y) z)) x1 x2 (rand.u 20)))
 (def reg (regress (list x1 x2) y))
 (ridge-regression (list x1 x2) y)

 (def x1 (list 7 1 11 11 7 11 3 1 2 21 1 11 10))
 (def x2 (list 26 29 56 31 52 55 71 31 54 47 40 66 68))
 (def x3 (list 6 15 8 8 6 9 17 22 18 4 23 9 8))
 (def x4 (list 60 52 20 47 33 22 6 44 22 26 34 12 12))
 (def x5 
      (/ (list 785 743 1043 876 959 1092 1027 725 931 1159 838 1133 1094) 10))
 (ridge-regression (list x1 x2 x3 x4) x5)
 (ridge-regression (list x1 x2 x3 x4) x5 :ridge .5)

 ;; This reproduce results of pages 316-317, D&S:
 (lines on)
 (def xs (rseq 0 1 100))
 (def ridges (ridge-regression (list x1 x2 x3 x4) x5 :ridge xs))
 (title \"Page 316\")
 (plot xs (mapcar #'first ridges))
 (plot xs (mapcar #'second ridges))
 (send my-plot :add-plot-item
    (list 'lines xs (mapcar #'third ridges) 'black)
    (list 'lines xs (mapcar #'fourth ridges) 'black)
    (list 'lines xs (mapcar #'fifth ridges) 'black)
    )
"
  (let* (
	 (n (length y))
	 (ones (repeat 1 n))
	 (x (turn-it-into-a-matrix x :from 'ridge-regression))
	 (means (mapcar #'mean (column-list x)))
	 (sds
	  (sqrt
	   (mapcar
	    (lambda(x mean) (sum (square (- x mean))))
	    (column-list x) means
	    )
	   )
	  )
	 (xmat (apply #'bind-columns
		      (mapcar
		       (lambda(x mean s) (/ (- x mean) s))
		       (column-list x) means sds)))
	 
	 (xmat (bind-columns (/ ones factor) xmat))
	 (sds (combine 1 sds))

	 (p (second (size xmat)))
	 (n-p (- n p))
	 (xmat-inv (pseudo-inverse xmat))
	 (prod (matmult xmat-inv (transpose xmat-inv)))
	 (beta (matmult xmat-inv y))

	 (svd (svd xmat))
	 (u (first svd))
	 (v (third svd))
	 (lambda (diagonal-to-matrix (second svd)))

	 (uut (matmult u (transpose u)))
	 (resss (square (norm (matmult (- (id n) uut) y))))
	 (regss (square (norm (matmult (- uut (/ (ones n n) n)) y))))
	 (ms (/ regss (1- p)))
	 (s-squared (/ resss n-p))
	 (f (/ ms s-squared))
	 (se 
	  (sqrt
	   (*
	    (diagonal (matmult xmat-inv (transpose xmat-inv)))
	    s-squared
	    )
	   )
	  )
	 (ridge
	  (if ridge ridge
	    (/ (* (1- p) s-squared) (dot (rest beta) (rest beta)))))
	 result
	 )

    (setq result
	  (if (numberp ridge)
	      (/ (matmult (inverse (+ (id p) (* ridge prod))) beta) sds)
	    (mapcar
	     (lambda(x) (/ (matmult (inverse (+ (id p) (* x prod)))beta) sds))
	     ridge
	     )
	    )
	  )
    (if (numberp ridge)
	(setf
	 (elt result 0)
	 (- (mean y) (sum (* (rest result) means)))
	 )
      (mapcar
       (lambda(x)
	 (setf (elt x 0) (- (mean y) (sum (* (rest x) means)))))
       result
       )
      )
    (values result
     (/ beta sds) beta se regss resss ms s-squared f ridge)
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 press
;;
(defun press(x y &rest reg-stuff)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          press Doc:
;;
;; Author: Andy Long
;; Created: Wed May 14 15:43:55 CDT 1997
;;
;; Description/Arguments:

       (press x y &rest reg-stuff)
  
	Took this from Draper and Smith, Applied Regression Analysis, 2nd
Edition, p. 325. The idea is to successively remove one row from the matrix,
run the regression, and then estimate y for that row. We return the squared
differences,
			(yhat_i-y_i)^2.

	Reg-stuff is any option that you might send to regression-model,
e.g. :intercept nil. The only default is :print nil.

;; Notes:

;; See Also:

;; Examples:
 ;; Hald data from Appendix B of Draper and Smith:
 (def x1 (list 7 1 11 11 7 11 3 1 2 21 1 11 10))
 (def x2 (list 26 29 56 31 52 55 71 31 54 47 40 66 68))
 (def x3 (list 6 15 8 8 6 9 17 22 18 4 23 9 8))
 (def x4 (list 60 52 20 47 33 22 6 44 22 26 34 12 12))
 (def x5 
      (/ (list 785 743 1043 876 959 1092 1027 725 931 1159 838 1133 1094) 10))
 (press (bind-columns x1 x2 x3 x4) x5)

 (def result (press (list x1 x2) x5))
 (plot result)
 (sum result)

 (def result (press (bind-columns x1 x2 x3 x4) x5))
 (plot result)
 (sum result)

 (title \"Plot of results in Table 6.2, p. 325\")
 (def result
   (mapcar
     (lambda(x) (sum (press x x5)))
     (rest (all-subsets (list x1 x2 x3 x4)))
     )
   )
 (mprint (bind-columns (rest (all-subsets '(x1 x2 x3 x4))) result))
 (plot result)

 ;; If you want to see the results of the 13 regressions, add the :print t
 ;; option: 
 (def result (press (list x1 x2) x5 :print t))
"
  (let* (
	 (x (turn-it-into-a-matrix x :from 'press))
	 (n (first (size x)))
	 keepers row reg result
	 (reg-stuff (if (first (catch-keys '(:print) reg-stuff))
			reg-stuff
		      (smoosh (list reg-stuff (list :print nil)))))
	 )
    (dotimes
     (i n)
     (setq
      keepers (remove i (iseq n))
      row (row (1+ i) x)
      ynot (select y (list i))
      reg (apply #'regress
		 (smoosh (list 
			  (list (select-rows x keepers) (select y keepers))
			  reg-stuff)))
      result
      (combine
       result
       (square (- ynot (funcall (send reg :regression-function) row)))))
     )
    (rest result)
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 pca-regression
;;
(defun pca-regression(x y &key percent use-variables verbose)
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          pca-regression Doc:
;;
;; Author: Andy Long
;; Created: Thu May 8 13:21:15 CDT 1997
;;
;; Description/Arguments:

       (pca-regression x y &key percent use-variables verbose)
  
See Draper and Smith, Applied Regression Analysis, 2nd Edition, p. 327. This 
Returns the regression-model, and the eigenvectors, eigenvalues, and scores in
a values statement.

;; Notes: The transformations are as follows:

	X -> Z = (X - Xbar)*(diagonal (/ sigma))*(sqrt n-1))

	  -> SVD(Z) = U L V^T

	  Scores:       (sqrt n-1)*U
	  Eigenvalues:  (square L)
	  Eigenvectors: V

	Then we regress Y on Scores*L.

	The coefficients in terms of the original X variables are then

		new-beta = (diagonal (/ sigma)) V (rest beta)

		beta_0 - Xbar*new-beta + X*new-beta

;; See Also:

;; Examples:
 ;; This is the Hald data from Draper and Smith (appendix B):
 (def x1 (list 7 1 11 11 7 11 3 1 2 21 1 11 10))
 (def x2 (list 26 29 56 31 52 55 71 31 54 47 40 66 68))
 (def x3 (list 6 15 8 8 6 9 17 22 18 4 23 9 8))
 (def x4 (list 60 52 20 47 33 22 6 44 22 26 34 12 12))
 (def x5 
      (/ (list 785 743 1043 876 959 1092 1027 725 931 1159 838 1133 1094) 10))

 (def reg (pca-regression (list x1 x2 x3 x4) x5))

 ;; You can catch the eigenvectors, eigenvalues, and scores, since they are
 ;; returned using the values statement. I use cmv to catch them to a list,
 ;; then parcel out the list values:
 (def stuff (cmv (pca-regression (list x1 x2 x3 x4) x5)))
 (def reg (first stuff))
 (def eigvecs (first (second stuff)))
 (def scores (second (second stuff)))
 (def eigvals (third (second stuff)))
 (send reg :anova-sequential)

 ;; Using all the vectors (i.e. percent 100):
 (def reg (pca-regression (list x1 x2 x3 x4) x5 :percent 100))
 (def z-times-sqrt-lambda (send reg :x))
 (mprint (matmult z-times-sqrt-lambda (diagonal (/ (sqrt eigvals)))))
 ;; equals the scores:
 (mprint scores)

 ;; Reproduces the results of p. 331 (note that W_2 in 6.9.10 should be W_3):
 (def reg (pca-regression (list x1 x2 x3 x4) x5
 		:percent 100
		:use-variables '(0)))
 (def reg (pca-regression (list x1 x2 x3 x4) x5
 		:percent 100
		:use-variables '(0 2)))
 (def reg (pca-regression (list x1 x2 x3 x4) x5
 		:percent 100
		:use-variables '(0 1 2)))

 (def reg (pca-regression (list x1 x2 x3 x4) x5 :use-variables '(0 2)))


 (def x (rand.u 20))
 (def y (+ x (rand.u 20)))
 (def z (+ (* x 2) y (rand.u 20)))
 (def reg (regress (list x y) z))
 (def reg (pca-regression (list x y) z))
 (def reg (pca-regression (list x y) z :use-variables '(0)))
"
  (let* (
	 (n (length y))
	 (x (turn-it-into-a-matrix x :from 'pca-regression))
	 (means (mapcar #'mean (column-list x)))
	 (sds
	  (sqrt
	   (mapcar
	    (lambda(x mean) (sum (square (- x mean))))
	    (column-list x) means
	    )
	   )
	  )
	 (xmat (apply #'bind-columns
		      (mapcar
		       (lambda(x mean s) (/ (- x mean) s))
		       (column-list x) means sds)))

	 (svd (svd xmat))
	 (eigenvectors (third svd))
	 (eigenvalues (coerce (square (second svd)) 'list))
	 (scores (* (sqrt (1- n)) (first svd)))

	 (proportion (/ (if percent percent 99) 100))
	 (total (sum eigenvalues))
	 (ncols (second (size xmat)))
	 (sum 0)
	 (p
	  (let (
		(val
		 (dotimes
		  (i ncols)
		  (setq sum (+ sum (/ (elt eigenvalues i) total)))
		  (if (>= sum proportion) (return (1+ i)))
		  )
		 )
		)
	    (if val val ncols)))
	 (use-variables (if use-variables use-variables (iseq p)))
	 reg result
	 )
    (if verbose
	(let ()
	  (mprint eigenvectors)
	  (mprint scores)
	  (print eigenvalues)
	  ))
    (setq reg
	  (regress
	   (matmult
	    (select-columns scores use-variables)
	    (diagonal (sqrt (select eigenvalues use-variables)))
	    )
	   y)
	  )

    (setq
     result
     (* (sqrt (1- n))
	(matmult
	 (select-columns
	  (matmult (diagonal (/ sds)) eigenvectors)
	  use-variables
	  )
	 (matrix (list (length use-variables) 1) (rest (send reg :beta)))
	 )
	))
    (setq result
	  (combine
	   (- (first (send reg :beta)) (sum (* result means)))
	   result
	   )
	  )
    (format t "~%Regression coefficients in terms of original variables:
Intercept: ~20t~,5g" (first result))
    (mapcar (lambda(x y) (format t "~%X~a: ~20t~,5g" x y))
	    (mapcar #'num-to-string (iseq (length (rest result))))
	    (rest result))
    (format t "~2%")
    (values reg (list eigenvectors scores eigenvalues))
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 step-wise
;;
(defun step-wise
  (x y &key (intercept t) already predictor-names kept n (alpha .05))
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          step-wise Doc:
;;
;; Author: Andy Long
;; Created: Thu May 15 12:58:59 CDT 1997
;;
;; Description/Arguments:

   (step-wise
       x y &key (intercept t) already predictor-names kept n (alpha .05))
  
	Step-wise is a regression procedure which starts with the best single
variable (as measured by the probability of its f-statistics), and adds in
additional variables until it can't find a single additional variable which
will give a significant (at the alpha level) f-statistic).

;; Notes: Draper and Smith, Applied Regression Analysis, 2nd Edition no longer
;; includes a section on stepwise regression (because it's too common, they
;; say!). They do describe the method in section 6.4, p. 307.

;; See Also:

;; Examples:
 (def x1 (list 7 1 11 11 7 11 3 1 2 21 1 11 10))
 (def x2 (list 26 29 56 31 52 55 71 31 54 47 40 66 68))
 (def x3 (list 6 15 8 8 6 9 17 22 18 4 23 9 8))
 (def x4 (list 60 52 20 47 33 22 6 44 22 26 34 12 12))
 (def x5
      (/ (list 785 743 1043 876 959 1092 1027 725 931 1159 838 1133 1094) 10))

 (step-wise (list x1 x2 x3 x4) x5) ;; alpha of .05:
 (step-wise (list x1 x2 x3 x4) x5 :alpha .1)

 ;; Notice that when we increase alpha to .5, we lose x2!
 (step-wise (list x1 x2 x3 x4) x5 :alpha .5)

 ;; When you'll take any old alpha, everyone gets into the act:
 (step-wise (list x1 x2 x3 x4) x5 :alpha 1)

 (step-wise (list x1 x2 x3 x4) x5 :intercept nil)
"
  (let* (
	 (x (turn-it-into-a-matrix x :from 'step-wise))
	 (singles
	  (mapcar
	   (lambda(x)
	     (send
	      (regress (if already (bind-columns x already) x) y
		       :print nil :intercept intercept)
	      :r-squared)
	     )
	   (column-list x)
	   )
	  )
	 (best (find-max singles))
	 (best-x (elt (column-list x) best))
	 (reg (regress (if already (bind-columns already best-x) best-x)
		       y :print nil :intercept intercept))
	 (vars (if intercept (1- (length (send reg :beta)))
		 (length (send reg :beta))))
	 (f (if (= vars 1)
		(send reg :prob-f)
	      (send reg :hypothesis (select-rows (id vars) '(0)))))
	 (p (second (size x)))
	 (n (if n n p))
	 (predictor-names
	  (if predictor-names predictor-names
	    (mapcar (lambda(x) (strcat "X" (num-to-string (1+ x)))) (iseq p))))
	 (kept (if kept (combine kept (elt predictor-names best))
		 (list (elt predictor-names best))))
	 (predictor-names (select predictor-names (remove best (iseq p))))
	 )
    (let* (
	   ;; Need to check if we want to keep all the variables.
	   ;; We test each partial f:
	   (fs (if already t nil))
	   (already (if already (bind-columns already best-x)
		      (bind-columns best-x)))
	   (fs (if fs
		   (send (regress already y :print nil :intercept intercept)
			 :anova-sequential
			 :partial t :verbose nil)))
	   (keepers (if fs (where fs alpha #'<) (list 0)))
	   (already (if keepers (select-columns already keepers)))
	   )
      (setq kept (select kept keepers))
      (if already
	  (if (> p 1)
	      (step-wise
	       (select-columns x (remove best (iseq p))) y
	       :already already
	       :kept kept
	       :predictor-names predictor-names
	       :n n
	       :alpha alpha
	       :intercept intercept
	       )
	    (regress already y :predictor-names kept :intercept intercept))
	nil)
      )
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                                 backward-step-wise
;;
(defun backward-step-wise
  (x y &key (intercept t) predictor-names (alpha .05))
     
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          backward-step-wise Doc:
;;
;; Author: Andy Long
;; Created: Thu May 15 12:58:59 CDT 1997
;;
;; Description/Arguments:

   (backward-step-wise
      (x y &key (intercept t) predictor-names (alpha .05))
     
	Backward-step-wise is a regression procedure which starts with all the
 variables and removes them one at a time if they are judged to be unnecessary
 (as measured by the probability of their f-statistics).

;; Notes: Draper and Smith, Applied Regression Analysis, 2nd Edition, section
;; 6.3, p. 305.

;; See Also:

;; Examples:
 (def x1 (list 7 1 11 11 7 11 3 1 2 21 1 11 10))
 (def x2 (list 26 29 56 31 52 55 71 31 54 47 40 66 68))
 (def x3 (list 6 15 8 8 6 9 17 22 18 4 23 9 8))
 (def x4 (list 60 52 20 47 33 22 6 44 22 26 34 12 12))
 (def x5
      (/ (list 785 743 1043 876 959 1092 1027 725 931 1159 838 1133 1094) 10))

 (backward-step-wise (list x1 x2 x3 x4) x5)
 (backward-step-wise (list x1 x2 x3 x4) x5 :alpha .1)
 (backward-step-wise (list x1 x2 x3 x4) x5 :alpha .5)

 ;; When you'll take any old alpha, everyone gets into the act:
 (backward-step-wise (list x1 x2 x3 x4) x5 :alpha 1)

 ;; you can turn off the intercept term:
 (backward-step-wise (list x1 x2 x3 x4) x5 :intercept nil)
 (backward-step-wise (list (repeat 1 (length x1)) x1 x2 x3 x4) x5
     :intercept nil)
"
  (let* (
	 (x (turn-it-into-a-matrix x :from 'backward-step-wise))
	 (p (second (size x)))
	 (reg (regress x y :print nil :intercept intercept))
	 (fs
	  (if (= p 1) (list (send reg :prob-f))
	    (mapcar
	     (lambda(x) (send reg :hypothesis (select-rows (id p) (list x))))
	     (iseq p)
	     )
	    )
	  )
	 (best (find-max fs))
	 (f (elt fs best))
	 (predictor-names
	  (if predictor-names predictor-names
	    (mapcar (lambda(x) (strcat "X" (num-to-string (1+ x)))) (iseq p))))
	 )
    (if (> f alpha)
	(if (> p 1)
	    (backward-step-wise
	     (select-columns x (remove best (iseq p)))
	     y
	     :predictor-names (select predictor-names (remove best (iseq p)))
	     :alpha alpha
	     :intercept intercept
	     )
	  nil
	  )
      (regress x y :predictor-names predictor-names :intercept intercept)
      )
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmeth regression-model-proto :menu(&optional title)
  "
;; Description: regression-model-proto :menu

	Installs a menu that permits easy access to the usual things that an
inquiring mind wants to know about a regression: plots, coefficients, anova,
that sort of thing.

;; Examples:
 (try-out regress)
 (send exn :menu \"exn\")
"
  (let* (
	 (title (if title title
		  (get-string-dialog "Name of Model:" :initial "Reg")))
	 (RXridge (send menu-proto :new (if title title "REG")))

	 (nvars (send self :num-coefs))
	 (nvars (if (send self :intercept) (1- nvars) nvars))

         (plot (send menu-item-proto
		     :new "Plot regression line"
		     :action
		     (lambda() (send (send self :plot) :plot-item))))

         (plot-2d (send menu-item-proto
		     :new "Plot regression surface"
		     :action
		     (lambda() (send self :plot-2d))))

         (plot-res (send menu-item-proto
			 :new "Plot Residuals against True "
			 :action
			 (lambda()
			   (lines off)
			   (let (
				 (myplot
				  (plotter
				   (send self :y) (send self :residuals))))
			     (send myplot :variable-label '(0 1)
				   '("True Y value" "Residual"))
			     (send myplot :plot-item)
			     ))))
         (plot-est (send menu-item-proto
			 :new "Plot True against Estimates "
			 :action
			 (lambda()
			   (lines off)
			   (let (
				 (myplot
				  (plotter
				   (send self :y) (send self :estimates))))
			     (send myplot :variable-label '(0 1)
				   '("Estimates" "True"))
			     (send myplot :add-plot-item
				   '(function (lambda(x) x) black lines))
			     (send myplot :plot-item)
			     ))
			 )
		   )

         (beta (send menu-item-proto
		     :new "Beta"
		     :action (lambda() (print (send self :beta))(terpri))))
         (se (send menu-item-proto
		   :new "Standard Errors"
		   :action (lambda() (print (send self :se))(terpri))))
         (display (send menu-item-proto
			:new "Display Regression Results"
			:action (lambda () (send self :display))))
         (anova (send menu-item-proto
		      :new "Anova"
		      :action (lambda() (send self :anova))))
	 
         (anovas (send menu-item-proto
		       :new "Anova Sequential"
		       :action (lambda() (send self :anova-sequential))))
	 
         (anovap (send menu-item-proto
		       :new "Anova Sequential Partial"
		       :action
		       (lambda() (send self :anova-sequential :partial t))))
         (ridgereg (send menu-item-proto
			 :new "Ridge-Regression"
			 :action
			 (lambda() 
			   (lines on)
			   (let* ((xs (rseq 0 1 100))
				  (ridges
				   (ridge-regression
				    (send self :x) (send self :y) :ridge xs))
				  (myplot (plotter xs (mapcar #'second ridges)))
				  )
			     (mapcar
			      (lambda(x)
				(send my-plot :add-plot-item
				      (list 'lines xs x 'black)
				      ))
			      (rest (rest (transpose ridges)))
			      )
			     (send myplot :update-plot)
			     )
			   )))
	 (backwise (send menu-item-proto
			 :new "Backward Step Regression"
			 :action
			 (lambda() 
			   (format t "~%Backward Step Regression:~%")
			   (backward-step-wise
			    (send self :x) (send self :y)
			    :intercept (send self :intercept))
			   )
			 ))
 	 (forewise (send menu-item-proto
			 :new "Forward Step Regression"
			 :action
			 (lambda() 
			   (format t "~%Forward Step Regression:~%")
			   (step-wise
			    (send self :x) (send self :y)
			    :intercept (send self :intercept))
			   )
			 ))
         (dash0     (send dash-item-proto :new))
         (killmenu  (send menu-item-proto :new "Remove this menu"
                          :action #'(lambda () 
				      (send RXridge :remove)
				      ))
		    )
	 )
    (send RXridge :append-items
	  beta se
	  dash0
	  display anova anovas anovap
	  dash0
	  (if (= 1 nvars) plot (if (= 2 nvars) plot-2d plot))
	  plot-res
	  plot-est
	  dash0
	  ridgereg
	  backwise
	  forewise
	  dash0
	  killmenu)
    (send RXridge :install)
    RXridge
    )
  )
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                         remove-menus-on-error-bar
;;
(defun remove-menus-on-menu-bar()
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          remove-menus-on-menu-bar Doc:
;;
;; Author: Andy Long
;; Created: Tue May 27 15:24:00 CDT 1997
;;
;; Description/Arguments:

       (remove-menus-on-menu-bar)
  
	Once a menu bar is established, this removes all menus from the
bar. Then drinks are served, on the house.

;; Notes: I'm not sure that this will work on anything other than unix
;; platform. It uses XLISP::*FAKE-MENU-BAR*.

;; Examples:
 (try-out :menu :proto regression-model-proto)
 (remove-menus-on-menu-bar)
 ;; Now we'll put one back, in case you want to try to remove it again!
 (send exn :menu \"Exn\")
"
  (mapcar (lambda(x) (send x :remove)) (send *fake-menu-bar* :menus))
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
