;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;	Runga-Kutta for general n-dimensional first order DE's
;;
;; From Burden and Faires, 'Numerical Analysis', 5th Edition, p. 259
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun rk-4 (f ynot start end num &key non-autonomous)
  "
;; Description: 

	(rk-4 f ynot start end num &key non-autonomous)

	rk-4 is Runga-Kutta 4, for general n-dimensional first order DE's.

;; Arguments:

	f - the right-hand side functions. It takes a single argument (see the
two dimensional example below to see how a higher order case is handled).
	
	ynot - the initial value(s) of the variables

	start - initial time
	end - end time
	num - the number of steps (+1) to take in going from start to end.
		i.e., (rseq 0 2 11) is (list 0 .2 .4 ... 2) - 10 steps...
;; Options:

	Set non-autonomous to true if the time is explicitly included in the
equations. Also make f a function of two variables, one a list of the
independent variables, and the second the time variable. See the example below.

;; Examples:
 ;; Let's turn the lines off so that we can see the points:
 (lines off)

 ;;			One-Dimensional examples:

 (title \"y'(x)=-1, where y_0=1\")
 (func f(x) -1)
 (def points (rk-4 #'f '(1) 0 3 20))
 (def points (mapcar #'first points))
 (plot (rseq 0 3 20) points)

 (title \"y'(x)=-y, where y_0=1\")
 (func f(x) -x)
 (def points (rk-4 #'f '(1) 0 3 20))
 (def points (mapcar #'first points))
 (plot (rseq 0 3 20) points)

 ;;			Two-Dimensional examples:

 (defun f (vals)
  (let* (
	 (x (first vals))
	 (y (second vals))
         (xval (- x (^ x 2) (* x y)))
         (yval (- (* x y) (/ y 2) (^ y 2)))
         )
    (list xval yval)
    )
  )

 (title \"f(x(t),y(t)), where (x_0,y_0)=(.1 .2)\")
 (def points (rk-4 #'f '(.1 .2) 0 100 200))
 (def xvals (mapcar #'first points))
 (def yvals (mapcar #'second points))
 (plot xvals yvals)

 ;;			Non-autonomous examples:

 (title \"y'(x)=-y/x, y(1)=1 (y(x)=1/x)\")
 (func f(x h) -x/h)
 (def points (rk-4 #'f '(1) 1 3 20 :non-autonomous t))
 (def points (mapcar #'first points))
 (plot (rseq 1 3 20) points)

 ;; Checks against Burden and Faires 'Numerical Analysis'....
 (title \"y'(x)=y-t^2+1, y(0)=.5\")
 (func f(x h) x-h^2+1)
 (def points (rk-4 #'f '(.5) 0 2 11 :non-autonomous t))
 (def points (mapcar #'first points))
 (plot (rseq 0 2 11) points)
 (send my-plot :add-plot-item
       (list 'function (func y(tt) (square (tt+1))-.5*(exp tt)) 'blue 'lines))
"
  (let* (
	 (h (/ (- end start) (1- num)))
	 (half-h (/ h 2))
	 (timesteps (rseq start end num))
	 (yvalues (list ynot))
	 (w ynot)
	 )
    (if non-autonomous 
	(dotimes
	 (i (1- num))
	 (let* (
		(time (elt timesteps i))
		(time+ (+ time half-h))
		(k1 (funcall f w time))
		(k2 (funcall f (+ w (* k1 half-h)) time+))
		(k3 (funcall f (+ w (* k2 half-h)) time+))
		(k4 (funcall f (+ w (* k3 h)) (+ time h)))
		(small-change (* h (/ (+ k1 (* 2 (+ k2 k3)) k4) 6.0)))
		)
	   (setq 
	    w (+ w small-change)
	    yvalues (append yvalues (list w))
	    )
	   )
	 )
      (dotimes
       (i (1- num))
       (let* (
	      (k1 (funcall f w))
	      (k2 (funcall f (+ w (* k1 half-h))))
	      (k3 (funcall f (+ w (* k2 half-h))))
	      (k4 (funcall f (+ w (* k3 h))))
	      (small-change (* h (/ (+ k1 (* 2 (+ k2 k3)) k4) 6.0)))
	      )
	 (setq 
	  w (+ w small-change)
	  yvalues (append yvalues (list w))
	  )
	 )
       )
      )      
    yvalues
    )
  )
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;  Runga-Kutta-Fehlberg for general n-dimensional first order DE's
;;
;;  From Burden and Faires, 'Numerical Analysis', 5th Edition, p. 266
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun rk-f (f
	     ynot start end num
	     &key
	     non-autonomous
	     (hmin 1e-6)
	     (hmax (/ (- end start) (1- num)))
	     (tol 1e-5)
	     )
  "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                          rk-f Doc:
;;
;; Author: Andy Long
;; Created: Fri Jan 23 00:27:55 1998
;;
;; Description/Arguments:

       (rk-f f
	     ynot start end num
	     &key
	     non-autonomous
	     (hmin 1e-6)
	     (hmax (/ (- end start) (1- num)))
	     (tol 1e-5)
	     )
  
  Runga-Kutta-Fehlberg (basically rk-4 with step-size control) for general
n-dimensional first order DE's.

;; Returns: a list of lists of times and y-values.

;; Notes: From Burden and Faires, 'Numerical Analysis', 5th Edition, p. 266.

;; See Also: rk-4, euler-nd

;; Examples:

 ;;		Autonomous (no time dependence), one-d:

 (title \"y'(x)=-1, where y_0=1\")
 (func f(x) -1)
 (def points (rk-f #'f '(1) 0 3 20))
 (def times (mapcar #'first points))
 (def points (mapcar #'second points))
 (lines off)
 (plot times points)

 (title \"y'(x)=-y, where y_0=1\")
 (func f(x) -x)
 (def points (rk-f #'f '(1) 0 3 20))
 (def times (mapcar #'first points))
 (def points (mapcar #'second points))
 (plot times points)

;;			Two-Dimensional autonomous:

 (defun f (vals)
  (let* (
	 (x (first vals))
	 (y (second vals))
         (xval (- x (^ x 2) (* x y)))
         (yval (- (* x y) (/ y 2) (^ y 2)))
         )
    (list xval yval)
    )
  )

 (title \"f(x(t),y(t)), where (x_0,y_0)=(.1 .2)\")
 (def points (rk-f #'f '(.1 .2) 0 100 200))
 (def times (mapcar #'first points))
 (def xvals (mapcar #'second points))
 (def yvals (mapcar #'third points))
 (plot xvals yvals)
 (plot times xvals yvals)

 ;;				Non-autonomous:

 ;; Checks against Burden and Faires 'Numerical Analysis'....
 (title \"y'(x)=y-t^2+1, y(0)=.5\")
 (func f(x h) x-h^2+1)
 (def points (rk-f #'f '(.5) 0 2 11 :non-autonomous t :hmin .01 :hmax .25))
 (pprint points)
"
  (let* (
	 (time start)
	 (w ynot)
	 (h hmax)
	 (yvalues (list (combine time ynot)))
	 (constants
	  (list (/ 3 32)	;; c0
		(/ 9 32)	;; c1
		(/ 3 8)		;; c2
		(/ 1932 2197)	;; c3
		(/ -7200 2197)	;; c4
		(/ 7296 2197)	;; c5
		(/ 12 13)	;; c6
		(/ 439  216)	;; c7
		(/ 3680  513)	;; c8
		(/  -845 4104)	;; c9
		(/ -8  27)	;; c10
		(/ -3544 2565)	;; c11
		(/  1859 4104)	;; c12
		(/  -11  40)	;; c13
		(/ -128  4275)	;; c14
		(/ -2197 75240)	;; c15
		(/ 2  55)	;; c16
		(/ 25 216)	;; c17
		(/ 1408 2565)	;; c18
		(/ 2197 4104)	;; c19
		))
	 
	 (c0 (elt constants 0))
	 (c1 (elt constants 1))
	 (c2 (elt constants 2))
	 (c3 (elt constants 3))
	 (c4 (elt constants 4))
	 (c5 (elt constants 5))
	 (c6 (elt constants 6))
	 (c7 (elt constants 7))
	 (c8 (elt constants 8))
	 (c9 (elt constants 9))
	 (c10 (elt constants 10))
	 (c11 (elt constants 11))
	 (c12 (elt constants 12))
	 (c13 (elt constants 13))
	 (c14 (elt constants 14))
	 (c15 (elt constants 15))
	 (c16 (elt constants 16))
	 (c17 (elt constants 17))
	 (c18 (elt constants 18))
	 (c19 (elt constants 19))
	 
	 k1 k2 k3 k4 k5 k6 r delta
	 )
    (if non-autonomous 
	(do
	 ((h-over-4 (/ h 4) (/ h 4)))
	 ()
	 (setq
	  k1 (funcall f w time)
	  k2 (funcall f (+ w (* k1 h-over-4)) (+ time h-over-4))
	  k3 (funcall f 
		      (+ w (* c0 k1 h) (* c1 k2 h))
		      (+ time (* c2 h)))
	  k4 (funcall f 
		      (+ w (* c3 k1 h) (* c4 k2 h) (* c5 k3 h))
		      (+ time (* c6 h)))
	  k5 (funcall f 
		      (+ w (* c7 h k1) (* -8 k2 h) (* c8 h k3) (* c9 h k4))
		      (+ time h ))
	  k6 (funcall f 
		      (+ w (* c10 h k1) (* 2 k2 h) (* c11 h k3) (* c12 h k4)
			 (* c13 h k5))
		      (+ time (/ h 2)))
	  r (abs (+ (/ k1 360) (* c14 k3) (* c15 k4) (/ k5 50) (* c16 k6)))
	  r (if (listp r) (first r) r)
	  delta (if (= r 0) 4 (* .84 (^ (/ tol r) .25)))
	  )
	 
	 (if (<= r tol)
	     (setq
	      time (+ time h)
	      w (+ w (* h (+ (* c17 k1) (* c18 k3) (* c19 k4) (/ k5 -5))))
	      yvalues (append yvalues (list (combine time w)))
	      )
	   )
	 (setq h 
	       (min
		(* h (min (max .1 delta) 4))
		hmax
		))
	 (if (>= time end) (return nil)
	   (if (> (+ time h) end) (setq h (- end time))
	     (if (< h hmin)
		 (let ()
		   (print "minimum h exceeded")
		   (return nil)))))
	 )
      (do
       () ()
       (setq
	k1 (funcall f w)
	k2 (funcall f (+ w (* k1 (/ h 4))))
	k3 (funcall f (+ w (* c0 k1 h) (* c1 k2 h)))
	k4 (funcall f (+ w (* c3 k1 h) (* c4 k2 h) (* c5 k3 h)))
	k5 (funcall f (+ w (* c7 h k1) (* -8 k2 h) (* c8 h k3) (* c9 h k4)))
	k6 (funcall f (+ w (* c10 h k1) (* 2 k2 h) 
			 (* c11 h k3) (* c12 h k4) (* c13 h k5)))
	r (abs (+ (/ k1 360) (* c14 k3) (* c15 k4) (/ k5 50) (* c16 k6)))
	r (if (listp r) (first r) r)
	delta (if (= r 0) 4 (* .84 (^ (/ tol r) .25)))
	)
       (if (<= r tol)
	   (setq
	    time (+ time h)
	    w (+ w (* h (+ (* c17 k1) (* c18 k3) (* c19 k4) (/ k5 -5))))
	    yvalues (append yvalues (list (combine time w)))
	    )
	 )
       (setq h 
	     (min
	      (* h (min (max .1 delta) 4))
	      hmax
	      ))
       (if (>= time end) (return nil)
	 (if (> (+ time h) end) (setq h (- end time))
	   (if (< h hmin)
	       (let ()
		 (print "minimum h exceeded")
		 (return nil)))))
       )
      )
    yvalues
    )
  )
