(message "
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;                     Complex Numbers and Functions
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

        Xlisp-stat handles complex numbers/functions pretty well, although a
few things will 'break' when they encounter complex numbers.

        There are two ways in which to define a complex number:

        (complex 1 2)
        #C(1 2)
        
both define the complex number 1 + 2i:
")
        (complex 1 2)
        #C(1 2)
(message "")

(message "
        You manipulate complex numbers much the same way as you do real
numbers: you can still use calc to do arithmetic: e.g.,

        (calc #c(1 2) * #c(0 1)) ;; (1 + 2i) * i
        (calc (sqrt #c(1 2)))
")
        (calc #c(1 2) * #c(0 1))
        (calc (sqrt #c(1 2)))
(message "")

(message "
        You might wonder why we don't just define a constant i, like we do pi,
and be done with it. The problem is that 'i' is used very often as an index,
and so we can't steal it so easily. If you're going to be doing a lot of
calculations, you can do this yourself using def:

        (def i #c(0 1))

Then computations could be written as, for example, 

        (calc (1+2*i)*i) ;; (1 + 2i) * i
")
        (def i #c(0 1))
        (calc (1+2*i)*i)

(message "
        One could write a modulus function (which is pretty easy!), using the
functions 'realpart' and 'imagpart':

        (func modulus(z) (sqrt ((realpart z) ^ 2 + (imagpart z) ^ 2)))
 
        But xlisp-stat uses 'abs' (for absolute value) to accomplish the same
thing: to find the modulus of a complex number (that is, its length) you
type
        (calc (abs i))
        (calc (abs (1+2*i)))
")
        (calc (abs i))
        (calc (abs (1+2*i)))

(message "")
(message "
        In order to express a complex number in terms of polar coordinates, you
need both a modulus and an angle (called the argument).

        Finding the argument is a little more complicated. You need to make
sure that you end up in the right quadrant. Fortunately there's a built in
function for that:

        (argument #c(0 1))
        (argument #c(1 1))
        (argument #c(-1 -1))

Note that these values are given on the interval [-pi,pi].
")
        (argument #c(0 1))
        (argument #c(1 1))
        (argument #c(-1 -1))
(message "")
(message "
        Now let's just check that we get the same thing using rectangular and
polar coordinates for the complex number:

        (def z #c(.7 -.3))
        (calc (modulus z) * (exp (i * (argument z))))

")
        (def z #c(.7 -.3))
        (calc (modulus z) * (exp (i * (argument z))))

(message "")
(message "
        One interesting application of complex numbers is to find out 'where
the roots go' when a polynomial doesn't have real ones. Here's where they are:

        (func f(x) x^3+1) ;; just -1 is a real root, but there are two
			      ;; complex ones
        (title \"modulus function of f\")
        (func g(x y) (abs (f (x + (#c(0 1) * y)))))
        (new-plot) ;; initialize the plot
        (spin-points 20) ;; 20 by 20 for the surface
        (spin-domain -1.2 1.2 -1.2 1.2) ;; set the domain to ...
        (plot g)
")
        (new-plot)
        (title "modulus function of f")
        (func f(x) x^3+1)
        (func g(x y) (abs (f (x + (#c(0 1) * y)))))
        (spin-points 20)
        (spin-domain -1.2 1.2 -1.2 1.2)
        (plot g)

(message "Can you see the root -1 (i.e., -1 + 0i)?")

(message "
        We can use the same procedure, with minor modification, to look at the
real and imaginary parts, rather than the modulus:

        (new-plot) ;; initialize the plot
        (title \"real part of f\")
        (func g(x y) (realpart (f (x + (#c(0 1) * y))))) ;; real part function
        (spin-points 20) ;; 20 by 20 for the surface
        (spin-domain -1.2 1.2 -1.2 1.2) ;; set the domain to ...
        (plot g)
        (title \"imaginary part of f\")
        (func g(x y) (imagpart (f (x + (#c(0 1) * y))))) ;; imaginar part
        (spin-points 20) ;; 20 by 20 for the surface
        (spin-domain -1.2 1.2 -1.2 1.2) ;; set the domain to ...
        (plot g)
")
        (new-plot) ;; initialize the plot
        (title "real part of f")
        (func g(x y) (realpart (f (x + (#c(0 1) * y))))) ;; real part function
        (spin-points 20) ;; 20 by 20 for the surface
        (spin-domain -1.2 1.2 -1.2 1.2) ;; set the domain to ...
        (plot g)
        (title "imaginary part of f")
        (func g(x y) (imagpart (f (x + (#c(0 1) * y))))) ;; imaginar part
        (spin-points 20) ;; 20 by 20 for the surface
        (spin-domain -1.2 1.2 -1.2 1.2) ;; set the domain to ...
        (plot g)
(message "")

(message "
        To find complex roots, use muller rather than newton: with Muller you
need to give three initial guesses, but even crude ones will work. There are
three complex roots, so we use the :num-roots option:

        (muller f -4 0 4 :num-roots 3)
")
        (muller f -4 0 4 :num-roots 3 :tol 1e-12)
(message "Back to you, Bob.")
