(message "
	Eigenvalues and eigenvectors can be computed for many cases, although
the routines provided are not optimized for speed or size.  Once again, the
pedogogical issues outweigh the research issues!

	Even so, I think that you can be fairly confident of the results for
run-of-the-mill matrices.

	Xlispstat comes with the routine 'eigen', for the computation of
eigenvalues and eigenvectors in the symmetric case. I added some routines (from
Numerical Recipes) which extend them to the general case. Thus, the general
routines are called 'eigenvalues' and 'eigenvectors', and if the given matrix
is symmetric, then eigen is called; otherwise, the more general routines are
called.
")
(message "
	Let's get down to business then: here's the help for eigen:

  (help eigen)
")
(help eigen)
(message "")

(message "

	Now let's have some examples. Symmetric matrices, like the two
following, are handled by eigen: 

 (eigenvalues (id 3))

 ;; from Strang's Introduction to Applied Mathematics:
 (eigenvalues (matrix '(3 3) '(1 -1 0 -1 2 -1 0 -1 1))) ;; 0, 1, 3
")
(eigenvalues (id 3))
(eigenvalues (matrix '(3 3) '(1 -1 0 -1 2 -1 0 -1 1))) ;; 0, 1, 3
(message "")

(message "

	If we now consider an unsymmetric matrix, then we're using the
Numerical recipes routines balanc, elmhes, and hqr (which had been converted
from fortran into lisp using f2cl):

 (eigenvalues (matrix '(3 3) '(10 2 3 2 3 4 30 4 20)))

This matrix has two complex eigenvalues (a pair of complex conjugates):

   26.770040788119477 
   #C(3.114979605940258 -1.8763368152370816) 
   #C(3.114979605940258 1.8763368152370816) 
")
(def eigs 
     (mapcar #'print (eigenvalues (matrix '(3 3) '(10 2 3 2 3 4 30 4 20)))))

(message "

	Burden and Faires, authors of a popular text on numerical analysis,
 give a function which generates eigenvalues of known values, which we can use
 to test our routine (note that these are symmetric, again, so eigen is used):

 (defun special-tridiag(n a)
  (let (
	(initial (make-array (list n n) :initial-element 0))
	(diag (+ 1 (* 2 a)))
	)
    (mapcar (lambda(i) (setf (aref initial i i) diag)) (iseq n))
    (mapcar (lambda(i)
	      (setf
	       (aref initial (1+ i) i) (- a)
	       (aref initial i (1+ i)) (- a)
	       ))
	    (iseq (1- n)))
    (mprint initial)
    (print \"eigenvalues:\")
    (print
     (1+ (* 4 a 
            (square
	    (mapcar (lambda(i) (sin (/ (* pi i) (* 2 (1+ n))))) (iseq 1 n)))
            )))
    initial
    )
  )
")
(defun special-tridiag(n a)
  (let (
	(initial (make-array (list n n) :initial-element 0))
	(diag (+ 1 (* 2 a)))
	)
    (mapcar (lambda(i) (setf (aref initial i i) diag)) (iseq n))
    (mapcar (lambda(i)
	      (setf
	       (aref initial (1+ i) i) (- a)
	       (aref initial i (1+ i)) (- a)
	       ))
	    (iseq (1- n)))
    (mprint initial)
    (print "eigenvalues:")
    (print
     (1+ (* 4 a 
            (square
	    (mapcar (lambda(i) (sin (/ (* pi i) (* 2 (1+ n))))) (iseq 1 n)))
            )))
    initial
    )
  )
(message "")


(message "
	We can now check our routine against these:

 (eigenvalues (special-tridiag 2 1))
 (eigenvalues (special-tridiag 3 1))
 (eigenvalues (special-tridiag 4 1))
")	
(eigenvalues (special-tridiag 2 1))
(eigenvalues (special-tridiag 3 1))
(eigenvalues (special-tridiag 4 1))
(message "")

(message "

	On to eigenvectors: that's a little more complicated. I'll just provide
a few examples: you'll recall the (symmetric) strang matrix:

 ;; Strang's Introduction to Applied Mathematics:
 (def strang (matrix '(3 3) '(1 -1 0 -1 2 -1 0 -1 1)))
 ;; eigenvalues: 0, 1, 3
 (def eig (eigenvectors strang))

")
(def strang (matrix '(3 3) '(1 -1 0 -1 2 -1 0 -1 1)))
(def eig (eigenvectors strang))
(message "")

(message "

	Notice that the default is to print certain diagnostics: for example,
the trace of the matrix (sum of the diagonal elements) should equal the sum of
the eigenvalues.

	Eigenvectors returns a pair of lists, containing the eigenvalues and
eigenvectors. Thus, as I defined eig above,

eig

produces this pair:
")
eig
(message "")

(message "

	Again, the non-symmetric case:

 (eigenvectors (matrix '(3 3) '(10 2 3 2 3 4 30 4 20)))
")
(eigenvectors (matrix '(3 3) '(10 2 3 2 3 4 30 4 20)))
(message "")

(message "
	That's all for eigen! Back to you Bill.
")
